;;; ada-format.el - An Emacs indentation mode for editing Ada source.
;;;
;;; Gary E. Barnes          July 1994           geb@rational.com

(defconst ada-format-version "ada-formal.el, v0.2, 09/07/94")

(defconst af-byte-compiling t) ;; Set to nil for debugging
(eval-when-compile (setq af-byte-compiling t))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; AF-DEBUG - Debug controls.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defconst af-debug-flag nil
  "*Set to t for debug printout in the *edebug-trace* buffer.  Set to any
numeric value for debug printout plus a pause as we lex each token.  Set to
nil to get us to shut up.")

(eval-and-compile
  (if af-byte-compiling

      ;;--We are compiling the package.  Set up everything as a normal library.

      (progn

        (setq af-debug-flag     nil)
        (if (boundp 'edebug-all-defuns)
            (setq edebug-all-defuns nil)
          )
        (defmacro af-defvar         (&rest rest) (cons 'defvar         rest))
        (defmacro af-defsubst       (&rest rest) (cons 'defsubst       rest))

        ;;--If we are byte-compiling, then we are not debugging and we want
        ;;  all of these function calls to "vanish" and not take up room or
        ;;  runtime.

        (defsubst af-debug (note &optional position)
          );; af-debug

        (defmacro af-debug-check (check &optional or-else)
          or-else
          );; af-debug-check

        )

    ;;--We are not compiling the package.  Force all variables to be set as if
    ;;  they were consts and all inline functions are forced out-of-line.

    (progn

      (defmacro af-defvar         (&rest rest)    (cons 'defconst rest))
      (defmacro af-defsubst       (&rest rest)    (cons 'defun    rest))

      (let ((buffer (current-buffer)))
        (switch-to-buffer "*edebug-trace*")
        (switch-to-buffer buffer)
        (kill-buffer "*edebug-trace*")
        )

      ;;--not af-byte-compiling

      (defun af-big-debug ()
        (interactive)
        (setq af-debug-flag  (not (or af-debug-flag
                                      edebug-all-defuns))
              debug-on-error t)
        (let ((buffer (current-buffer)))
          (switch-to-buffer "ada-format.el")
          (if af-debug-flag
              (progn
                (if (not (boundp 'edebug-all-defuns))
                    (progn
                      (load-library "edebug")
                      (load-library "cust-print")
                      (install-custom-print-funcs)
                      (setq print-circle t)
                      )
                  )
                (if (not edebug-all-defuns) (edebug-all-defuns))
                (eval-current-buffer)
                (setq af-debug-flag 0.5)
                )
            (progn
              (if (not (boundp 'edebug-all-defuns))
                  (progn
                    (load-library "edebug")
                    (load-library "cust-print")
                    (install-custom-print-funcs)
                    (setq print-circle t)
                    )
                )
              (if edebug-all-defuns (edebug-all-defuns))
              (eval-current-buffer)
              (setq af-debug-flag nil)
              )
            )
          (switch-to-buffer buffer)
          )
        (setq edebug-trace   af-debug-flag
              debug-on-error af-debug-flag)
        (setq-default debug-on-error af-debug-flag)
        (message "af: Debugging is %s." af-debug-flag)
        );; af-big-debug

      ;;--not af-byte-compiling

      (defun af-toggle-debug ()
        "Turn on/off af-debug mode for debugging Ada Format mode."
        (interactive)
        (setq af-debug-flag (if af-debug-flag nil 0.5))
        (message "af-debug-flag is %S." af-debug-flag)
        );; af-toggle-debug

      ;;--not af-byte-compiling

      (defmacro af-debug (note &optional position)
        (if af-debug-flag
            (if position
                (` (af-debug-inside2 (, note) (, position)))
              (` (af-debug-inside (, note)))
              )
          )
        );; af-debug

      ;;--not af-byte-compiling

      (defun af-debug-inside2 (note position)
        (let ((start (point)))
          (goto-char position)
          (af-debug-inside note)
          (goto-char start)
          )
        );; af-debug-inside2

      ;;--not af-byte-compiling

      (defun af-debug-inside (note)
        (let ((buffer (current-buffer))
              (line   (af-what-line-number)))
          (message "%s" note)
          (if (numberp af-debug-flag)
              (progn
                (sit-for af-debug-flag)
                )
            )
          (switch-to-buffer "*edebug-trace*")
          (end-of-buffer)
          (insert (format "'%s' at line %d\n" note line))
          (switch-to-buffer buffer)
;;;       (y-or-n-p (concat note "; Continue?"))
          )
        );; af-debug-inside

      ;;--not af-byte-compiling

      (defmacro af-debug-check (check &optional or-else)
        (if af-byte-compiling
            or-else
          check
          )
        );; af-debug-check

      ;;--not af-byte-compiling

      (defun af-debug-error (&rest args)
        ;;--Put out a debug error message that stops everything.

        (split-window-vertically)
        (select-window (next-window))
        (switch-to-buffer "*ada-format-error*")
        (erase-buffer)
        (let ((msg (apply 'format args)))
          (insert msg)
          (beginning-of-buffer)
          (select-window (previous-window))
          (error "%s" msg)
          )
        );; af-debug-error

      ;;--not af-byte-compiling

      (defun af-debug-dump2 (list &optional elem)
        (if (setq elem (memq list done))
            (progn
              (insert (format " #%d" (length elem)))
              )
          (progn
            (setq done (cons list done))
            (cond ((null list)
                   (insert (format "#%d=nil" (length done))))
                  ((consp list)
                   (insert (format "#%d=(" (length done)))
                   (af-debug-dump2 (car list))
                   (while (setq list (cdr list))
                     (insert " ")
                     (af-debug-dump2 (car list))
                     )
                   (insert ")")
                   )
                  ((and (sequencep list)
                        (= 0 (length list)))
                   (insert (format "#%d=[" (length done))))
                  ((sequencep list)
                   (let ((len (length list))
                         (n 0))
                     (insert (format "#%d=[" (length done)))
                     (af-debug-dump2 (elt list 0))
                     (while (< (setq n (1+ n)) len)
                       (insert " ")
                       (af-debug-dump2 (elt list n))
                       )
                     (insert "]")
                     )
                   )
                  (t
                   (insert (format "#%d=%S" (length done) list)))
                  )
            )
          )
        );; af-debug-dump2

      ;;--not af-byte-compiling

      (defun af-debug-dump (name &rest args)
        (let ((done nil))
          (switch-to-buffer-other-window "*edebug-trace*")  
          (end-of-buffer)
          (if (stringp name)
              (insert name "\n"))
          (while args
            (af-debug-dump2 (car args))
            (insert "\n")
            (setq args (cdr args))
            )
          (sit-for 0)
          (select-window (previous-window))
          )
        );; af-debug-dump

      )
    );; (if (not af-byte-compiling) (progn
  );; eval-and-compile


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; AF-INDENT-<what> - indentation amounts
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-defvar af-indent-new-element ada-indent
  "*Default AF indentation amount for new statements, declarations, etc.")

(af-defvar af-indent-continuation ada-broken-indent
  "*Default AF indentation amount for continuation lines.")

(af-defvar af-indent-after-left-paren 1
  "*Default AF indentation amount for first element of a () group which comes
on a new line after the (.  Use a numeric value to get indentation as for
AAA (indent relative to the '(') and use a value of nil to indent as for
BBB (indent relative to the tight expression which precedes the '('):

    AAA.AAA (                BBB.BBB (
             FIRST);            FIRST);
");

(af-defvar af-indent-record-after-is af-indent-new-element
  "*Default AF indentation amount for a RECORD on a new line after an IS,
as in (TYPE x IS \\n RECORD ... END RECORD;).")

(af-defvar af-indent-block-name        (- (/ (1+ af-indent-new-element) 2))
  "*Indentation amount for a NAME:DECLARE style block name.")
(af-defvar af-indent-use-of-for        af-indent-continuation
  "*Indentation amount for USE of FOR xxx USE when USE is on separate line.")
(af-defvar af-indent-left-label        (- (/ (1+ af-indent-new-element) 2))
  "*Indentation amount for a <<abc>> label.")
(af-defvar af-indent-loop-name         (- (/ (1+ af-indent-new-element) 2))
  "*Indentation amount for a NAME:FOR style loop name.")
(af-defvar af-indent-middle-complex    af-indent-new-element
  "*Amount to indent the second half of complex tokens which are broken across
lines.  e.g. the IF of END IF")
(af-defvar af-indent-tight-expression  af-indent-continuation
  "*Amount to indent things like ID2 of ID1.ID2 when ID2 is on a separate line.
  Includes things like ID1\\n'\\nID2 and ID\\n(...)")
(af-defvar af-indent-range             af-indent-continuation
  "*Indentation amount for a RANGE keyword at the start of a line.")
(af-defvar af-indent-when              af-indent-new-element
  "*Indentation amount for a WHEN keyword (except for an EXIT WHEN).")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; AF-CONTROL-<what> - general control flags
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-defvar af-control-and-then-or-else-special t
  "*t if and-then/or-else clauses are supposed to line up inside ()'s, IF's
and whatnot.  nil if they are to be considered just-another-operator.")

(af-defvar af-control-indent-short-cut t
  "*t if indentation short cuts are ok, nil if not.  For example, we stop
searching when we find END IF as the first token on a line.  We assume that
the IF is also the first token on a line and that the END IF is indented the
same as the IF.  If a user has a coding style where this is not true, set
this to nil, or, turn off the af-flag-major-end bits for end tokens where
searching should not terminate this way.")

(af-defvar af-blink-matching-group-begin 0.1
  "*t if we are to 'blink' when we indent a group end or a group continue
token.  Set to a numeric value, eg. 1, 1.0, or 0.5, to change the blink time.
The default is 1 second.  Set to a cons with two integers, eg. (0 . 500)
if your Emacs doesn't have floating point and you want to wait for a
fraction of a second.  The two numbers are passed as the 1st & 2nd arguments
for the sit-for function.")

(af-defvar af-unix-eol nil
  "*Set to t if only \\n is to be recognized as an end of line indication.
Normal Ada syntax rules require that \\f\\n\\r\\v all be recognized as
end of line characters, whether used separately or together in combination.
However, on Unix, \\r and \\v are seldom used, and \\f is usually followed
by \\n to make the text \"print right\" on a printer or in an Emacs buffer.
If af-unix-eol is t, then \\f\\r\\v are treated simply as whitespace.
Set this to nil if all four characters are to be treated as end of line (which
also means treat them as end-of-comment as well).")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-defsubst af-testbit (data mask)
  ;;--Returns true if any of the mask bits are present in the data.
  (/= 0 (logand data mask))
  );; af-testbit

(af-defsubst af-zerobit (data mask)
  ;;--Returns true if none of the mask bits are present in the data.
  (= 0 (logand data mask))
  );; af-testbit


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; AF-TOKEN-LIST - Implements an abstraction which is a "slice" out of the
;;;; lexical stream which would represent the current buffer.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defconst af-token-list-mask (1- 256))   ;; Size of our token store.

(af-defvar af-token-symbols    (make-vector (1+ af-token-list-mask) nil))
(af-defvar af-token-points     (make-vector (1+ af-token-list-mask) nil))
(af-defvar af-token-flags      (make-vector (1+ af-token-list-mask) nil))
(af-defvar af-token-list-first 0)
(af-defvar af-token-list-next  0)

(defconst af-token-flag-bol              (lsh 1 0))
                                        ; Indicates that the token is
                                        ;  positioned at bol.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-defsubst af-lex-one ()

  ;;--Used to call the af-state-1 FSM.  We lex one lexeme and return.
  ;;
  ;;--Skip forward over all whitespace.

  (if af-unix-eol
      (skip-chars-forward " \t\f\r\v")
    (skip-chars-forward " \t")
    )

  ;;--Remember this point, we may need it for the indentation calculation.

  (setq lex-result-point (point))

  ;;--Actually lex and return the token that starts at (point).

  (af-debug-check 

   ;;--Do this when we are not byte compiling.

   (if (vectorp
        (setq lex-result-symbol (funcall (aref af-state-1 (following-char)))))
       lex-result-symbol
     ;;--The return value does not matter, but, optimizations we have
     ;;  put into place require that there *be* a non-nil token return value
     ;;  from all af-state-1 funcalls.
     (error "funcall on af-state-1 returned nil.")
     )

   ;;--Do this we are are byte compiling

   (setq lex-result-symbol (funcall (aref af-state-1 (following-char))))
   )

  );; af-lex-one

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun clear-af-token-list (bol-point)

  ;;--Sets the af-token-list to \"empty\".  The argument is a beginning-of-line
  ;;  point.  It indicates where lexing is to start when next/previous-af-token
  ;;  is called.

  (setq af-token-list-first 0
        af-token-list-next  1)
  (af-debug-check (progn
                    (fillarray af-token-symbols nil)
                    (fillarray af-token-points  nil)
                    (fillarray af-token-flags   nil)
                    )
                  )
  (aset af-token-points af-token-list-first bol-point)
  (aset af-token-flags  af-token-list-first 0)
  (aset af-token-points af-token-list-next  bol-point)
  (aset af-token-flags  af-token-list-next  af-token-flag-bol)
  );; clear-af-token-list

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun next-af-token-lex ( &optional          ; Local variables
                           lex-result-symbol
                           lex-result-point
                           lex-result-flag )

  ;;--Subroutine for next-af-token.  Called when we need to perform some
  ;;  lexing.
  ;;
  ;;--Goto wherever we stopped lexing in the past.  The af-token-flags
  ;;  value for this spot (basically the BOL flag) is already set.
  ;;  Continue to lex forward until we find a non-eol token or until we
  ;;  hit end-of-buffer.

  (setq lex-result-flag (aref af-token-flags af-token-list-next))
  (goto-char (aref af-token-points af-token-list-next))
  (while (and (or (eq (af-lex-one)      af-syntax-end-of-line)
                  (eq lex-result-symbol af-syntax-comment))
              (setq lex-result-flag af-token-flag-bol)
              (not (eobp)))
    )
  (aset af-token-symbols af-token-list-next lex-result-symbol)
  (aset af-token-points  af-token-list-next lex-result-point)
  (aset af-token-flags   af-token-list-next lex-result-flag)
  (af-debug-check
   (if (not (and (vectorp lex-result-symbol)
                 (numberp lex-result-point)
                 (numberp lex-result-flag)))
       (error "Something went wrong in the parse results.")
     )
   )

  ;;--Advance the next pointer.  If this conflicts with the first pointer
  ;;  then advance that point until we find a BOL token.

  (setq af-token-list-next (logand (1+ af-token-list-next)
                                   af-token-list-mask))
  (if (= af-token-list-next af-token-list-first)
      (while
          (setq af-token-list-first (logand (1+ af-token-list-first)
                                            af-token-list-mask)
                lex-result-flag      ;;; 2 setq's are faster than a progn+setq.
                                     ;;; This assigment is an optimization.
                (af-zerobit (aref af-token-flags af-token-list-first)
                            af-token-flag-bol))
        (if (= af-token-list-next af-token-list-first)
            (error "next-af-token: more than %d tokens on one line?"
                   (1+ af-token-list-mask))
          )
        )
    )

  ;;--Remember this location so we can continue lexing in the future.
  ;;  Preset the flags for this next time to be not-BOL.

  (aset af-token-flags  af-token-list-next 0)
  (aset af-token-points af-token-list-next (point))
  
  );; next-af-token-lex

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-defsubst next-af-token (reference)

  ;;--After calling clear-af-token-list, call this with 0 to get the first
  ;;  token for the empty token list.  Returns a reference to the token which
  ;;  follows the argument token reference in the token list.  We
  ;;  automatically invoke the token lexing when necessary.
  ;;
  ;;--Advance the reference number.

  (setq reference (logand (1+ reference) af-token-list-mask))

   ;;--If we just hit the "next" slot then we need to do some lexing in the
   ;;  forward direction.

  (if (= reference af-token-list-next)
      (next-af-token-lex)
    )
  (af-debug (format "%S, reference: %d"
                    (aref (aref af-token-symbols reference) 0)
                    reference)
            (aref af-token-points reference))
  reference

  );; next-af-token

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun initial-indent-af-token ()

  ;;--Only call this routine right after clear-af-token-list.  It is just
  ;;  (next-af-token 0) in that case, but it never scans past a comment or
  ;;  end of line.
  ;;
  ;;--We need to do some lexing in the forward direction.

  (let ((starting-point (point))
        lex-result-symbol
        lex-result-point
        (lex-result-flag (aref af-token-flags af-token-list-next)))

    ;;--Goto wherever we stopped lexing in the past.  The af-token-flags
    ;;  value for this spot (basically the BOL flag) is already set.
    ;;  Continue to lex forward until we find a non-eol token or until we
    ;;  hit end-of-buffer.

    (goto-char (aref af-token-points af-token-list-next))
    (aset af-token-symbols af-token-list-next (af-lex-one))
    (aset af-token-points  af-token-list-next lex-result-point)
    (aset af-token-flags   af-token-list-next lex-result-flag)
    (af-debug-check
     (if (not (and (vectorp lex-result-symbol)
                   (numberp lex-result-point)
                   (numberp lex-result-flag)))
         (error "Something went wrong in the parse results.")
       )
     )

    ;;--Advance the next pointer.  If this conflicts with the first pointer
    ;;  then advance that point until we find a BOL token.

    (setq af-token-list-next (logand (1+ af-token-list-next)
                                     af-token-list-mask))
    (if (= af-token-list-next af-token-list-first)
        (while
            (setq af-token-list-first (logand (1+ af-token-list-first)
                                              af-token-list-mask)
                  lex-result-flag       ; 2*setq is faster than progn+setq.
                                        ; This assigment is an optimization.
                  (af-zerobit (aref af-token-flags af-token-list-first)
                              af-token-flag-bol))
          (if (= af-token-list-next af-token-list-first)
              (error
               "initial-indent-af-token: more than %d tokens on one line?"
               (1+ af-token-list-mask))
            )
          )
      )

    ;;--Remember this location so we can continue lexing in the future.
    ;;  Preset the flags for this next time to be not-BOL.

    (if (or (eq lex-result-symbol af-syntax-end-of-line)
            (eq lex-result-symbol af-syntax-comment))
        (aset af-token-flags  af-token-list-next af-token-flag-bol)
      (aset af-token-flags  af-token-list-next 0)
      )
    (aset af-token-points af-token-list-next (point))
    (goto-char starting-point)
    )
  (af-debug (format "%S, reference: %d"
                    (aref (aref af-token-symbols (1- af-token-list-next)) 0)
                    (1- af-token-list-next))
            (aref af-token-points (1- af-token-list-next)))
  (1- af-token-list-next)
  );; initial-indent-af-token

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun af-token-previous (lex-result-symbol
                          lex-result-point
                          lex-result-flag)

  ;;--Utility subprogram for the previous-af-token routine.  We recursively
  ;;  lex forward to eol.  Then we unrecurse and use successive slots
  ;;  in the token list.  Doing it recursively eliminates having to shuffle
  ;;  the arrays of the list.

  ;;--Lex forward one token.  If it isn't an eol token, recurse and get
  ;;  more tokens.  Then record what we have using the -first index.

  (setq lex-result-flag 0)
  (if (or (eq (af-lex-one)      af-syntax-end-of-line)
          (eq lex-result-symbol af-syntax-comment))

      nil                               ; Simply exit this function.

    (af-token-previous nil nil 0)       ; Found one, recurse for rest.
        
    (aset af-token-symbols af-token-list-first lex-result-symbol)
    (aset af-token-points  af-token-list-first lex-result-point)
    (aset af-token-flags   af-token-list-first lex-result-flag)
    (af-debug-check
     (if (not (and (vectorp lex-result-symbol)
                   (numberp lex-result-point)
                   (numberp lex-result-flag)))
         (error "Something went wrong in the parse results.")
       )
     )

    ;;--Retard the first pointer.  If this conflicts with the next
    ;;  pointer then retard that pointer by one.

    (setq af-token-list-first (logand (1- af-token-list-first)
                                      af-token-list-mask))
    (if (= af-token-list-next af-token-list-first)
        (setq af-token-list-next (logand (1- af-token-list-next)
                                         af-token-list-mask))
      )
    )

  );; af-token-previous

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun previous-af-token-lex ( &optional      ; Local variables
                               bol
                               lex-result-symbol
                               lex-result-point
                               lex-result-flag)

  ;;--Subroutine for previous-af-token.  Called when we need to perform some
  ;;  lexing.
  ;;
  ;;--Goto wherever we stopped lexing in the past.  It will be a bol
  ;;  point.  We want to back up to the previous bol point and start
  ;;  lexing from there.  We keep backing up until we find a
  ;;  non-blank/comment line or until we hit beginning-of-buffer.
  ;;  Then we lex the entire line.

  (goto-char (aref af-token-points af-token-list-first))
  (setq lex-result-flag af-token-flag-bol);
  (while
      (if (bobp)

          ;;--If there is no more text up there, record an eol and
          ;;  stop.

          (progn
            (setq lex-result-symbol af-syntax-end-of-line
                  lex-result-point  (point)
                  bol               (point))
            nil                   ; Terminate loop.
            )

        ;;--Back up to the beginning of the next previous line and
        ;;  try to lex it.  If we get something, recurse and get the
        ;;  rest of it as well.

        (backward-char 1)       ; Skip preceding eol character.
        (af-beginning-of-line)  ; Go to start of new current line.
        (setq bol (point))      ; Remember this point for next iter.
        (if (or (eq (af-lex-one)      af-syntax-end-of-line)
                (eq lex-result-symbol af-syntax-comment))
            t                   ; Blank/comment line, loop again.
          (af-token-previous nil nil 0) ; Found one, recurse for rest.
          nil                   ; Terminate loop.
          )
        ) 
    (goto-char bol)
    )
  (aset af-token-symbols af-token-list-first lex-result-symbol)
  (aset af-token-points  af-token-list-first lex-result-point)
  (aset af-token-flags   af-token-list-first lex-result-flag)
  (af-debug-check
   (if (not (and (vectorp lex-result-symbol)
                 (numberp lex-result-point)
                 (numberp lex-result-flag)))
       (error "Something went wrong in the parse results.")
     )
   )

  ;;--Retard the first pointer.  If this conflicts with the next
  ;;  pointer then retard that pointer by one.

  (setq af-token-list-first (logand (1- af-token-list-first)
                                    af-token-list-mask))
  (if (= af-token-list-next af-token-list-first)
      (setq af-token-list-next (logand (1- af-token-list-next)
                                       af-token-list-mask))
    )
  (aset af-token-points af-token-list-first bol)

  );; previous-af-token-lex

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-defsubst previous-af-token (reference)

  ;;--After calling clear-af-token-list, call this with 0 to get the first
  ;;  token would precede the location of the empty token list.  Returns a
  ;;  reference to the token which follows the argument token reference in
  ;;  the token list.  We automatically invoke the token lexing when necessary.
  ;;
  ;;--Retard the reference number.

  (setq reference (logand (1- reference) af-token-list-mask))

  ;;--If we just hit the "first" slot then we need to do some lexing in the
  ;;  backward direction.

  (if (= reference af-token-list-first)
      (previous-af-token-lex)
    )
  (af-debug (format "%S, reference: %d"
                    (aref (aref af-token-symbols reference) 0)
                    reference)
            (aref af-token-points reference))
  reference

  );; previous-af-token

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun af-token-shrink-next (reference current &optional next)

  ;;--Called to combine the reference token with the one that follows it.
  ;;  This results in one token where there were two.  We shuffle tokens
  ;;  as appropriate.

  (aset af-token-symbols reference current)

  (setq current (logand (1+ reference) af-token-list-mask)
        next    (logand (1+ current)   af-token-list-mask))
  (aset af-token-symbols current (aref af-token-symbols next))
  (aset af-token-points  current (aref af-token-points  next))
  (aset af-token-flags   current (aref af-token-flags   next))
  (af-debug-check
   (if (and (/= next af-token-list-next)
            (not (and (vectorp (aref af-token-symbols current))
                      (numberp (aref af-token-points  current))
                      (numberp (aref af-token-flags   current)))))
       (error "Something went wrong in the parse results.")
     )
   )

  (while (/= next af-token-list-next)
    (setq current next
          next    (logand (1+ next) af-token-list-mask))
    (aset af-token-symbols current (aref af-token-symbols next))
    (aset af-token-points  current (aref af-token-points  next))
    (aset af-token-flags   current (aref af-token-flags   next))
    (af-debug-check
     (if (and (/= next af-token-list-next)
              (not (and (vectorp (aref af-token-symbols current))
                        (numberp (aref af-token-points  current))
                        (numberp (aref af-token-flags   current)))))
         (error "Something went wrong in the parse results.")
       )
     )
    )
  (setq af-token-list-next current)
  (af-debug (format "%S, reference: %d"
                    (aref (aref af-token-symbols reference) 0)
                    reference)
            (aref af-token-points reference))
  reference
  );; af-token-shrink-next

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun af-token-shrink-previous (reference current &optional next)

  ;;--Called to combine the reference token with the one that follows it.
  ;;  This results in one token where there were two.  We shuffle tokens
  ;;  as appropriate.

  (aset af-token-symbols reference current)

  (setq current (logand (1- reference) af-token-list-mask)
        next    (logand (1- current)   af-token-list-mask))

  (aset af-token-points  reference (aref af-token-points  current))
  (aset af-token-flags   reference (aref af-token-flags   current))

  (aset af-token-symbols current (aref af-token-symbols next))
  (aset af-token-points  current (aref af-token-points  next))
  (aset af-token-flags   current (aref af-token-flags   next))
  (af-debug-check
   (if (and (/= next af-token-list-first)
            (not (and (vectorp (aref af-token-symbols current))
                      (numberp (aref af-token-points  current))
                      (numberp (aref af-token-flags   current)))))
       (error "Something went wrong in the parse results.")
     )
   )

  (while (/= next af-token-list-first)
    (setq current next
          next    (logand (1- current) af-token-list-mask))
    (aset af-token-symbols current (aref af-token-symbols next))
    (aset af-token-points  current (aref af-token-points  next))
    (aset af-token-flags   current (aref af-token-flags   next))
    (af-debug-check
     (if (and (/= next af-token-list-first)
              (not (and (vectorp (aref af-token-symbols current))
                        (numberp (aref af-token-points  current))
                        (numberp (aref af-token-flags   current)))))
         (error "Something went wrong in the parse results.")
       )
     )
    )
  (setq af-token-list-first current)
  (af-debug (format "%S, reference: %d"
                    (aref (aref af-token-symbols reference) 0)
                    reference)
            (aref af-token-points reference))
  reference
  );; af-token-shrink-previous

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-defsubst af-token-change-symbol (reference symbol)

  ;;--Called to change the symbol associated with a reference.

  (aset af-token-symbols reference symbol)
  reference

  );; af-token-change-symbol


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; AF-TOKEN-LIST Query Functions - Functions for asking who/what about
;;;; elements in the lexical stream.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-defsubst af-token-reference-check (reference)
  (af-debug-check
   (cond ((not (numberp reference))
          (error "af-token-reference-check: Reference is not a number [%s]"
                 (prin1-to-string reference)))

         ((if (> af-token-list-next af-token-list-first)
              ;;--List is first..next.
              (or (<= reference af-token-list-first)
                  (>= reference af-token-list-next))
            ;;--List is first..mask and 0..next.
            (or (and (>= reference af-token-list-next)
                     (<= reference af-token-list-first))
                (< reference 0)
                (> reference af-token-list-mask)))
          (error
           "af-token-reference-check: Reference is out of bounds [%d]"
           reference))

         ((not (and (vectorp (aref af-token-symbols reference))
                    (numberp (aref af-token-points  reference))
                    (numberp (aref af-token-flags   reference))))
          (error "Bad data in the recorded parse results."))
         );; cond
   )
  );; af-token-reference-check

(af-defsubst af-token-symbol (reference)
  ;;--Returns the token symbol.
  (af-token-reference-check reference)
  (aref af-token-symbols reference)
  );; af-token-symbol

(af-defsubst af-token-point (reference)
  ;;--Returns the (point) where the token text began.
  (af-token-reference-check reference)
  (aref af-token-points reference)
  );; af-token-point

(af-defsubst af-token-flags (reference)
  ;;--Returns the lexer flags for the token.
  (af-token-reference-check reference)
  (aref af-token-flags reference)
  );; af-token-flags

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; Our per-token data structure has this shape:
;;;;
;;;;[self data indent check-previous check-next name match-previous match-next]
;;;;  0    1    2      3              4          5    6              7
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defconst af-token-data-offset           1)
(defconst af-token-indent-offset         2)
(defconst af-token-check-previous-offset 3)
(defconst af-token-check-next-offset     4)
(defconst af-token-name-offset           5)
(defconst af-token-match-previous-offset 6)
(defconst af-token-match-next-offset     7)

(af-defsubst af-token-data (reference)
  ;;--Returns the token symbol's data.
  (af-token-reference-check reference)
  (aref (aref af-token-symbols reference) af-token-data-offset)
  );; af-token-data

(af-defsubst af-token-check-previous-function (reference)
  ;;--Returns the token symbol's check-previous function.
  (af-token-reference-check reference)
  (aref (aref af-token-symbols reference) af-token-check-previous-offset)
  );; af-token-check-previous-function

(af-defsubst af-token-check-next-function (reference)
  ;;--Returns the token symbol's check-next function.
  (af-token-reference-check reference)
  (aref (aref af-token-symbols reference) af-token-check-next-offset)
  );; af-token-check-next-function

(af-defsubst af-token-special-indentation (reference)
  ;;--Returns the token symbol's special indentation value.
  (af-token-reference-check reference)
  (symbol-value
   (aref (aref af-token-symbols reference) af-token-indent-offset))
  );; af-token-special-indentation

(af-defsubst af-token-string-name (reference)
  ;;--Returns the token symbol's string name.
  (af-token-reference-check reference)
  (let ((str (aref (aref af-token-symbols reference) af-token-name-offset)))
    (if (stringp str)
        str
      (prin1-to-string (aref af-token-symbols reference))
      )
    )
  );; af-token-string-name

(af-defsubst af-token-match-previous (reference)
  ;;--Returns the token symbol's matching previous-token list.
  (af-token-reference-check reference)
  (aref (aref af-token-symbols reference) af-token-match-previous-offset)
  );; af-token-match-previous

(af-defsubst af-token-match-next (reference)
  ;;--Returns the token symbol's matching next-token list.
  (af-token-reference-check reference)
  (aref (aref af-token-symbols reference) af-token-match-next-offset)
  );; af-token-match-next

(defun af-token-match-check (match-list reference &optional tmp)
  ;;--Checks the reference against the current match-list.
  ;;
  ;;  Returns a (consp) if we are ok and we still have more to match.
  ;;
  ;;  Returns 0 if we have just passed the last known token in a sequence.
  ;;  This means that we have just successfully passed the end of a group.
  ;;
  ;;  Returns 1 if we have fallen off the end of an optional sequence; meaning
  ;;  we did not accept the token, but there is no syntax error.
  ;;
  ;;  Returns 2 if we have fallen off the end of a mandatory sequence; meaning
  ;;  that we did not accept the token and there is a syntax error.

  (setq reference (af-token-symbol reference))
  (if (eq reference (car (setq tmp (car match-list))))
      (if (eq t (car (car (setq tmp (cdr tmp)))))
          0                                   ; Just passed the last known.
        tmp)                                  ; We are ok, and continue.
    (catch 'return
      (while (setq match-list (cdr match-list))
        (if (eq reference (car (setq tmp (car match-list))))
            (throw 'return
                   (if (eq t (car (car (setq tmp (cdr tmp)))))
                       0                      ; Just passed the last known.
                     tmp))                    ; We are ok, and continue.
          )
        );; while
      (if (eq t (car tmp))
          1                                   ; It was optional, we are ok.
        2)                                    ; It wasn't optional, give error.
      );; catch
    )

  );; af-token-match-check

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-defsubst af-token-check-previous (reference &optional func)

  ;;--Checks the indicated symbol for being a \"complex\" lexical token (more
  ;;  than one part, e.g. OR ELSE).  After the call, no other reference values
  ;;  remain valid.  The af-token-list has been arbitrarily rearranged around 
  ;;  the argument reference.
  ;;
  ;;  The return value is the one and only reference value that is still
  ;;  guaranteed to be valid for the caller (or anyone else) to use.
  ;;
  ;;--We don't bother to check things twice.

  (af-token-reference-check reference)

  ;;--Call the complexity checking routine for this token.

  (if (setq func (af-token-check-previous-function reference))
      (funcall func reference)
    reference
    )

  );; af-token-check-previous

(af-defsubst af-token-check-next (reference &optional func)

  ;;--Checks the indicated symbol for being a \"complex\" lexical token (more
  ;;  than one part, e.g. OR ELSE).  After the call, no other reference values
  ;;  remain valid.  The af-token-list has been arbitrarily rearranged around 
  ;;  the argument reference.
  ;;
  ;;  The return value is the one and only reference value that is still
  ;;  guaranteed to be valid for the caller (or anyone else) to use.
  ;;
  ;;--We don't bother to check things twice.

  (af-token-reference-check reference)

  ;;--Call the complexity checking routine for this token.

  (if (setq func (af-token-check-next-function reference))
      (setq reference (funcall func reference))
    reference
    )

  );; af-token-check-next

(af-defsubst af-token-check-both (reference)

  ;;--Checks the indicated symbol for being a \"complex\" lexical token (more
  ;;  than one part, e.g. OR ELSE).  After the call, no other reference values
  ;;  remain valid.  The af-token-list has been arbitrarily rearranged around 
  ;;  the argument reference.
  ;;
  ;;  The return value is the one and only reference value that is still
  ;;  guaranteed to be valid for the caller (or anyone else) to use.
  ;;
  ;;--We don't bother to check things twice.

  (af-token-check-previous (af-token-check-next reference))

  );; af-token-check-both


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; AF-FLAG-<name> - The various symbol flags we have defined.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;--Any token that can only appear as the first token of a statement.

(defconst af-flag-stmt-start           (lsh 1 0))

;;--Any token that can follow an END.  END IF, END abc, END "abc", END CASE,
;;  etc.  Does not include ";".  This is used to test for compound end tokens.

(defconst af-flag-end-follower         (lsh 1 1))

;;--Any token that can be used to short-cut the search for the start of an
;;  element group during indentation.

(defconst af-flag-major-end            (lsh 1 2))

;;--Any token that begins a group (DECLARE), continues a group (ELSE), ends
;;  a group (END IF), or may begin *or* continue a group depending on context
;;  (BEGIN).

(defconst af-flag-group-begin          (lsh 1 3))
(defconst af-flag-group-continue       (lsh 1 4))
(defconst af-flag-group-end            (lsh 1 5))
(defconst af-flag-group-begin-continue (lsh 1 6))
(defconst af-flag-element-separator    (lsh 1 7))

;;--Any begin-continue/continue token where all internal elements are indented
;;  relative to this token regardless of where it appears on the line.

(defconst af-flag-force-indent         (lsh 1 8))

;;--These tokens are markers that indicate we are in a "tight" expression 
;;  region for the purposes of indentation.

(defconst af-flag-tight-expression     (lsh 1 9))

;;--These tokens are markers that indicate that a preceding IS is just an
;;  IS and is not a group continue token.

(defconst af-flag-simple-is-follower   (lsh 1 10))

;;--These tokens are markers that indicate that a following IS is just an
;;  IS and is not a group continue token.  eg. SUBTYPE

(defconst af-flag-simple-is-start      (lsh 1 11))

;;--These tokens are markers that indicate that a following IS is to be
;;  treated as a group-continue token.  eg. PACKAGE

(defconst af-flag-is-group-start       (lsh 1 12))

;;--These tokens are things that matter when we are trying to figure out
;;  what to do with a FUNC/PROC/PACKAGE keyword when going backwards.
;;  Basically, these are things that tell us whether they are appearing
;;  as begins (PACKAGE x IS) or as ends (GENERIC x FUNCTION...;) or as
;;  ordinary tokens (PROCEDURE x;).

(defconst af-flag-is-going-backwards   (lsh 1 13))

;;--These tokens are things that we treat as pseudo begins when we are
;;  indenting on a line immediately after them.

(defconst af-flag-after-pseudo-begin   (lsh 1 14))

;;--These tokens are things that cannot appear in a GENERIC formal declarative
;;  region.  They are used to short-circuit an af-calc-backup-inside-group.

(defconst af-flag-not-generic-formal   (lsh 1 15))

;;--These tokens are things that cannot appear in a DECLARE declarative region.
;;  They are used to short-circuit an af-calc-backup-inside-group.

(defconst af-flag-not-declare-begin    (lsh 1 16))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;--Any begin, continue, or separator such as DECLARE, THEN, or ','.  Used to
;;  "look back" and determine that the current token must be the start of a
;;  grouped element (an element that occurs between begin-end markers and which
;;  is indented relative to those markers.

(defconst af-flag-any-sep              (logior af-flag-group-begin
                                               af-flag-group-begin-continue
                                               af-flag-group-continue
                                               af-flag-element-separator))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; AF-SYNTAX-<token> declarations.  Create data structures to describe them.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defconst af-always-zero 0 "Must always be zero.") ;; Please don't change it.

(let ((fixup-list     nil)                    ; Local Initialized Variables
      (preload-states nil)
      (cons-pool      nil)

      af-token)                               ; Local Functions
  
  (fset
   'af-token
   (lambda (token-name &rest args)

     ;;--The first 2..N arguments are numbers.  The rest of the arguments are
     ;;  optional.  The possibilities are:
     ;;
     ;;  ('check-previous <function>) - Function used to check a simple token
     ;;                        and its previous neighbor to see if they are
     ;;                        really a single complex token.  The
     ;;                        default is nil.
     ;;  ('check-next <function>) - Function used to check a simple token and
     ;;                        its next neighbor to see if  they are really a
     ;;                        single complex token.  The default is nil.
     ;;  ('indent <symbol>)  - A symbol for a variable containing a special
     ;;                        (usually negative) numeric indentation amount to
     ;;                        be used when indenting this token on a newline.
     ;;  ('name "token name")- A string name used for error messages during
     ;;                        group skip operations.

     (let ((data           0)
           (check-previous nil)
           (check-next     nil)
           (indent         'af-always-zero)
           (name           nil)
           sub-list
           symbol)

       (af-debug-check (message (prin1-to-string token-name)))

       ;;--Eat up all of the numbers appearing at the head of the list.

       (while (numberp (car args))
         (setq data (logior data (car args))
               args (cdr args))
         )

       ;;--Now eat up all of the (<value> <symbol>) pairs that come next.

       (while (and args
                   (symbolp (setq symbol (car args))))
         (setq args (cdr args))
         (cond

          ;;--Record a 'check-previous/'check-next value here.

          ((eq symbol 'check-previous)
           (setq check-previous (car args)))
          ((eq symbol 'check-next)
           (setq check-next (car args)))
               
          ;;--Record an 'indent value here.

          ((eq symbol 'indent)
           (setq indent (car args))
           (if (not (symbolp indent))
               (af-debug-error
                "af-flags: 'indent value not a symbol? [%S]"
                token-name)))

          ;;--Record a 'name value here.

          ((eq symbol 'name)
           (setq name (car args))
           (if (not (stringp name))
               (af-debug-error "af-flags: 'name value not a string? [%S]"
                               token-name)))

          ;;--Anything else is some kind of error.

          (t
           (af-debug-error "af-flags: Symbol [%S] not recognized? [%S]"
                           symbol token-name))
          );; cond

         ;;--Loop for next list element.

         (setq args (cdr args))
         );; while

       ;;--Now create the data structure for this token.  It is a simple
       ;;  array.
       ;;
       ;; [data indent check-previous check-next name {match-prev match-next}]

       (if (af-testbit data (logior af-flag-group-begin
                                    af-flag-group-continue
                                    af-flag-group-begin-continue
                                    af-flag-group-end))

           ;;--We have a group token.  It must have a name.  Create a vector.
           ;;  Add every symbol of this sort to the fixup-list.

           (progn
             (if (and (af-zerobit data (logior af-flag-group-begin-continue
                                               af-flag-group-continue))
                      (af-testbit data af-flag-force-indent))
                 (af-debug-error
                  "group token has unneeded -force-indent? [%S]"
                  token-name))
             (if (not name)
                 (af-debug-error "group token has no name? [%S]"
                                 token-name))
             (setq tmp
                   (eval (` (defconst (, token-name)
                              (vector (quote (, token-name))
                                      data indent check-previous check-next
                                      name nil nil))))
                   fixup-list (cons tmp fixup-list))
             (car fixup-list)
             )

         ;;--We have a non-group token.  It must not have a name.

         (progn
           (if (and name
                    (af-zerobit data af-flag-element-separator))
               (af-debug-error "non-group token has a name? [%S]"
                               token-name))
           (if name
               (eval (` (defconst (, token-name)
                          (vector (quote (, token-name))
                                  data indent check-previous check-next
                                  name))))
             (eval (` (defconst (, token-name)
                        (vector (quote (, token-name))
                                data indent check-previous check-next))))
             )
           )
         );; if
       );; let
     );; af-token
   );; fset 'af-token


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; AF-SYNTAX-<token name> - These are the control flags associated with
;;;; the various kinds of tokens.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;--af-syntax-end-of-line will occur at (bobp) and it may occur at (eobp),
;;  although af-syntax-comment may occur at (eobp) instead.
;;
;;  It will only appear in the syntax list at those two points.  Because we
;;  spend virtually all of our time scanning backwards, we need a "cheap" way
;;  to test for (bobp) that will "automatically" happen in all of our 
;;  scanning loops.  We achieve this by making end-of-line be a group-begin
;;  and by setting force-indent on it.  It is a begin that will never match
;;  any end (so we'll get proper errors when skipping groups which are not
;;  syntactically correct) and it will always stop the indentation search.

(af-token 'af-syntax-end-of-line              ;;; end-of-line or buffer
          af-flag-group-begin
          'name "Begin/End of Buffer")

(af-token 'af-syntax-comment)                 ;;; comment (to end-of-line)

(af-token 'af-syntax-character                ;;; 'x'
          af-flag-tight-expression)

(af-token 'af-syntax-identifier               ;;; abc
          af-flag-end-follower
          af-flag-tight-expression
          'check-previous
          (lambda (reference)
            (if (eq (af-token-symbol (previous-af-token reference))
                    af-syntax-end)
                ;;--END ABC;
                (af-token-shrink-previous reference af-syntax-end-name)
              )
            reference
            );; lambda
          'check-next
          (lambda (reference
                   &optional            ; Local variables
                   id-position
                   next-token
                   symbol
                   tmp)
            (setq next-token (next-af-token reference)
                  symbol     (af-token-symbol next-token))
            (cond

             ;;--NAME:LOOP, NAME:BLOCK

             ((eq symbol af-syntax-colon)
              (setq next-token (next-af-token next-token)
                    symbol     (af-token-symbol next-token))
              (cond ((or (eq symbol af-syntax-while)
                         (eq symbol af-syntax-for)
                         (eq symbol af-syntax-loop))
                     (af-token-shrink-next reference af-syntax-loop-name))

                    ((or (eq symbol af-syntax-declare)
                         (eq symbol af-syntax-begin))
                     (af-token-shrink-next reference af-syntax-block-name))
                    );; cond
              )

             ;;--A ID with a previous ' cannot be a name-at.

             ((eq (af-token-symbol (previous-af-token reference))
                  af-syntax-apostrophe)
              )

             ;;--ID at x range x..x;

             ((eq symbol af-syntax-at)
              (af-token-change-symbol reference af-syntax-name-at))
                 
             ;;--ID['id]+ at x range x..x;

             ((eq symbol af-syntax-apostrophe)
              (setq id-position (af-token-point reference))
              (while                    ; Search past all ' ID pairs
                  (and
                   (setq next-token (next-af-token next-token)
                         symbol     (af-token-symbol next-token))
                   (eq symbol af-syntax-identifier)
                   (setq next-token (next-af-token next-token)
                         symbol     (af-token-symbol next-token))
                   (eq symbol af-syntax-apostrophe))
                )
              (while                    ; Now search back to where we started.
                  (setq next-token (previous-af-token next-token)
                        tmp (/= id-position (af-token-point next-token)))
                )
              (if (eq symbol af-syntax-at) ; Maybe change the symbol.
                  (af-token-change-symbol reference af-syntax-name-at)
                )
              )
             );; cond
            reference
            );; lambda
          )
(af-token 'af-syntax-number)                  ;;; 123 16#123# 16:123:
(af-token 'af-syntax-string                   ;;; "x" %x%
          af-flag-end-follower
          af-flag-tight-expression
          'check-previous
          (lambda (reference)
            (if (eq (af-token-symbol (previous-af-token reference))
                    af-syntax-end)
                ;;--Change the STRING to an END FUNC.
                (af-token-shrink-previous reference af-syntax-end-func)
              )
            reference
            );; lambda
          )

(af-token 'af-syntax-left-curly               ;;; {
          af-flag-group-begin
          'name "{")
(af-token 'af-syntax-right-curly              ;;; }
          af-flag-group-end
          'name "}")

(af-token 'af-syntax-left-label               ;;; <<
          af-flag-group-begin
          af-flag-stmt-start
          af-flag-not-declare-begin
          'indent 'af-indent-left-label
          'name "<<")
(af-token 'af-syntax-right-label              ;;; >>
          af-flag-group-end
          af-flag-not-declare-begin
          'name ">>")

(af-token 'af-syntax-left-paren               ;;; (
          af-flag-tight-expression
          af-flag-group-begin
          af-flag-is-going-backwards
          'name "(")
(af-token 'af-syntax-right-paren              ;;; )
          af-flag-tight-expression
          af-flag-group-end
          'name ")")

(af-token 'af-syntax-left-square              ;;; [
          af-flag-group-begin
          'name "[")
(af-token 'af-syntax-right-square             ;;; ]
          af-flag-group-end
          'name "]")


(af-token 'af-syntax-ampersand)               ;;; &
(af-token 'af-syntax-apostrophe               ;;; '
          af-flag-tight-expression)
(af-token 'af-syntax-arrow                    ;;; =>
          af-flag-element-separator
          af-flag-is-going-backwards
          'name "=>")
(af-token 'af-syntax-assignment               ;;; :=
          af-flag-is-going-backwards)
(af-token 'af-syntax-box)                     ;;; <>
(af-token 'af-syntax-colon                    ;;; :
          'check-previous
          (lambda (reference &optional symbol)
            (if (and (eq (af-token-symbol (previous-af-token reference))
                         af-syntax-identifier)
                     (setq symbol (af-token-symbol (next-af-token reference))))
                (cond

                 ;;--NAME:LOOP

                 ((or (eq symbol af-syntax-while)
                      (eq symbol af-syntax-for)
                      (eq symbol af-syntax-loop))
                  (af-token-shrink-previous reference af-syntax-loop-name)
                  )

                 ;;--NAME:BLOCK

                 ((or (eq symbol af-syntax-declare)
                      (eq symbol af-syntax-begin))
                  (af-token-shrink-previous reference af-syntax-block-name)
                  )
                 );; cond
              )
            reference
            );; lambda
          )
(af-token 'af-syntax-comma                    ;;; ,
          af-flag-element-separator
          af-flag-is-going-backwards
          'name ",")
(af-token 'af-syntax-divide)                  ;;; /
(af-token 'af-syntax-dot                      ;;; .
          af-flag-tight-expression)
(af-token 'af-syntax-double-dot)              ;;; ..
(af-token 'af-syntax-double-star)             ;;; **
(af-token 'af-syntax-equal)                   ;;; =
(af-token 'af-syntax-greater-than)            ;;; >
(af-token 'af-syntax-greater-than-or-equal)   ;;; >=
(af-token 'af-syntax-less-than)               ;;; <
(af-token 'af-syntax-less-than-or-equal)      ;;; <= 
(af-token 'af-syntax-minus)                   ;;; -
(af-token 'af-syntax-multiply)                ;;; *
(af-token 'af-syntax-not-equal)               ;;; /=
(af-token 'af-syntax-plus)                    ;;; +
(af-token 'af-syntax-semicolon                ;;; ;
          af-flag-element-separator
          af-flag-is-going-backwards
          'name ";")
(af-token 'af-syntax-vertical-bar             ;;; | !
          af-flag-element-separator
          'name "|")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; AF-KEYWORDS-? - What keywords start with what letter and what flags go
;;;; with each keyword.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-token 'af-syntax-loop-name                ;;; NAME:LOOP
          af-flag-not-declare-begin
          'indent 'af-indent-loop-name)
(af-token 'af-syntax-block-name               ;;; NAME:DECLARE
          'indent 'af-indent-block-name)
(af-token 'af-syntax-name-at                  ;;; NAME AT x rng a..b;
          af-flag-tight-expression)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-token 'af-syntax-abort                    ;;; "ABORT"
          af-flag-stmt-start
          af-flag-not-declare-begin)
(af-token 'af-syntax-abs)                     ;;; "ABS"
(af-token 'af-syntax-abstract                 ;;; "ABSTRACT"
          af-flag-simple-is-follower)
(af-token 'af-syntax-accept                   ;;; "ACCEPT"
          af-flag-group-begin
          af-flag-stmt-start
          af-flag-not-declare-begin
          'name "ACCEPT")
(af-token 'af-syntax-access)                  ;;; "ACCESS"
(af-token 'af-syntax-aliased)                 ;;; "ALIASED"
(af-token 'af-syntax-all                      ;;; "ALL"
          af-flag-tight-expression)
(af-token 'af-syntax-and                      ;;; "AND"
          'check-next
          (lambda (reference)
            (if (eq (af-token-symbol (next-af-token reference))
                    af-syntax-then)
                ;;--Change the AND to an AND THEN.
                (af-token-shrink-next reference af-syntax-and-then)
              )
            reference
            );; lambda
          )
(af-token 'af-syntax-and-then                 ;;; "AND THEN"
          af-flag-element-separator
          'name "AND THEN")
(af-token 'af-syntax-array)                   ;;; "ARRAY"
(af-token 'af-syntax-at)                      ;;; "AT"

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-token 'af-syntax-begin                    ;;; "BEGIN"
          af-flag-group-begin-continue
          af-flag-stmt-start
          af-flag-force-indent
          af-flag-not-declare-begin
          'name "BEGIN")
(af-token 'af-syntax-body                     ;;; "BODY"
          af-flag-not-declare-begin
          'check-previous
          (lambda (reference &optional symbol)
            (setq symbol (af-token-symbol (previous-af-token reference)))
            (cond
             ;;--PACKAGE BODY
             ((eq symbol af-syntax-package)
              (af-token-shrink-previous reference af-syntax-package-body))
             ;;--TASK BODY
             ((eq symbol af-syntax-task)
              (af-token-shrink-previous reference af-syntax-task-body))
             ;;--PROTECTED BODY
             ((eq symbol af-syntax-protected)
              (af-token-shrink-previous reference af-syntax-protected-body))
             )
            reference
            );; lambda
          )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-token 'af-syntax-case                     ;;; "CASE"
          af-flag-is-group-start
          af-flag-group-begin
          af-flag-stmt-start
          af-flag-end-follower
          af-flag-not-declare-begin
          'name "CASE"
          'check-previous
          (lambda (reference)
            (if (eq (af-token-symbol (previous-af-token reference))
                    af-syntax-end)
                ;;--Change the CASE to an END CASE.
                (af-token-shrink-previous reference af-syntax-end-case)
              )
            reference
            );; lambda
          )
(af-token 'af-syntax-constant)                ;;; "CONSTANT"

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-token 'af-syntax-declare                  ;;; "DECLARE"
          af-flag-group-begin
          af-flag-stmt-start
          'name "DECLARE")
(af-token 'af-syntax-delay                    ;;; "DELAY"
          af-flag-stmt-start
          af-flag-not-declare-begin)
(af-token 'af-syntax-delta)                   ;;; "DELTA"
(af-token 'af-syntax-digits)                  ;;; "DIGITS"
(af-token 'af-syntax-do                       ;;; "DO"
          af-flag-group-continue
          af-flag-not-declare-begin
          'name "DO")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-token 'af-syntax-else                     ;;; "ELSE"
          af-flag-group-continue
          af-flag-force-indent
          af-flag-not-declare-begin
          'name "ELSE"
          'check-previous
          (lambda (reference)
            (if (eq (af-token-symbol (previous-af-token reference))
                    af-syntax-or)
                ;;--Change the ELSE to an OR ELSE.
                (af-token-shrink-previous reference af-syntax-or-else)
              )
            reference
            );; lambda
          )
(af-token 'af-syntax-elsif                    ;;; "ELSIF"
          af-flag-force-indent
          af-flag-group-continue
          af-flag-not-declare-begin
          'name "ELSIF")
(af-token 'af-syntax-end                      ;;; "END"
          af-flag-major-end
          af-flag-group-end
          af-flag-not-generic-formal
          'name "END"
          'check-next
          (lambda (reference &optional symbol)
            (setq symbol (next-af-token reference))
            (if (af-testbit (af-token-data symbol)
                            af-flag-end-follower)

                ;;--We have a compound END.  Figure out which one we have.

                (cond
                 ((eq (setq symbol (af-token-symbol symbol))
                      af-syntax-if)
                  (af-token-shrink-next reference af-syntax-end-if))
                 ((eq symbol af-syntax-case)
                  (af-token-shrink-next reference af-syntax-end-case))
                 ((eq symbol af-syntax-loop)
                  (af-token-shrink-next reference af-syntax-end-loop))
                 ((eq symbol af-syntax-identifier)
                  (af-token-shrink-next reference af-syntax-end-name))
                 ((eq symbol af-syntax-string)
                  (af-token-shrink-next reference af-syntax-end-func))
                 ((eq symbol af-syntax-record)
                  (af-token-shrink-next reference af-syntax-end-record))
                 ((eq symbol af-syntax-select)
                  (af-token-shrink-next reference af-syntax-end-select))
                 (t
                  (error "END complexity: unexpected token [%s]"
                         (prin1-to-string symbol))
                  )
                 );; cond
              );; if
            reference
            );; lambda
          )
(af-token 'af-syntax-end-if                   ;;; "END IF"
          af-flag-major-end
          af-flag-group-end
          af-flag-not-generic-formal
          af-flag-not-declare-begin
          'name "END IF")
(af-token 'af-syntax-end-case                 ;;; "END CASE"
          af-flag-major-end
          af-flag-group-end
          af-flag-not-generic-formal
          af-flag-not-declare-begin
          'name "END CASE")
(af-token 'af-syntax-end-loop                 ;;; "END LOOP"
          af-flag-major-end
          af-flag-group-end
          af-flag-not-generic-formal
          af-flag-not-declare-begin
          'name "END LOOP")
(af-token 'af-syntax-end-name                 ;;; "END abc"
          af-flag-major-end
          af-flag-group-end
          af-flag-not-generic-formal
          'name "END <id>")
(af-token 'af-syntax-end-func                 ;;; "END "abc""
          af-flag-major-end
          af-flag-group-end
          af-flag-not-generic-formal
          'name "END \"op\"")
(af-token 'af-syntax-end-record               ;;; "END RECORD"
          af-flag-group-end
          af-flag-not-generic-formal
          'name "END RECORD")
(af-token 'af-syntax-end-select               ;;; "END SELECT"
          af-flag-major-end
          af-flag-group-end
          af-flag-not-generic-formal
          af-flag-not-declare-begin
          'name "END SELECT")
(af-token 'af-syntax-entry                    ;;; "ENTRY"
          af-flag-is-group-start)
(af-token 'af-syntax-exception                ;;; "EXCEPTION"
          af-flag-not-generic-formal
          'check-previous
          (lambda (reference)
            (if (not (eq (af-token-symbol (previous-af-token reference))
                         af-syntax-colon))
                ;;--Change the EXCEPTION to an EXCEPTION-section.
                (af-token-change-symbol reference
                                        af-syntax-exception-section)
              )
            reference
            );; lambda
          )
(af-token 'af-syntax-exception-section        ;;; "EXCEPTION" part in block or
          af-flag-group-continue              ;;;  package body
          af-flag-force-indent
          af-flag-not-generic-formal
          af-flag-not-declare-begin
          'name "EXCEPTION (block part)")
(af-token 'af-syntax-exit                     ;;; "EXIT"
          af-flag-stmt-start
          af-flag-not-declare-begin)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-token 'af-syntax-for                      ;;; "FOR"
          af-flag-group-begin
          af-flag-stmt-start
          af-flag-not-declare-begin
          'name "FOR")
(af-token 'af-syntax-function                 ;;; "FUNCTION"
          af-flag-is-group-start
          af-flag-group-begin-continue
          af-flag-force-indent
          af-flag-not-generic-formal
          'name "FUNCTION"
          'check-previous
          (lambda (reference)
            (if (eq (af-token-symbol (previous-af-token reference))
                    af-syntax-with)
                ;;--We have a WITH FUNCTION.
                (af-token-shrink-previous reference af-syntax-with-function)
              )
            reference
            );; lambda
          )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-token 'af-syntax-generic                  ;;; "GENERIC"
          af-flag-group-begin
          'name "GENERIC")
(af-token 'af-syntax-goto                     ;;; "GOTO"
          af-flag-stmt-start
          af-flag-not-declare-begin)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-token 'af-syntax-if                       ;;; "IF"
          af-flag-group-begin
          af-flag-stmt-start
          af-flag-end-follower
          af-flag-not-declare-begin
          'name "IF"
          'check-previous
          (lambda (reference)
            (if (eq (af-token-symbol (previous-af-token reference))
                    af-syntax-end)
                ;;--Change the IF to an END IF.
                (af-token-shrink-previous reference af-syntax-end-if)
              )
            reference
            );; lambda
          )
(af-token 'af-syntax-in)                      ;;; "IN"
(af-token 'af-syntax-is                       ;;; "IS"
          af-flag-is-going-backwards)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-token 'af-syntax-limited                  ;;; "LIMITED"
          af-flag-simple-is-follower)
(af-token 'af-syntax-loop                     ;;; "LOOP"
          af-flag-group-begin-continue
          af-flag-stmt-start
          af-flag-end-follower
          af-flag-not-declare-begin
          'name "LOOP"
          'check-previous
          (lambda (reference)
            (if (eq (af-token-symbol (previous-af-token reference))
                    af-syntax-end)
                ;;--Change the LOOP to an END LOOP.
                (af-token-shrink-previous reference af-syntax-end-loop)
              )
            reference
            );; lambda
          )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-token 'af-syntax-mod)                     ;;; "MOD"

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-token 'af-syntax-new                      ;;; "NEW"
          af-flag-simple-is-follower
          af-flag-after-pseudo-begin)
(af-token 'af-syntax-not)
(af-token 'af-syntax-null
          'check-next
          (lambda (reference)
            (if (eq (af-token-symbol (next-af-token reference))
                    af-syntax-record)
                ;;--Change the NULL to a NULL RECORD.
                (af-token-shrink-next reference af-syntax-null-record)
              )
            reference
            );; lambda
          )
(af-token 'af-syntax-null-record)             ;;; "NULL RECORD"

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-token 'af-syntax-of)                      ;;; "OF"
(af-token 'af-syntax-or                       ;;; "OR"
          'check-next
          (lambda (reference &optional next-token)
            (setq next-token (next-af-token reference))
            (cond

             ;;--Do we have OR ELSE?

             ((eq (af-token-symbol next-token) af-syntax-else)
              (af-token-shrink-next reference af-syntax-or-else))

             ;;--If we see any kind of begin-of-statement
             ;;  keyword forward then we have the SELECT/OR case.
               
             ((af-testbit (af-token-data next-token)
                          af-flag-stmt-start)
              (af-token-change-symbol reference af-syntax-or-select))

             );; cond
            reference 
            );; lambda
          'check-previous
          (lambda (reference)
            (if (af-testbit (af-token-data (previous-af-token reference))
                            af-flag-any-sep)
                (af-token-change-symbol reference af-syntax-or-select)
              )
            reference 
            );; lambda
          )
(af-token 'af-syntax-or-else                  ;;; "OR ELSE"
          af-flag-element-separator
          'name "OR ELSE")
(af-token 'af-syntax-or-select                ;;; "OR" in a SELECT
          af-flag-group-continue
          af-flag-force-indent
          af-flag-stmt-start
          af-flag-not-declare-begin
          'name "OR (in SELECT)")
(af-token 'af-syntax-others)                  ;;; "OTHERS"
(af-token 'af-syntax-out)                     ;;; "OUT"

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-token 'af-syntax-package                  ;;; "PACKAGE"
          af-flag-is-group-start
          af-flag-group-begin-continue
          af-flag-force-indent
          'name "PACKAGE"
          'check-previous
          (lambda (reference)
            (if (eq (af-token-symbol (previous-af-token reference))
                    af-syntax-with)
                ;;--We have a WITH PACKAGE.
                (af-token-shrink-previous reference af-syntax-with-package)
              )
            reference
            );; lambda
          'check-next
          (lambda (reference)
            (if (eq (af-token-symbol (next-af-token reference))
                    af-syntax-body)
                ;;--We have a PACKAGE BODY.
                (af-token-shrink-next reference af-syntax-package-body)
              )
            reference
            );; lambda
          )
(af-token 'af-syntax-package-body             ;;; "PACKAGE BODY"
          af-flag-is-group-start
          af-flag-group-begin
          af-flag-not-generic-formal
          'name "PACKAGE BODY")
(af-token 'af-syntax-pragma                   ;;; "PRAGMA"
          af-flag-stmt-start)
(af-token 'af-syntax-private                  ;;; "PRIVATE"
          af-flag-simple-is-follower
          'check-previous
          (lambda (reference)
            (if (af-testbit (af-token-data (previous-af-token reference))
                            af-flag-any-sep)
                ;;--Change the PRIVATE to a PRIVATE section
                (af-token-change-symbol reference af-syntax-private-section)
              )
            reference
            );; lambda
          )
(af-token 'af-syntax-private-section          ;;; "PRIVATE" part in
          af-flag-group-continue              ;;;  package
          af-flag-force-indent
          af-flag-not-generic-formal
          'name "PRIVATE (PACKAGE part)")
(af-token 'af-syntax-procedure                ;;; "PROCEDURE"
          af-flag-is-group-start
          af-flag-group-begin-continue
          af-flag-force-indent
          af-flag-not-generic-formal
          'name "PROCEDURE"
          'check-previous
          (lambda (reference)
            (if (eq (af-token-symbol (previous-af-token reference))
                    af-syntax-with)
                ;;--We have a WITH PROCEDURE
                (af-token-shrink-previous reference af-syntax-with-procedure)
              )
            reference
            );; lambda
          )
(af-token 'af-syntax-protected                ;;; "PROTECTED"
          af-flag-is-group-start
          af-flag-group-begin
          'check-next
          (lambda (reference &optional symbol)
            (setq symbol (af-token-symbol (next-af-token reference)))
            (cond
             ;;--PROTECTED TYPE
             ((eq symbol af-syntax-type)
              (af-token-shrink-next reference af-syntax-protected-type))
             ;;--PROTECTED-BODY
             ((eq symbol af-syntax-body)
              (af-token-shrink-next reference af-syntax-protected-body))
             )
            reference
            );; lambda
          'name "PROTECTED"
          )
(af-token 'af-syntax-protected-body           ;;; "PROTECTED BODY"
          af-flag-is-group-start
          af-flag-group-begin
          af-flag-not-generic-formal
          'name "PROTECTED BODY"
          )
(af-token 'af-syntax-protected-type           ;;; "PROTECTED TYPE" x IS
          af-flag-is-group-start
          af-flag-group-begin
          af-flag-not-generic-formal
          'name "PROTECTED TYPE"
          )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-token 'af-syntax-raise                    ;;; "RAISE"
          af-flag-stmt-start
          af-flag-not-declare-begin)
(af-token 'af-syntax-range                    ;;; "RANGE"
          af-flag-after-pseudo-begin
          'indent 'af-indent-range)
(af-token 'af-syntax-record                   ;;; "RECORD"
          af-flag-group-begin
          af-flag-end-follower
          'name "RECORD"
          'check-previous
          (lambda (reference &optional symbol)
            (setq symbol (af-token-symbol (previous-af-token reference)))
            (cond

             ;;--Watch for END RECORD

             ((eq symbol af-syntax-end)
              (af-token-shrink-previous reference af-syntax-end-record))

             ;;--Watch for NULL RECORD

             ((eq symbol af-syntax-null)
              (af-token-shrink-previous reference af-syntax-null-record))

             ;;--Is it followed by an AT (AT MOD)?

             ((eq (af-token-symbol (next-af-token reference))
                  af-syntax-at)
              (af-token-change-symbol reference af-syntax-record-at))

             );; cond
            reference
            );; lambda
          )
(af-token 'af-syntax-record-at                ;;; "RECORD AT" (MOD)
          af-flag-group-begin
          'name "RECORD AT")
(af-token 'af-syntax-rem)                     ;;; "REM"
(af-token 'af-syntax-renames)                 ;;; "RENAMES"
(af-token 'af-syntax-requeue                  ;;; "REQUEUE"
          af-flag-stmt-start
          af-flag-not-declare-begin)
(af-token 'af-syntax-return                   ;;; "RETURN"
          af-flag-stmt-start
          af-flag-after-pseudo-begin
;;;;      af-flag-not-declare-begin           ;;; could be FUNC X RETURN Y;
          )
(af-token 'af-syntax-reverse)                 ;;; "REVERSE"

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-token 'af-syntax-select                   ;;; "SELECT"
          af-flag-group-begin
          af-flag-stmt-start
          af-flag-end-follower
          af-flag-not-declare-begin
          'name "SELECT"
          'check-previous
          (lambda (reference)
            (if (eq (af-token-symbol (previous-af-token reference))
                    af-syntax-end)
                ;;--Change the SELECT to an END SELECT.
                (af-token-shrink-previous reference af-syntax-end-select)
              )
            reference
            );; lambda
          )
(af-token 'af-syntax-separate                 ;;; "SEPARATE"
          af-flag-simple-is-follower
          af-flag-not-generic-formal)
(af-token 'af-syntax-subtype                  ;;; "SUBTYPE"
          af-flag-simple-is-start)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-token 'af-syntax-tagged                   ;;; "TAGGED"
          af-flag-simple-is-follower)
(af-token 'af-syntax-task                     ;;; "TASK"
          af-flag-is-group-start
          af-flag-group-begin
          'name "TASK"
          'check-next
          (lambda (reference)
            (if (eq (af-token-symbol (next-af-token reference)) af-syntax-type)
                ;;--We have a TASK TYPE
                (af-token-shrink-next reference af-syntax-task-type)
              )
            reference
            );; lambda
          )
(af-token 'af-syntax-task-body                ;;; "TASK BODY"
          af-flag-is-group-start
          af-flag-group-begin
          af-flag-not-generic-formal
          'name "TASK BODY"
          )
(af-token 'af-syntax-task-type                ;;; "TASK TYPE"
          af-flag-group-begin
          af-flag-is-group-start
          af-flag-not-generic-formal
          'name "TASK TYPE")
(af-token 'af-syntax-terminate                ;;; "TERMINATE"
          af-flag-stmt-start
          af-flag-not-declare-begin)
(af-token 'af-syntax-then                     ;;; "THEN"
          af-flag-group-continue
          af-flag-not-declare-begin
          'name "THEN"
          'check-previous
          (lambda (reference)
            (if (eq (af-token-symbol (previous-af-token reference))
                    af-syntax-and)
                ;;--Change the THEN to an AND THEN.
                (af-token-shrink-previous reference af-syntax-and-then)
              )
            reference
            );; lambda
          )
(af-token 'af-syntax-type                     ;;; "TYPE"
          af-flag-simple-is-start
          'check-previous
          (lambda (reference &optional symbol)
            (setq symbol (af-token-symbol (previous-af-token reference)))
            (cond
             ;;--TASK TYPE
             ((eq symbol af-syntax-task)
              (af-token-shrink-previous reference af-syntax-task-type))
             ;;--PROTECTED TYPE
             ((eq symbol af-syntax-protected)
              (af-token-shrink-previous reference af-syntax-protected-type))
             );; COND
            reference
            );; lambda
          )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-token 'af-syntax-until                    ;;; "UNTIL"
          af-flag-not-declare-begin)
(af-token 'af-syntax-use                      ;;; "USE"
          af-flag-not-generic-formal
          'check-previous
          (lambda (reference)
            (if (af-zerobit (af-token-data (previous-af-token reference))
                            af-flag-any-sep)
                (af-token-change-symbol reference af-syntax-use-of-for)
              )
            reference
            );; lambda
          )

(af-token 'af-syntax-use-of-for               ;;; "FOR xxx USE"
          af-flag-group-end
          'indent 'af-indent-use-of-for
          'name "USE")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-token 'af-syntax-when                     ;;; "WHEN"
          af-flag-group-continue
          af-flag-force-indent
          af-flag-not-declare-begin
          'indent 'af-indent-when
          'name "WHEN"
          'check-previous
          (lambda (reference &optional previous-token symbol)
            (setq previous-token (previous-af-token reference)
                  symbol         (af-token-symbol previous-token))
            (if (or (eq symbol af-syntax-is)
                    (eq symbol af-syntax-exception)
                    (af-testbit (af-token-data (previous-af-token reference))
                                af-flag-any-sep))
                nil
              (af-token-change-symbol reference af-syntax-when-of-exit)
              )
            reference
            );; lambda
          )
(af-token 'af-syntax-when-of-exit)            ;;; "EXIT ... WHEN"
(af-token 'af-syntax-while                    ;;; "WHILE"
          af-flag-group-begin
          af-flag-stmt-start
          af-flag-not-declare-begin
          'name "WHILE")
(af-token 'af-syntax-with                     ;;; "WITH"
          af-flag-not-generic-formal
          af-flag-not-declare-begin
          'check-next
          (lambda (reference &optional symbol)
            (setq symbol (af-token-symbol (next-af-token reference)))
            (cond
             ;;--WITH PACKAGE
             ((eq symbol af-syntax-package)
              (af-token-shrink-next reference af-syntax-with-package))
             ;;--WITH PROCEDURE
             ((eq symbol af-syntax-procedure)
              (af-token-shrink-next reference af-syntax-with-procedure))
             ;;--WITH FUNCTION
             ((eq symbol af-syntax-function)
              (af-token-shrink-next reference af-syntax-with-function))
             );; cond
            reference
            );; lambda

          )
(af-token 'af-syntax-with-function            ;;; "WITH FUNCTION"
          af-flag-simple-is-start
          af-flag-not-declare-begin)
(af-token 'af-syntax-with-package             ;;; "WITH PACKAGE"
          af-flag-simple-is-start
          af-flag-not-declare-begin)
(af-token 'af-syntax-with-procedure           ;;; "WITH PROCEDURE"
          af-flag-simple-is-start
          af-flag-not-declare-begin)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-token 'af-syntax-xor)                     ;;; "XOR"


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; AF-KEYWORDS-? - The keywords we find for each letter of the alphabet.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defconst af-keywords-a (list (cons "ABORT"    af-syntax-abort)
                              (cons "ABS"      af-syntax-abs)
                              (cons "ABSTRACT" af-syntax-abstract)
                              (cons "ACCEPT"   af-syntax-accept)
                              (cons "ACCESS"   af-syntax-access)
                              (cons "ALIASED"  af-syntax-aliased)
                              (cons "ALL"      af-syntax-all)
                              (cons "AND"      af-syntax-and)
                              (cons "ARRAY"    af-syntax-array)
                              (cons "AT"       af-syntax-at))
  )

(defconst af-keywords-b (list (cons "BEGIN" af-syntax-begin)
                              (cons "BODY"  af-syntax-body))
  )

(defconst af-keywords-c (list (cons "CASE"     af-syntax-case)
                              (cons "CONSTANT" af-syntax-constant))
  )

(defconst af-keywords-d (list (cons "DECLARE" af-syntax-declare)
                              (cons "DELAY"   af-syntax-delay)
                              (cons "DELTA"   af-syntax-delta)
                              (cons "DIGITS"  af-syntax-digits)
                              (cons "DO"      af-syntax-do))
  )

(defconst af-keywords-e (list (cons "ELSE"      af-syntax-else)
                              (cons "ELSIF"     af-syntax-elsif)
                              (cons "END"       af-syntax-end)
                              (cons "ENTRY"     af-syntax-entry)
                              (cons "EXCEPTION" af-syntax-exception)
                              (cons "EXIT"      af-syntax-exit))
  )

(defconst af-keywords-f (list (cons "FOR"      af-syntax-for)
                              (cons "FUNCTION" af-syntax-function))
  )

(defconst af-keywords-g (list (cons "GENERIC" af-syntax-generic)
                              (cons "GOTO"    af-syntax-goto))
  )

(defconst af-keywords-i (list (cons "IF" af-syntax-if)
                              (cons "IN" af-syntax-in)
                              (cons "IS" af-syntax-is))
  )

(defconst af-keywords-l (list (cons "LIMITED" af-syntax-limited)
                              (cons "LOOP"    af-syntax-loop))
  )

(defconst af-keywords-m (list (cons "MOD" af-syntax-mod))
  )

(defconst af-keywords-n (list (cons "NEW"  af-syntax-new)
                              (cons "NOT"  af-syntax-not)
                              (cons "NULL" af-syntax-null))
  )

(defconst af-keywords-o (list (cons "OF"     af-syntax-of)
                              (cons "OR"     af-syntax-or)
                              (cons "OTHERS" af-syntax-others)
                              (cons "OUT"    af-syntax-out)
                              )
  )

(defconst af-keywords-p (list (cons "PACKAGE"   af-syntax-package)
                              (cons "PRAGMA"    af-syntax-pragma)
                              (cons "PRIVATE"   af-syntax-private)
                              (cons "PROCEDURE" af-syntax-procedure)
                              (cons "PROTECTED" af-syntax-protected)
                              )
  )

(defconst af-keywords-r (list (cons "RAISE"   af-syntax-raise)
                              (cons "RANGE"   af-syntax-range)
                              (cons "RECORD"  af-syntax-record)
                              (cons "REM"     af-syntax-rem)
                              (cons "RENAMES" af-syntax-renames)
                              (cons "REQUEUE" af-syntax-requeue)
                              (cons "RETURN"  af-syntax-return)
                              (cons "REVERSE" af-syntax-reverse)
                              )
  )

(defconst af-keywords-s (list (cons "SELECT"   af-syntax-select)
                              (cons "SEPARATE" af-syntax-separate)
                              (cons "SUBTYPE"  af-syntax-subtype)
                              )
  )

(defconst af-keywords-t (list (cons "TAGGED"    af-syntax-tagged)
                              (cons "TASK"      af-syntax-task)
                              (cons "TERMINATE" af-syntax-terminate)
                              (cons "THEN"      af-syntax-then)
                              (cons "TYPE"      af-syntax-type)
                              )
  )

(defconst af-keywords-u (list (cons "UNTIL" af-syntax-until)
                              (cons "USE"   af-syntax-use)
                              )
  )

(defconst af-keywords-w (list (cons "WHEN"  af-syntax-when)
                              (cons "WHILE" af-syntax-while)
                              (cons "WITH"  af-syntax-with)
                              )
  )

(defconst af-keywords-x (list (cons "XOR" af-syntax-xor)
                              )
  )


(if nil  ;;;--Change this to t if you need to regenerate the match-previous
  (progn ;;;  and match-next fields of all af-syntax-* tokens.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; Tools for building Finite-State-Machines that describe portions of Ada
;;;; syntax.  We construct pseudo LL(0) FSM's to help us decide when we
;;;; have/have-not found sytnax errors.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar cons-pool-list nil)

(defun cons-pool (car cdr &optional l1 l2)

  ;;--Find a cons in the pool that is exactly (car . cdr).  The pool is
  ;;  an association list with association lists as members.

  (if (setq l1 (assq car cons-pool-list))     ; Locate all cons's this car
      (if (setq l2 (assq cdr (cdr l1)))       ; Locate one cons with this cdr
          (cdr l2)                            ; We are done
        (setq l2 (cons car cdr))              ; No cdr found, need new cons
        (setcdr l1 (cons (cons cdr            ; Insert in cdr list
                               l2)
                         (cdr l1)))
        l2
        )
    (setq l2 (cons car cdr))                  ; No car found, need new cons
    (setq cons-pool-list (cons (cons car      ; Insert in car list
                                (list (cons cdr l2)))
                          cons-pool-list))
    l2
    )

  );; cons-pool

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun sift-pool (func seq &optional tmp)

  ;;--Much like mapcar for a list except we create the new list with
  ;;  cons-pool and we drop all nil elements.

  (if seq
      (if (setq tmp (funcall func (car seq)))
          (cons-pool tmp
                     (sift-pool func (cdr seq)))
        (sift-pool func (cdr seq))
        )
    )

  );; sift-pool

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun append-pool (&rest lists)

  ;;--Just like append for lists, but we pool everything.

  (if lists
      (if (cdr lists)
          (if (consp (car lists))
              
              ;;--Append two lists.

              (let* ((head   (car lists))
                     (tail   (if (cdr (cdr lists))
                                 (apply 'append-pool (cdr lists))
                               (car (cdr lists))))
                     (length (length head)))
                (while (> length 0)
                  (setq length (1- length))
                  (setq tail (cons-pool (nth length head)
                                        tail))
                  )
                tail)

            ;;--Append single element and a list.

            (let* ((head   (car lists))
                   (tail   (if (cdr (cdr lists))
                               (apply 'append-pool (cdr lists))
                             (car (cdr lists)))))
              (cons-pool head tail))
            )

        ;;--Just a single element in argument.

        (car lists)
        )
    )

  );; append-pool

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun share-a-list (fsm)

  ;;--Recursively walk list, pool all cons's found.  Cannot handle circular
  ;;  lists.

  (if (consp fsm)
      (cons-pool (share-a-list (car fsm))
                 (share-a-list (cdr fsm)))
    fsm
    )

  );; share-a-list

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun unique-elem (elem list)

  ;;--Called to add elem to the list.  If elem is already on the list then
  ;;  just return the list.  Otherwise elem becomes the new list head.

  (if (memq elem list)
      list
    (cons-pool elem list)
    )

  );; unique-elem

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun unique-assoc (elem list)

  ;;--Called to add elem to the list.  If elem is already on the list then
  ;;  just return the list.  Otherwise elem becomes the new list head.

  (if (assq (car elem) list)
      list
    (cons-pool elem list)
    )

  );; unique-assoc

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun reverse-an-fsm (fsm &optional result)

  ;;--We need to be able to process our FSM's backwards (from the END to the
  ;;  BEGIN, so reverse the fsm and pool the result.

  (if (consp fsm)
      (cond

       ;;--To reverse an OR, keep elements in same sequence, just reverse
       ;;  them internally.

       ((eq (car fsm) 'or)
        (if (null (cdr fsm))
            fsm
          (cons-pool 'or
                     (reverse-an-fsm (cons 'or-internal
                                           (cdr fsm))))))

       ((eq (car fsm) 'or-internal)
        (setq fsm (cdr fsm))
        (if (null fsm)
            nil
          (cons-pool (reverse-an-fsm (car fsm))
                     (reverse-an-fsm (cons 'or-internal
                                           (cdr fsm))))))

       ;;--To reverse OPTIONAL or REPEAT, reverse the contents of the list
       ;;  but keep the symbol at the head of the new list.

       ((or (eq (car fsm) 'optional)
            (eq (car fsm) 'repeat))
        (cons-pool (car fsm)
                   (reverse-an-fsm (cdr fsm))))

       ;;--All other lists reverse car/cdr.

       (t
        (while fsm
          (setq result (cons-pool (reverse-an-fsm (car fsm))
                                  result))
          (setq fsm (cdr fsm))
          )
        result)
       );; cond

    ;;--Can't reverse anything that isn't a cons.

    (if (and fsm
             (not af-byte-compiling))
        (if (not (symbolp fsm))
            (af-debug-error "Finite state machine contains non-symbol [%S]."
                            fsm)
          (if (eq (symbol-value fsm) t)
              nil
            (if (af-zerobit (aref (symbol-value fsm)
                                  af-token-data-offset)
                            (logior af-flag-group-begin
                                    af-flag-group-begin-continue
                                    af-flag-group-continue
                                    af-flag-group-end))
                (af-debug-error "Symbol %S in FSM not a group token."
                                fsm)
              )
            )
          )
      )
    fsm
    )

  );; reverse-an-fsm

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun preload-an-fsm (fsm)

  ;;--Takes each FSM, create a shared forward copy and a shared backward copy.
  ;;  Then create new preload states for both of those.  We will process the
  ;;  preloads later.

  (let* (;;--Create two shared FSM's.  One forward, one backward.  Put
         ;;  them into lists of length 1.

         (shared-forward   (cons-pool (share-a-list   (append fsm
                                                              '(t)
                                                              nil))
                                      nil))
         (shared-backward  (cons-pool (reverse-an-fsm (append '(t)
                                                              fsm
                                                              nil))
                                      nil))

         ;;--Create two state associations for these preload states.

         (preload-forward  (cons-pool shared-forward
                                      'forward))
         (preload-backward (cons-pool shared-backward
                                      'backward)))

    ;;--Add these preloads onto our list of preloaded states.
    ;;  Create a new association element with the preload as the key and
    ;;  the 'forward/'backward (eventually the postload state) to be filled
    ;;  in later.  Add the new association to the preload-states association
    ;;  list.

    (setq preload-states (unique-elem preload-forward  preload-states))
    (setq preload-states (unique-elem preload-backward preload-states))
    )

  );; preload-an-fsm

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun interpret-preload (head tail &optional symbol)

  ;;--Called with a single preload element.  We look at it and try to
  ;;  insert it into the caller's new-postload list.

  (if (consp head)

      ;;--We are a list.  It may be a special list or it may just be a
      ;;  sequential list.

      (cond

       ;;--Sequence is an or element.  We want to treat the contents of the
       ;;  or as being a series of lists, each with one element from the or
       ;;  followed by the tail.

       ((eq 'or (setq symbol (car head)))
        (setq head (cdr head))
        (while head
          (interpret-preload (car head) tail)
          (setq head (cdr head))
          )
         )

       ;;--Sequence is an optional element.  We want to treat the contents of
       ;;  the optional as (a) being just the tail, and (b) being a sequence
       ;;  followed again by the tail.

       ((eq 'optional symbol)
        (if (cdr head)
            (if (cdr (cdr head))
                (interpret-preload (car (cdr head))
                                   (append (cdr (cdr head))
                                           tail))
              (interpret-preload (car (cdr head))
                                 tail)
              )
          )
        (if tail
            (interpret-preload (car tail) (cdr tail))
          )
        )

       ;;--Sequence is a repeat element.  We want to treat the contents of
       ;;  the repeat as (a) being just the tail, and (b) being a sequence,
       ;;  followed again by the repeat, followed again by the tail.

       ((eq 'repeat symbol)
        (if (cdr head)
            (if (cdr (cdr head))
                (interpret-preload (car (cdr head))
                                   (append (cdr (cdr head))
                                           (list head)
                                           tail))
              (interpret-preload (car (cdr head))
                                 (append (list head)
                                         tail))
              )
          )
        (if tail
            (interpret-preload (car tail) (cdr tail))
          )
        )

       ;;--Ordinary sequence, the new element is the head of the sequence
       ;;  plus the rest of the sequence concatenated with the tail.

       (t
        (if (cdr head)
            (interpret-preload symbol
                               (append (cdr head)
                                       tail))
          (interpret-preload symbol
                             tail)
          )
        )
       );; cond
     
    ;;--We are a simple symbol.  The postload state gets an unpooled
    ;;  association of (head . tail).

    (setq head (cons head tail))
    (if (not (member head new-postload))
        (setq new-postload (cons head new-postload))
      )
    )

  );; interpret-preload

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun handle-preload (direction
                       preload
                       &optional              ; Local variables
                       postload
                       new-postload
                       symbol
                       new-preload tmp)

  ;;--Returns the postload form of the preload state.  The preload state is
  ;;  a simple list of lists.  The head of each list is what-to-do-next
  ;;  and the tail of each list is what-to-do-after-that.  In the simple
  ;;  case, the head is just a symbol and it is the next symbol to read.
  ;;  We look at the head and decide what to do.  Mainly we try to come
  ;;  up with things that start with symbols.
  ;;
  ;;  The final postload state is another association list.  There is one
  ;;  element in the postload for each unique (eq) symbol in the preload
  ;;  state.  The car of the postload is the symbol and the cdr is the
  ;;  preload-states element that contains the new preload state.
  ;;
  ;;  Interpret each preload list and use it to create 1..N new-postload
  ;;  elements.

  (while preload
    (interpret-preload (car (car preload)) (cdr (car preload)))
    (setq preload (cdr preload))
    )

  ;;--We now have a new-postload association list.  These are associations
  ;;  of the form (symbol . FSM).  We want to group all entries with the
  ;;  same symbol together and form new preload states.

  (setq new-postload 
        (sort new-postload
              (lambda (left right)
                (setq left  (car left)
                      right (car right))
                (af-debug-check
                 (if (or (not (symbolp left))
                         (not (symbolp right)))
                     (error "new-postload sort, one CAR is not a symbolp.")
                     )
                 )
                (not (string< (symbol-name left) (symbol-name right)))
                )))
  (while new-postload

    ;;--symbol is the symbol we are using to create the new state.
    ;;  new-preload is the cdr's of all elements with symbol as their symbol.
    ;;  new-postload has all symbol entries removed.

    (setq symbol      (car (car new-postload))
          new-preload nil)

    ;;--Move all symbol entries at the front of new-postload that use symbol.

    (while (eq symbol (car (car new-postload)))
      (setq tmp new-preload
            new-preload new-postload
            new-postload (cdr new-postload))
      (setcdr new-preload tmp)
      (setcar new-preload (cdr (car new-preload)))
      (if (null (car new-preload))
          (setq new-preload (cdr new-preload))
        )
      )

    ;;--Now move all symbol entries within new-postload that use symbol.

    (setq preload new-postload)
    (while (cdr preload)
      (if (eq symbol (car (car (cdr preload))))
          (progn
            (setq tmp new-preload
                  new-preload (cdr preload))
            (setcdr preload (cdr new-preload))
            (setcdr new-preload tmp)
            (setcar new-preload (cdr (car new-preload)))
            (if (null (car new-preload))
                (setq new-preload (cdr new-preload))
              )
            )
        (setq preload (cdr preload))
        )
      )

    ;;--Put this new-preload into the preload-states list.

    (if new-preload
        (setq new-preload    (share-a-list new-preload)
              preload-states (unique-assoc (cons-pool new-preload
                                                      direction)
                                           preload-states))
      )

    ;;--Add a new association onto the start of the postload state.

    (setq postload (unique-elem (cons-pool symbol
                                           (assq new-preload preload-states))
                                postload))

    ;;--Add all of these new-preload items to the symbol's initial preload
    ;;  list.

    (if (and new-preload
             preload-direction
             (not (eq (symbol-value symbol) t)))
        (if (setq preload (aref (symbol-value symbol) preload-direction))
            (progn
              (setq new-preload
                    (sift-pool
                     (lambda (fsm) (if (memq fsm preload)
                                       nil
                                     fsm))
                     new-preload))
              (if new-preload
                  (aset (symbol-value symbol) preload-direction
                        (append-pool preload
                                     new-preload))
                )
              )
          (aset (symbol-value symbol) preload-direction new-preload)
          )
      )
    );; while

  ;;--Return what we found.

  postload


  );; handle-preload


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun create-all-fsms ( &optional fixup symbol vector preload )

  ;;--Create all of the syntax FSM's that we need for the ada-format library.
  ;;  These describe Ada syntax in various ways and we use them to detect
  ;;  certain sorts of syntax errors.

  (setq cons-pool-list nil)

  (preload-an-fsm '(af-syntax-left-curly
                    af-syntax-right-curly))

  (preload-an-fsm '(af-syntax-left-label
                    af-syntax-right-label))

  (preload-an-fsm '(af-syntax-left-square
                    af-syntax-right-square))

  (preload-an-fsm '(af-syntax-left-paren
                    af-syntax-right-paren))

  (preload-an-fsm '(af-syntax-accept
                    (optional
                     af-syntax-do
                     (or af-syntax-end-name
                         af-syntax-end))))

  (preload-an-fsm '(af-syntax-case
                    (repeat af-syntax-when)
                    af-syntax-end-case))

  (preload-an-fsm '((optional af-syntax-declare)
                    af-syntax-begin
                    (optional af-syntax-exception-section
                              (repeat af-syntax-when))
                    (or af-syntax-end
                        af-syntax-end-name)))

  (preload-an-fsm '(af-syntax-for
                    af-syntax-loop
                    af-syntax-end-loop))

  (preload-an-fsm '(af-syntax-for
                    af-syntax-use-of-for))

  (preload-an-fsm '(af-syntax-function
                    (optional af-syntax-begin
                              (optional af-syntax-exception-section
                                        (repeat af-syntax-when))
                              (or af-syntax-end
                                  af-syntax-end-name
                                  af-syntax-end-func))))

  (preload-an-fsm '(af-syntax-generic
                    af-syntax-function))

  (preload-an-fsm '(af-syntax-generic
                    af-syntax-package
                    (optional af-syntax-private-section)
                    (or af-syntax-end
                        af-syntax-end-name)))

  (preload-an-fsm '(af-syntax-generic
                    af-syntax-procedure))

  (setq preload-states (unique-elem
                        (share-a-list   '(((af-syntax-function t)) .
                                          backward))
                        preload-states))

  (setq preload-states (unique-elem
                        (share-a-list   '(((af-syntax-package t)) .
                                          backward))
                        preload-states))

  (setq preload-states (unique-elem
                        (share-a-list   '(((af-syntax-procedure t)) .
                                          backward))
                        preload-states))

  (preload-an-fsm '(af-syntax-if
                    af-syntax-then
                    (repeat af-syntax-elsif af-syntax-then)
                    (optional af-syntax-else)
                    af-syntax-end-if))

  (preload-an-fsm '(af-syntax-loop
                    af-syntax-end-loop))

  (preload-an-fsm '(af-syntax-package-body
                    (optional af-syntax-begin
                              (optional af-syntax-exception-section
                                        (repeat af-syntax-when))
                              )
                    (or af-syntax-end
                        af-syntax-end-name)))

  (preload-an-fsm '(af-syntax-package
                    (optional af-syntax-private-section)
                    (or af-syntax-end
                        af-syntax-end-name)))

  (preload-an-fsm '(af-syntax-protected-body
                    (or af-syntax-end
                        af-syntax-end-name)))

  (preload-an-fsm '(af-syntax-protected
                    (optional af-syntax-private-section)
                    (or af-syntax-end
                        af-syntax-end-name)))

  (preload-an-fsm '(af-syntax-protected-type
                    (optional af-syntax-private-section)
                    (or af-syntax-end
                        af-syntax-end-name)))

  (preload-an-fsm '(af-syntax-procedure
                    (optional af-syntax-begin
                              (optional af-syntax-exception-section
                                        (repeat af-syntax-when))
                              (or af-syntax-end
                                  af-syntax-end-name))))

  (preload-an-fsm '((or af-syntax-record
                        af-syntax-record-at)
                    af-syntax-end-record))

  (preload-an-fsm '(af-syntax-select
                    (repeat (or af-syntax-when
                                af-syntax-or-select))
                    (optional af-syntax-else)
                    af-syntax-end-select))

  (preload-an-fsm '(af-syntax-task-body
                    af-syntax-begin
                    (optional af-syntax-exception-section
                              (repeat af-syntax-when))
                    (or af-syntax-end
                        af-syntax-end-name)))

  (preload-an-fsm '(af-syntax-task
                    (optional af-syntax-private-section)
                    (or af-syntax-end
                        af-syntax-end-name)))

  (preload-an-fsm '(af-syntax-task-type
                    (optional af-syntax-private-section)
                    (or af-syntax-end
                        af-syntax-end-name)))

  (preload-an-fsm '(af-syntax-while
                    af-syntax-loop
                    af-syntax-end-loop))

  ;;--Loop over all preload states which have not been postloaded yet.

  (setq preload-direction af-token-match-next-offset)
  (while (setq preload (rassq 'forward preload-states))
    (setcdr preload (handle-preload 'forward (car preload)))
    )

  (setq preload-direction af-token-match-previous-offset)
  (while (setq preload (rassq 'backward preload-states))
    (setcdr preload (handle-preload 'backward (car preload)))
    )

  ;;--Loop over all fixup symbols.  Take both preload states from each symbol 
  ;;  and add them to our preload-states list.  Then process any unprocessed
  ;;  states.

  (setq fixup fixup-list)
  (while fixup
    (setq symbol (car fixup)
          vector (symbol-value symbol))

    ;;--Get the backward preload.

    (setq preload        (cons-pool (aref vector
                                          af-token-match-previous-offset)
                                    'backward)
          preload-states (unique-assoc preload
                                       preload-states))
    (aset vector af-token-match-previous-offset
          (assq (aref vector af-token-match-previous-offset) preload-states))

    ;;--Get the forward preload.

    (setq preload        (cons-pool (aref vector af-token-match-next-offset)
                                    'forward)
          preload-states (unique-assoc preload
                                       preload-states))
    (aset vector af-token-match-next-offset
          (assq (aref vector af-token-match-next-offset) preload-states))

    ;;--Get the next symbol.

    (setq fixup (cdr fixup))
    )

  ;;--Loop again over all preload states which have not been postloaded yet.

  (setq preload-direction nil)
  (while (setq preload (rassq 'forward preload-states))
    (setcdr preload (handle-preload 'forward (car preload)))
    )

  (setq preload-direction nil)
  (while (setq preload (rassq 'backward preload-states))
    (setcdr preload (handle-preload 'backward (car preload)))
    )

  ;;--Loop over all of the symbols in the fixup-list.  Convert their
  ;;  paired (preload . postload) states into pure postload values.

  (setq fixup fixup-list)
  (while fixup

    (setq symbol         (car fixup)
          vector         (symbol-value symbol))

    (aset vector af-token-match-previous-offset
          (cdr (aref vector af-token-match-previous-offset)))
    (aset vector af-token-match-next-offset
          (cdr (aref vector af-token-match-next-offset)))

    (setq fixup (cdr fixup))
    );; while fixup

  ;;--preload-states is a list of (preload . postload) elements.  We use
  ;;  mapcar to get a list of just the postload states.

  (setq preload (mapcar 'cdr preload-states))

  ;;--Each postload state is a list if (symbol . (preload . postload)).
  ;;  We want to create a new single list containing every one of these
  ;;  elements.  We don't want to modify any existing lists.
  ;;
  ;;  The nconc will create a list of all the postload states plus a nil
  ;;  on the end.  The append will copy all of the postload states and create
  ;;  a new list with the nil on the end.

  (setq preload (apply 'append
                       (nconc preload
                              (list nil))))

  ;;--Now go through the list of postload elements and eliminate all duplicate
  ;;  elements.

  (setq fixup preload)
  (while (cdr fixup)
    (setcdr fixup (delq (car fixup) (cdr fixup)))
    (setq fixup (cdr fixup))
    )

  ;;--Now go through all of the unique (symbol . (preload .postload)) elements
  ;;  and change them to be (symbol . postload) only.

  (while preload
    (setcdr (car preload)
            (cdr                              ; postload
             (cdr                             ; (preload .postload)
              (car preload))))                ; (symbol . (preload . postload))
    (setq preload (cdr preload))
    )

  );; create-all-fsms


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun print-all-fsms ( &optional             ; Local variables
                        cons-list
                        fixup
                        symbol
                        vector
                        fsm
                        len
                        func )

  ;;--Print all of the match-previous/next values so we can initialize them
  ;;  more rapidly than we can if we were to fully recompute them each time.

  (switch-to-buffer-other-window "*ada-format.el init data*")
  (erase-buffer)

  ;;--First we create a cons-list that contains one element for every
  ;;  element in all of the match lists for all of the group tokens.
  ;;  Those are the ones in the fixup-list.

  (setq fixup fixup-list)
  (setq func (lambda (fsm)
               ;;--We recursively mapcar on every argument that is not
               ;;  already on the cons-list.
               (if (memq fsm cons-list)
                   nil
                 (setq cons-list (cons fsm cons-list))
                 (if (consp fsm)
                     (mapcar func fsm)
                   )
                 )
               ))
  (while fixup
    (setq symbol         (car fixup)
          vector         (symbol-value symbol))

    (insert (format ";;; A%04d : %S previous\n"
                    (1+ (length cons-list)) symbol))
    (setq fsm (aref vector af-token-match-previous-offset))
    (if (not (memq fsm cons-list))
        (setq cons-list (cons fsm cons-list))
      )
    (mapcar func fsm)

    (insert (format ";;; A%04d : %S next\n" (1+ (length cons-list)) symbol))
    (setq fsm (aref vector af-token-match-next-offset))
    (if (not (memq fsm cons-list))
        (setq cons-list (cons fsm cons-list))
      )
    (mapcar func fsm)

    (setq fixup (cdr fixup))
    );; while fixup

  ;;--Loop over all of the elements in the cons-list.  Put out a (let)
  ;;  for every element.

  (insert "\n  (let (\n")
  (setq fixup cons-list)
  (while fixup

    (cond ((consp (car fixup))
           (let ((line
                  (format "         (A%04d (make-list %4d nil)) ;;; %S"
                          (length fixup) (length (car fixup)) (car fixup))))
             (if (> (length line) 1000)
                 (insert (substring line 0 1001) "...\n")
               (insert line "\n")
               )
             )
           )

          ((symbolp (car fixup))
           (insert (format "         (A%04d %S)\n"
                           (length fixup) (car fixup))))

          (t
           (af-debug-error "Match-list element not cons/symbol <<%S>>?"
                           (car fixup)))
          );; cond

    (setq fixup (cdr fixup))
    );; while
  (insert "      )\n")

  ;;--Loop over all elements in the cons-list.  Put out assignments to
  ;;  create our circular data structures.

  (insert "\n")
  (setq fixup cons-list)
  (while fixup

    (if (consp (car fixup))
        (progn
          (setq fsm (car fixup)
                len (length fsm))
          (while (> len 0)
            (setq len (1- len))
            (let ((line
                   (format "    (setcar (nthcdr %4d A%04d) A%04d) ;;; %S"
                           len
                           (length fixup)
                           (length (memq (nth len fsm) cons-list))
                           (nth len fsm))))
              (if (> (length line) 900)
                  (insert (substring line 0 901) "...\n")
                (insert line "\n")
                )
              )
            )
          )
      )

    (setq fixup (cdr fixup))
    );; while

  ;;--Loop over all fixup symbols and put out assignments to set their
  ;;  match-previous and match-next fields.

  (insert "\n")
  (setq fixup fixup-list)
  (while fixup
    (setq symbol         (car fixup)
          vector         (symbol-value symbol))

    (insert (format "    (aset %-31S af-token-match-previous-offset A%04d)\n"
                    symbol
                    (length (memq (aref vector af-token-match-previous-offset)
                                  cons-list))))
                            
    (insert (format "    (aset %-31S af-token-match-next-offset     A%04d)\n"
                    symbol
                    (length (memq (aref vector af-token-match-next-offset)
                                  cons-list))))
                            
    (setq fixup (cdr fixup))
    );; while fixup

  ;;--Get back in the correct window.

  (insert "  )\n")
  (select-window (previous-window))

  );; print-all-fsms

(create-all-fsms)
(print-all-fsms)
);; if nil (progn


;; if (not) nil
(progn

;;--These are the constant declarations created by print-all-fsms above when
;;  we need to regenerate the match lists for the tokens.

;;; A0001 : af-syntax-while previous
;;; A0004 : af-syntax-while next
;;; A0009 : af-syntax-when previous
;;; A0037 : af-syntax-when next
;;; A0054 : af-syntax-use-of-for previous
;;; A0057 : af-syntax-use-of-for next
;;; A0057 : af-syntax-then previous
;;; A0064 : af-syntax-then next
;;; A0070 : af-syntax-task-type previous
;;; A0070 : af-syntax-task-type next
;;; A0073 : af-syntax-task-body previous
;;; A0073 : af-syntax-task-body next
;;; A0077 : af-syntax-task previous
;;; A0077 : af-syntax-task next
;;; A0077 : af-syntax-select previous
;;; A0077 : af-syntax-select next
;;; A0078 : af-syntax-record-at previous
;;; A0078 : af-syntax-record-at next
;;; A0081 : af-syntax-record previous
;;; A0081 : af-syntax-record next
;;; A0081 : af-syntax-protected-type previous
;;; A0081 : af-syntax-protected-type next
;;; A0081 : af-syntax-protected-body previous
;;; A0081 : af-syntax-protected-body next
;;; A0082 : af-syntax-protected previous
;;; A0082 : af-syntax-protected next
;;; A0082 : af-syntax-procedure previous
;;; A0085 : af-syntax-procedure next
;;; A0086 : af-syntax-private-section previous
;;; A0097 : af-syntax-private-section next
;;; A0097 : af-syntax-package-body previous
;;; A0097 : af-syntax-package-body next
;;; A0098 : af-syntax-package previous
;;; A0098 : af-syntax-package next
;;; A0098 : af-syntax-or-select previous
;;; A0099 : af-syntax-or-select next
;;; A0099 : af-syntax-loop previous
;;; A0102 : af-syntax-loop next
;;; A0103 : af-syntax-if previous
;;; A0103 : af-syntax-if next
;;; A0104 : af-syntax-generic previous
;;; A0104 : af-syntax-generic next
;;; A0106 : af-syntax-function previous
;;; A0106 : af-syntax-function next
;;; A0110 : af-syntax-for previous
;;; A0110 : af-syntax-for next
;;; A0113 : af-syntax-exception-section previous
;;; A0114 : af-syntax-exception-section next
;;; A0117 : af-syntax-end-select previous
;;; A0119 : af-syntax-end-select next
;;; A0119 : af-syntax-end-record previous
;;; A0124 : af-syntax-end-record next
;;; A0124 : af-syntax-end-func previous
;;; A0128 : af-syntax-end-func next
;;; A0128 : af-syntax-end-name previous
;;; A0139 : af-syntax-end-name next
;;; A0139 : af-syntax-end-loop previous
;;; A0141 : af-syntax-end-loop next
;;; A0141 : af-syntax-end-case previous
;;; A0143 : af-syntax-end-case next
;;; A0143 : af-syntax-end-if previous
;;; A0145 : af-syntax-end-if next
;;; A0145 : af-syntax-end previous
;;; A0145 : af-syntax-end next
;;; A0145 : af-syntax-elsif previous
;;; A0146 : af-syntax-elsif next
;;; A0146 : af-syntax-else previous
;;; A0147 : af-syntax-else next
;;; A0148 : af-syntax-do previous
;;; A0149 : af-syntax-do next
;;; A0149 : af-syntax-declare previous
;;; A0149 : af-syntax-declare next
;;; A0149 : af-syntax-case previous
;;; A0149 : af-syntax-case next
;;; A0151 : af-syntax-begin previous
;;; A0152 : af-syntax-begin next
;;; A0154 : af-syntax-accept previous
;;; A0154 : af-syntax-accept next
;;; A0156 : af-syntax-right-square previous
;;; A0159 : af-syntax-right-square next
;;; A0159 : af-syntax-left-square previous
;;; A0159 : af-syntax-left-square next
;;; A0162 : af-syntax-right-paren previous
;;; A0165 : af-syntax-right-paren next
;;; A0165 : af-syntax-left-paren previous
;;; A0165 : af-syntax-left-paren next
;;; A0168 : af-syntax-right-label previous
;;; A0171 : af-syntax-right-label next
;;; A0171 : af-syntax-left-label previous
;;; A0171 : af-syntax-left-label next
;;; A0174 : af-syntax-right-curly previous
;;; A0177 : af-syntax-right-curly next
;;; A0177 : af-syntax-left-curly previous
;;; A0177 : af-syntax-left-curly next
;;; A0180 : af-syntax-end-of-line previous
;;; A0181 : af-syntax-end-of-line next

  (let (
         (A0180 "nil")
         (A0179 af-syntax-right-curly)
         (A0178 (make-list    2 nil)) 
         (A0177 (make-list    1 nil)) 
         (A0176 af-syntax-left-curly)
         (A0175 (make-list    2 nil)) 
         (A0174 (make-list    1 nil)) 
         (A0173 af-syntax-right-label)
         (A0172 (make-list    2 nil)) 
         (A0171 (make-list    1 nil)) 
         (A0170 af-syntax-left-label)
         (A0169 (make-list    2 nil)) 
         (A0168 (make-list    1 nil)) 
         (A0167 af-syntax-right-paren)
         (A0166 (make-list    2 nil)) 
         (A0165 (make-list    1 nil)) 
         (A0164 af-syntax-left-paren)
         (A0163 (make-list    2 nil)) 
         (A0162 (make-list    1 nil)) 
         (A0161 af-syntax-right-square)
         (A0160 (make-list    2 nil)) 
         (A0159 (make-list    1 nil)) 
         (A0158 af-syntax-left-square)
         (A0157 (make-list    2 nil)) 
         (A0156 (make-list    1 nil)) 
         (A0155 (make-list    3 nil)) 
         (A0154 (make-list    2 nil)) 
         (A0153 (make-list    5 nil)) 
         (A0152 (make-list    4 nil)) 
         (A0151 (make-list    6 nil)) 
         (A0150 (make-list    3 nil)) 
         (A0149 (make-list    2 nil)) 
         (A0148 (make-list    1 nil)) 
         (A0147 (make-list    2 nil)) 
         (A0146 (make-list    4 nil)) 
         (A0145 (make-list    1 nil)) 
         (A0144 (make-list    2 nil)) 
         (A0143 (make-list    2 nil)) 
         (A0142 (make-list    3 nil)) 
         (A0141 (make-list    2 nil)) 
         (A0140 (make-list    4 nil)) 
         (A0139 (make-list    1 nil)) 
         (A0138 (make-list    3 nil)) 
         (A0137 (make-list    3 nil)) 
         (A0136 af-syntax-protected-body)
         (A0135 (make-list    2 nil)) 
         (A0134 (make-list    3 nil)) 
         (A0133 (make-list    6 nil)) 
         (A0132 af-syntax-accept)
         (A0131 (make-list    2 nil)) 
         (A0130 af-syntax-do)
         (A0129 (make-list    2 nil)) 
         (A0128 (make-list   12 nil)) 
         (A0127 (make-list    3 nil)) 
         (A0126 (make-list    2 nil)) 
         (A0125 (make-list    2 nil)) 
         (A0124 (make-list    3 nil)) 
         (A0123 af-syntax-record-at)
         (A0122 (make-list    2 nil)) 
         (A0121 af-syntax-record)
         (A0120 (make-list    2 nil)) 
         (A0119 (make-list    2 nil)) 
         (A0118 (make-list    4 nil)) 
         (A0117 (make-list    4 nil)) 
         (A0116 (make-list    5 nil)) 
         (A0115 (make-list    5 nil)) 
         (A0114 (make-list    4 nil)) 
         (A0113 (make-list    1 nil)) 
         (A0112 af-syntax-use-of-for)
         (A0111 (make-list    2 nil)) 
         (A0110 (make-list    2 nil)) 
         (A0109 (make-list    5 nil)) 
         (A0108 (make-list    5 nil)) 
         (A0107 (make-list    5 nil)) 
         (A0106 (make-list    2 nil)) 
         (A0105 (make-list    4 nil)) 
         (A0104 (make-list    3 nil)) 
         (A0103 (make-list    1 nil)) 
         (A0102 (make-list    1 nil)) 
         (A0101 af-syntax-while)
         (A0100 (make-list    2 nil)) 
         (A0099 (make-list    3 nil)) 
         (A0098 (make-list    3 nil)) 
         (A0097 (make-list    3 nil)) 
         (A0096 af-syntax-task-type)
         (A0095 (make-list    2 nil)) 
         (A0094 af-syntax-task)
         (A0093 (make-list    2 nil)) 
         (A0092 af-syntax-protected-type)
         (A0091 (make-list    2 nil)) 
         (A0090 af-syntax-protected)
         (A0089 (make-list    2 nil)) 
         (A0088 af-syntax-package)
         (A0087 (make-list    3 nil)) 
         (A0086 (make-list    5 nil)) 
         (A0085 (make-list    2 nil)) 
         (A0084 af-syntax-generic)
         (A0083 (make-list    2 nil)) 
         (A0082 (make-list    2 nil)) 
         (A0081 (make-list    2 nil)) 
         (A0080 af-syntax-end-record)
         (A0079 (make-list    2 nil)) 
         (A0078 (make-list    1 nil)) 
         (A0077 (make-list    4 nil)) 
         (A0076 (make-list    4 nil)) 
         (A0075 (make-list    4 nil)) 
         (A0074 (make-list    4 nil)) 
         (A0073 (make-list    1 nil)) 
         (A0072 af-syntax-private-section)
         (A0071 (make-list    3 nil)) 
         (A0070 (make-list    3 nil)) 
         (A0069 (make-list    4 nil)) 
         (A0068 (make-list    2 nil)) 
         (A0067 af-syntax-end-if)
         (A0066 (make-list    2 nil)) 
         (A0065 (make-list    2 nil)) 
         (A0064 (make-list    3 nil)) 
         (A0063 af-syntax-if)
         (A0062 (make-list    2 nil)) 
         (A0061 af-syntax-then)
         (A0060 (make-list    3 nil)) 
         (A0059 af-syntax-elsif)
         (A0058 (make-list    2 nil)) 
         (A0057 (make-list    2 nil)) 
         (A0056 af-syntax-for)
         (A0055 (make-list    2 nil)) 
         (A0054 (make-list    1 nil)) 
         (A0053 (make-list    9 nil)) 
         (A0052 (make-list    9 nil)) 
         (A0051 (make-list    5 nil)) 
         (A0050 (make-list    5 nil)) 
         (A0049 af-syntax-end-name)
         (A0048 (make-list    2 nil)) 
         (A0047 af-syntax-end-func)
         (A0046 (make-list    2 nil)) 
         (A0045 af-syntax-end-case)
         (A0044 (make-list    2 nil)) 
         (A0043 af-syntax-end)
         (A0042 (make-list    2 nil)) 
         (A0041 af-syntax-end-select)
         (A0040 (make-list    2 nil)) 
         (A0039 af-syntax-else)
         (A0038 (make-list    2 nil)) 
         (A0037 (make-list    8 nil)) 
         (A0036 (make-list    6 nil)) 
         (A0035 (make-list    7 nil)) 
         (A0034 (make-list    2 nil)) 
         (A0033 (make-list    6 nil)) 
         (A0032 af-syntax-when)
         (A0031 (make-list    4 nil)) 
         (A0030 af-syntax-select)
         (A0029 (make-list    2 nil)) 
         (A0028 af-syntax-or-select)
         (A0027 (make-list    4 nil)) 
         (A0026 (make-list    1 nil)) 
         (A0025 af-syntax-task-body)
         (A0024 (make-list    2 nil)) 
         (A0023 af-syntax-procedure)
         (A0022 (make-list    2 nil)) 
         (A0021 af-syntax-package-body)
         (A0020 (make-list    2 nil)) 
         (A0019 af-syntax-function)
         (A0018 (make-list    2 nil)) 
         (A0017 af-syntax-declare)
         (A0016 (make-list    2 nil)) 
         (A0015 af-syntax-begin)
         (A0014 (make-list    7 nil)) 
         (A0013 af-syntax-exception-section)
         (A0012 (make-list    2 nil)) 
         (A0011 af-syntax-case)
         (A0010 (make-list    2 nil)) 
         (A0009 (make-list    5 nil)) 
         (A0008 af-syntax-end-loop)
         (A0007 (make-list    2 nil)) 
         (A0006 af-syntax-loop)
         (A0005 (make-list    2 nil)) 
         (A0004 (make-list    1 nil)) 
         (A0003 t)
         (A0002 (make-list    1 nil)) 
         (A0001 (make-list    1 nil)) 
      )

    (setcar (nthcdr    1 A0178) A0002) 
    (setcar (nthcdr    0 A0178) A0179) ;;; af-syntax-right-curly
    (setcar (nthcdr    0 A0177) A0178) 
    (setcar (nthcdr    1 A0175) A0002) 
    (setcar (nthcdr    0 A0175) A0176) ;;; af-syntax-left-curly
    (setcar (nthcdr    0 A0174) A0175) 
    (setcar (nthcdr    1 A0172) A0002) 
    (setcar (nthcdr    0 A0172) A0173) ;;; af-syntax-right-label
    (setcar (nthcdr    0 A0171) A0172) 
    (setcar (nthcdr    1 A0169) A0002) 
    (setcar (nthcdr    0 A0169) A0170) ;;; af-syntax-left-label
    (setcar (nthcdr    0 A0168) A0169) 
    (setcar (nthcdr    1 A0166) A0002) 
    (setcar (nthcdr    0 A0166) A0167) ;;; af-syntax-right-paren
    (setcar (nthcdr    0 A0165) A0166) 
    (setcar (nthcdr    1 A0163) A0002) 
    (setcar (nthcdr    0 A0163) A0164) ;;; af-syntax-left-paren
    (setcar (nthcdr    0 A0162) A0163) 
    (setcar (nthcdr    1 A0160) A0002) 
    (setcar (nthcdr    0 A0160) A0161) ;;; af-syntax-right-square
    (setcar (nthcdr    0 A0159) A0160) 
    (setcar (nthcdr    1 A0157) A0002) 
    (setcar (nthcdr    0 A0157) A0158) ;;; af-syntax-left-square
    (setcar (nthcdr    0 A0156) A0157) 
    (setcar (nthcdr    2 A0155) A0048) 
    (setcar (nthcdr    1 A0155) A0042) 
    (setcar (nthcdr    0 A0155) A0130) ;;; af-syntax-do
    (setcar (nthcdr    1 A0154) A0002) 
    (setcar (nthcdr    0 A0154) A0155) 
    (setcar (nthcdr    4 A0153) A0116) 
    (setcar (nthcdr    3 A0153) A0048) 
    (setcar (nthcdr    2 A0153) A0046) 
    (setcar (nthcdr    1 A0153) A0042) 
    (setcar (nthcdr    0 A0153) A0013) ;;; af-syntax-exception-section
    (setcar (nthcdr    3 A0152) A0153) 
    (setcar (nthcdr    2 A0152) A0048) 
    (setcar (nthcdr    1 A0152) A0046) 
    (setcar (nthcdr    0 A0152) A0042) 
    (setcar (nthcdr    5 A0151) A0026) 
    (setcar (nthcdr    4 A0151) A0024) 
    (setcar (nthcdr    3 A0151) A0022) 
    (setcar (nthcdr    2 A0151) A0020) 
    (setcar (nthcdr    1 A0151) A0018) 
    (setcar (nthcdr    0 A0151) A0016) 
    (setcar (nthcdr    2 A0150) A0150) 
    (setcar (nthcdr    1 A0150) A0044) 
    (setcar (nthcdr    0 A0150) A0032) ;;; af-syntax-when
    (setcar (nthcdr    1 A0149) A0150) 
    (setcar (nthcdr    0 A0149) A0044) 
    (setcar (nthcdr    0 A0148) A0131) 
    (setcar (nthcdr    1 A0147) A0040) 
    (setcar (nthcdr    0 A0147) A0066) 
    (setcar (nthcdr    3 A0146) A0031) 
    (setcar (nthcdr    2 A0146) A0060) 
    (setcar (nthcdr    1 A0146) A0029) 
    (setcar (nthcdr    0 A0146) A0027) 
    (setcar (nthcdr    0 A0145) A0060) 
    (setcar (nthcdr    1 A0144) A0060) 
    (setcar (nthcdr    0 A0144) A0039) ;;; af-syntax-else
    (setcar (nthcdr    1 A0143) A0060) 
    (setcar (nthcdr    0 A0143) A0144) 
    (setcar (nthcdr    2 A0142) A0142) 
    (setcar (nthcdr    1 A0142) A0010) 
    (setcar (nthcdr    0 A0142) A0032) ;;; af-syntax-when
    (setcar (nthcdr    1 A0141) A0142) 
    (setcar (nthcdr    0 A0141) A0010) 
    (setcar (nthcdr    3 A0140) A0026) 
    (setcar (nthcdr    2 A0140) A0100) 
    (setcar (nthcdr    1 A0140) A0055) 
    (setcar (nthcdr    0 A0140) A0006) ;;; af-syntax-loop
    (setcar (nthcdr    0 A0139) A0140) 
    (setcar (nthcdr    2 A0138) A0137) 
    (setcar (nthcdr    1 A0138) A0012) 
    (setcar (nthcdr    0 A0138) A0032) ;;; af-syntax-when
    (setcar (nthcdr    2 A0137) A0138) 
    (setcar (nthcdr    1 A0137) A0034) 
    (setcar (nthcdr    0 A0137) A0032) ;;; af-syntax-when
    (setcar (nthcdr    1 A0135) A0002) 
    (setcar (nthcdr    0 A0135) A0136) ;;; af-syntax-protected-body
    (setcar (nthcdr    2 A0134) A0026) 
    (setcar (nthcdr    1 A0134) A0083) 
    (setcar (nthcdr    0 A0134) A0088) ;;; af-syntax-package
    (setcar (nthcdr    5 A0133) A0095) 
    (setcar (nthcdr    4 A0133) A0093) 
    (setcar (nthcdr    3 A0133) A0091) 
    (setcar (nthcdr    2 A0133) A0089) 
    (setcar (nthcdr    1 A0133) A0134) 
    (setcar (nthcdr    0 A0133) A0072) ;;; af-syntax-private-section
    (setcar (nthcdr    1 A0131) A0002) 
    (setcar (nthcdr    0 A0131) A0132) ;;; af-syntax-accept
    (setcar (nthcdr    1 A0129) A0131) 
    (setcar (nthcdr    0 A0129) A0130) ;;; af-syntax-do
    (setcar (nthcdr   11 A0128) A0137) 
    (setcar (nthcdr   10 A0128) A0095) 
    (setcar (nthcdr    9 A0128) A0093) 
    (setcar (nthcdr    8 A0128) A0091) 
    (setcar (nthcdr    7 A0128) A0135) 
    (setcar (nthcdr    6 A0128) A0089) 
    (setcar (nthcdr    5 A0128) A0133) 
    (setcar (nthcdr    4 A0128) A0020) 
    (setcar (nthcdr    3 A0128) A0087) 
    (setcar (nthcdr    2 A0128) A0012) 
    (setcar (nthcdr    1 A0128) A0129) 
    (setcar (nthcdr    0 A0128) A0035) 
    (setcar (nthcdr    2 A0127) A0127) 
    (setcar (nthcdr    1 A0127) A0126) 
    (setcar (nthcdr    0 A0127) A0032) ;;; af-syntax-when
    (setcar (nthcdr    1 A0126) A0125) 
    (setcar (nthcdr    0 A0126) A0013) ;;; af-syntax-exception-section
    (setcar (nthcdr    1 A0125) A0018) 
    (setcar (nthcdr    0 A0125) A0015) ;;; af-syntax-begin
    (setcar (nthcdr    2 A0124) A0127) 
    (setcar (nthcdr    1 A0124) A0126) 
    (setcar (nthcdr    0 A0124) A0125) 
    (setcar (nthcdr    1 A0122) A0002) 
    (setcar (nthcdr    0 A0122) A0123) ;;; af-syntax-record-at
    (setcar (nthcdr    1 A0120) A0002) 
    (setcar (nthcdr    0 A0120) A0121) ;;; af-syntax-record
    (setcar (nthcdr    1 A0119) A0122) 
    (setcar (nthcdr    0 A0119) A0120) 
    (setcar (nthcdr    3 A0118) A0031) 
    (setcar (nthcdr    2 A0118) A0029) 
    (setcar (nthcdr    1 A0118) A0027) 
    (setcar (nthcdr    0 A0118) A0039) ;;; af-syntax-else
    (setcar (nthcdr    3 A0117) A0031) 
    (setcar (nthcdr    2 A0117) A0029) 
    (setcar (nthcdr    1 A0117) A0027) 
    (setcar (nthcdr    0 A0117) A0118) 
    (setcar (nthcdr    4 A0116) A0115) 
    (setcar (nthcdr    3 A0116) A0048) 
    (setcar (nthcdr    2 A0116) A0046) 
    (setcar (nthcdr    1 A0116) A0042) 
    (setcar (nthcdr    0 A0116) A0032) ;;; af-syntax-when
    (setcar (nthcdr    4 A0115) A0116) 
    (setcar (nthcdr    3 A0115) A0048) 
    (setcar (nthcdr    2 A0115) A0046) 
    (setcar (nthcdr    1 A0115) A0042) 
    (setcar (nthcdr    0 A0115) A0032) ;;; af-syntax-when
    (setcar (nthcdr    3 A0114) A0115) 
    (setcar (nthcdr    2 A0114) A0048) 
    (setcar (nthcdr    1 A0114) A0046) 
    (setcar (nthcdr    0 A0114) A0042) 
    (setcar (nthcdr    0 A0113) A0035) 
    (setcar (nthcdr    1 A0111) A0002) 
    (setcar (nthcdr    0 A0111) A0112) ;;; af-syntax-use-of-for
    (setcar (nthcdr    1 A0110) A0111) 
    (setcar (nthcdr    0 A0110) A0005) 
    (setcar (nthcdr    4 A0109) A0109) 
    (setcar (nthcdr    3 A0109) A0048) 
    (setcar (nthcdr    2 A0109) A0046) 
    (setcar (nthcdr    1 A0109) A0042) 
    (setcar (nthcdr    0 A0109) A0032) ;;; af-syntax-when
    (setcar (nthcdr    4 A0108) A0109) 
    (setcar (nthcdr    3 A0108) A0048) 
    (setcar (nthcdr    2 A0108) A0046) 
    (setcar (nthcdr    1 A0108) A0042) 
    (setcar (nthcdr    0 A0108) A0013) ;;; af-syntax-exception-section
    (setcar (nthcdr    4 A0107) A0108) 
    (setcar (nthcdr    3 A0107) A0048) 
    (setcar (nthcdr    2 A0107) A0046) 
    (setcar (nthcdr    1 A0107) A0042) 
    (setcar (nthcdr    0 A0107) A0015) ;;; af-syntax-begin
    (setcar (nthcdr    1 A0106) A0002) 
    (setcar (nthcdr    0 A0106) A0107) 
    (setcar (nthcdr    3 A0105) A0071) 
    (setcar (nthcdr    2 A0105) A0048) 
    (setcar (nthcdr    1 A0105) A0042) 
    (setcar (nthcdr    0 A0105) A0088) ;;; af-syntax-package
    (setcar (nthcdr    2 A0104) A0022) 
    (setcar (nthcdr    1 A0104) A0105) 
    (setcar (nthcdr    0 A0104) A0018) 
    (setcar (nthcdr    0 A0103) A0069) 
    (setcar (nthcdr    0 A0102) A0007) 
    (setcar (nthcdr    1 A0100) A0002) 
    (setcar (nthcdr    0 A0100) A0101) ;;; af-syntax-while
    (setcar (nthcdr    2 A0099) A0026) 
    (setcar (nthcdr    1 A0099) A0100) 
    (setcar (nthcdr    0 A0099) A0055) 
    (setcar (nthcdr    2 A0098) A0031) 
    (setcar (nthcdr    1 A0098) A0029) 
    (setcar (nthcdr    0 A0098) A0027) 
    (setcar (nthcdr    2 A0097) A0048) 
    (setcar (nthcdr    1 A0097) A0042) 
    (setcar (nthcdr    0 A0097) A0074) 
    (setcar (nthcdr    1 A0095) A0002) 
    (setcar (nthcdr    0 A0095) A0096) ;;; af-syntax-task-type
    (setcar (nthcdr    1 A0093) A0002) 
    (setcar (nthcdr    0 A0093) A0094) ;;; af-syntax-task
    (setcar (nthcdr    1 A0091) A0002) 
    (setcar (nthcdr    0 A0091) A0092) ;;; af-syntax-protected-type
    (setcar (nthcdr    1 A0089) A0002) 
    (setcar (nthcdr    0 A0089) A0090) ;;; af-syntax-protected
    (setcar (nthcdr    2 A0087) A0026) 
    (setcar (nthcdr    1 A0087) A0083) 
    (setcar (nthcdr    0 A0087) A0088) ;;; af-syntax-package
    (setcar (nthcdr    4 A0086) A0095) 
    (setcar (nthcdr    3 A0086) A0093) 
    (setcar (nthcdr    2 A0086) A0091) 
    (setcar (nthcdr    1 A0086) A0089) 
    (setcar (nthcdr    0 A0086) A0087) 
    (setcar (nthcdr    1 A0085) A0002) 
    (setcar (nthcdr    0 A0085) A0074) 
    (setcar (nthcdr    1 A0083) A0002) 
    (setcar (nthcdr    0 A0083) A0084) ;;; af-syntax-generic
    (setcar (nthcdr    1 A0082) A0026) 
    (setcar (nthcdr    0 A0082) A0083) 
    (setcar (nthcdr    1 A0081) A0048) 
    (setcar (nthcdr    0 A0081) A0042) 
    (setcar (nthcdr    1 A0079) A0002) 
    (setcar (nthcdr    0 A0079) A0080) ;;; af-syntax-end-record
    (setcar (nthcdr    0 A0078) A0079) 
    (setcar (nthcdr    3 A0077) A0051) 
    (setcar (nthcdr    2 A0077) A0050) 
    (setcar (nthcdr    1 A0077) A0040) 
    (setcar (nthcdr    0 A0077) A0038) 
    (setcar (nthcdr    3 A0076) A0076) 
    (setcar (nthcdr    2 A0076) A0048) 
    (setcar (nthcdr    1 A0076) A0042) 
    (setcar (nthcdr    0 A0076) A0032) ;;; af-syntax-when
    (setcar (nthcdr    3 A0075) A0076) 
    (setcar (nthcdr    2 A0075) A0048) 
    (setcar (nthcdr    1 A0075) A0042) 
    (setcar (nthcdr    0 A0075) A0013) ;;; af-syntax-exception-section
    (setcar (nthcdr    3 A0074) A0075) 
    (setcar (nthcdr    2 A0074) A0048) 
    (setcar (nthcdr    1 A0074) A0042) 
    (setcar (nthcdr    0 A0074) A0015) ;;; af-syntax-begin
    (setcar (nthcdr    0 A0073) A0074) 
    (setcar (nthcdr    2 A0071) A0048) 
    (setcar (nthcdr    1 A0071) A0042) 
    (setcar (nthcdr    0 A0071) A0072) ;;; af-syntax-private-section
    (setcar (nthcdr    2 A0070) A0071) 
    (setcar (nthcdr    1 A0070) A0048) 
    (setcar (nthcdr    0 A0070) A0042) 
    (setcar (nthcdr    3 A0069) A0066) 
    (setcar (nthcdr    2 A0069) A0068) 
    (setcar (nthcdr    1 A0069) A0065) 
    (setcar (nthcdr    0 A0069) A0061) ;;; af-syntax-then
    (setcar (nthcdr    1 A0068) A0069) 
    (setcar (nthcdr    0 A0068) A0059) ;;; af-syntax-elsif
    (setcar (nthcdr    1 A0066) A0002) 
    (setcar (nthcdr    0 A0066) A0067) ;;; af-syntax-end-if
    (setcar (nthcdr    1 A0065) A0066) 
    (setcar (nthcdr    0 A0065) A0039) ;;; af-syntax-else
    (setcar (nthcdr    2 A0064) A0066) 
    (setcar (nthcdr    1 A0064) A0068) 
    (setcar (nthcdr    0 A0064) A0065) 
    (setcar (nthcdr    1 A0062) A0002) 
    (setcar (nthcdr    0 A0062) A0063) ;;; af-syntax-if
    (setcar (nthcdr    2 A0060) A0062) 
    (setcar (nthcdr    1 A0060) A0058) 
    (setcar (nthcdr    0 A0060) A0061) ;;; af-syntax-then
    (setcar (nthcdr    1 A0058) A0060) 
    (setcar (nthcdr    0 A0058) A0059) ;;; af-syntax-elsif
    (setcar (nthcdr    1 A0057) A0062) 
    (setcar (nthcdr    0 A0057) A0058) 
    (setcar (nthcdr    1 A0055) A0002) 
    (setcar (nthcdr    0 A0055) A0056) ;;; af-syntax-for
    (setcar (nthcdr    0 A0054) A0055) 
    (setcar (nthcdr    8 A0053) A0052) 
    (setcar (nthcdr    7 A0053) A0050) 
    (setcar (nthcdr    6 A0053) A0040) 
    (setcar (nthcdr    5 A0053) A0048) 
    (setcar (nthcdr    4 A0053) A0046) 
    (setcar (nthcdr    3 A0053) A0044) 
    (setcar (nthcdr    2 A0053) A0042) 
    (setcar (nthcdr    1 A0053) A0038) 
    (setcar (nthcdr    0 A0053) A0032) ;;; af-syntax-when
    (setcar (nthcdr    8 A0052) A0053) 
    (setcar (nthcdr    7 A0052) A0050) 
    (setcar (nthcdr    6 A0052) A0040) 
    (setcar (nthcdr    5 A0052) A0048) 
    (setcar (nthcdr    4 A0052) A0046) 
    (setcar (nthcdr    3 A0052) A0044) 
    (setcar (nthcdr    2 A0052) A0042) 
    (setcar (nthcdr    1 A0052) A0038) 
    (setcar (nthcdr    0 A0052) A0032) ;;; af-syntax-when
    (setcar (nthcdr    4 A0051) A0051) 
    (setcar (nthcdr    3 A0051) A0050) 
    (setcar (nthcdr    2 A0051) A0040) 
    (setcar (nthcdr    1 A0051) A0038) 
    (setcar (nthcdr    0 A0051) A0032) ;;; af-syntax-when
    (setcar (nthcdr    4 A0050) A0051) 
    (setcar (nthcdr    3 A0050) A0050) 
    (setcar (nthcdr    2 A0050) A0040) 
    (setcar (nthcdr    1 A0050) A0038) 
    (setcar (nthcdr    0 A0050) A0028) ;;; af-syntax-or-select
    (setcar (nthcdr    1 A0048) A0002) 
    (setcar (nthcdr    0 A0048) A0049) ;;; af-syntax-end-name
    (setcar (nthcdr    1 A0046) A0002) 
    (setcar (nthcdr    0 A0046) A0047) ;;; af-syntax-end-func
    (setcar (nthcdr    1 A0044) A0002) 
    (setcar (nthcdr    0 A0044) A0045) ;;; af-syntax-end-case
    (setcar (nthcdr    1 A0042) A0002) 
    (setcar (nthcdr    0 A0042) A0043) ;;; af-syntax-end
    (setcar (nthcdr    1 A0040) A0002) 
    (setcar (nthcdr    0 A0040) A0041) ;;; af-syntax-end-select
    (setcar (nthcdr    1 A0038) A0040) 
    (setcar (nthcdr    0 A0038) A0039) ;;; af-syntax-else
    (setcar (nthcdr    7 A0037) A0052) 
    (setcar (nthcdr    6 A0037) A0050) 
    (setcar (nthcdr    5 A0037) A0040) 
    (setcar (nthcdr    4 A0037) A0048) 
    (setcar (nthcdr    3 A0037) A0046) 
    (setcar (nthcdr    2 A0037) A0044) 
    (setcar (nthcdr    1 A0037) A0042) 
    (setcar (nthcdr    0 A0037) A0038) 
    (setcar (nthcdr    5 A0036) A0033) 
    (setcar (nthcdr    4 A0036) A0029) 
    (setcar (nthcdr    3 A0036) A0027) 
    (setcar (nthcdr    2 A0036) A0012) 
    (setcar (nthcdr    1 A0036) A0010) 
    (setcar (nthcdr    0 A0036) A0032) ;;; af-syntax-when
    (setcar (nthcdr    6 A0035) A0026) 
    (setcar (nthcdr    5 A0035) A0024) 
    (setcar (nthcdr    4 A0035) A0022) 
    (setcar (nthcdr    3 A0035) A0020) 
    (setcar (nthcdr    2 A0035) A0018) 
    (setcar (nthcdr    1 A0035) A0016) 
    (setcar (nthcdr    0 A0035) A0015) ;;; af-syntax-begin
    (setcar (nthcdr    1 A0034) A0035) 
    (setcar (nthcdr    0 A0034) A0013) ;;; af-syntax-exception-section
    (setcar (nthcdr    5 A0033) A0036) 
    (setcar (nthcdr    4 A0033) A0029) 
    (setcar (nthcdr    3 A0033) A0027) 
    (setcar (nthcdr    2 A0033) A0034) 
    (setcar (nthcdr    1 A0033) A0010) 
    (setcar (nthcdr    0 A0033) A0032) ;;; af-syntax-when
    (setcar (nthcdr    3 A0031) A0031) 
    (setcar (nthcdr    2 A0031) A0029) 
    (setcar (nthcdr    1 A0031) A0027) 
    (setcar (nthcdr    0 A0031) A0032) ;;; af-syntax-when
    (setcar (nthcdr    1 A0029) A0002) 
    (setcar (nthcdr    0 A0029) A0030) ;;; af-syntax-select
    (setcar (nthcdr    3 A0027) A0031) 
    (setcar (nthcdr    2 A0027) A0029) 
    (setcar (nthcdr    1 A0027) A0027) 
    (setcar (nthcdr    0 A0027) A0028) ;;; af-syntax-or-select
    (setcar (nthcdr    0 A0026) A0003) ;;; t
    (setcar (nthcdr    1 A0024) A0002) 
    (setcar (nthcdr    0 A0024) A0025) ;;; af-syntax-task-body
    (setcar (nthcdr    1 A0022) A0002) 
    (setcar (nthcdr    0 A0022) A0023) ;;; af-syntax-procedure
    (setcar (nthcdr    1 A0020) A0002) 
    (setcar (nthcdr    0 A0020) A0021) ;;; af-syntax-package-body
    (setcar (nthcdr    1 A0018) A0002) 
    (setcar (nthcdr    0 A0018) A0019) ;;; af-syntax-function
    (setcar (nthcdr    1 A0016) A0002) 
    (setcar (nthcdr    0 A0016) A0017) ;;; af-syntax-declare
    (setcar (nthcdr    6 A0014) A0026) 
    (setcar (nthcdr    5 A0014) A0024) 
    (setcar (nthcdr    4 A0014) A0022) 
    (setcar (nthcdr    3 A0014) A0020) 
    (setcar (nthcdr    2 A0014) A0018) 
    (setcar (nthcdr    1 A0014) A0016) 
    (setcar (nthcdr    0 A0014) A0015) ;;; af-syntax-begin
    (setcar (nthcdr    1 A0012) A0014) 
    (setcar (nthcdr    0 A0012) A0013) ;;; af-syntax-exception-section
    (setcar (nthcdr    1 A0010) A0002) 
    (setcar (nthcdr    0 A0010) A0011) ;;; af-syntax-case
    (setcar (nthcdr    4 A0009) A0033) 
    (setcar (nthcdr    3 A0009) A0029) 
    (setcar (nthcdr    2 A0009) A0027) 
    (setcar (nthcdr    1 A0009) A0012) 
    (setcar (nthcdr    0 A0009) A0010) 
    (setcar (nthcdr    1 A0007) A0002) 
    (setcar (nthcdr    0 A0007) A0008) ;;; af-syntax-end-loop
    (setcar (nthcdr    1 A0005) A0007) 
    (setcar (nthcdr    0 A0005) A0006) ;;; af-syntax-loop
    (setcar (nthcdr    0 A0004) A0005) 
    (setcar (nthcdr    0 A0002) A0003) ;;; t
    (setcar (nthcdr    0 A0001) A0002) 

    (aset af-syntax-while                 af-token-match-previous-offset A0001)
    (aset af-syntax-while                 af-token-match-next-offset     A0004)
    (aset af-syntax-when                  af-token-match-previous-offset A0009)
    (aset af-syntax-when                  af-token-match-next-offset     A0037)
    (aset af-syntax-use-of-for            af-token-match-previous-offset A0054)
    (aset af-syntax-use-of-for            af-token-match-next-offset     A0001)
    (aset af-syntax-then                  af-token-match-previous-offset A0057)
    (aset af-syntax-then                  af-token-match-next-offset     A0064)
    (aset af-syntax-task-type             af-token-match-previous-offset A0001)
    (aset af-syntax-task-type             af-token-match-next-offset     A0070)
    (aset af-syntax-task-body             af-token-match-previous-offset A0001)
    (aset af-syntax-task-body             af-token-match-next-offset     A0073)
    (aset af-syntax-task                  af-token-match-previous-offset A0001)
    (aset af-syntax-task                  af-token-match-next-offset     A0070)
    (aset af-syntax-select                af-token-match-previous-offset A0001)
    (aset af-syntax-select                af-token-match-next-offset     A0077)
    (aset af-syntax-record-at             af-token-match-previous-offset A0001)
    (aset af-syntax-record-at             af-token-match-next-offset     A0078)
    (aset af-syntax-record                af-token-match-previous-offset A0001)
    (aset af-syntax-record                af-token-match-next-offset     A0078)
    (aset af-syntax-protected-type        af-token-match-previous-offset A0001)
    (aset af-syntax-protected-type        af-token-match-next-offset     A0070)
    (aset af-syntax-protected-body        af-token-match-previous-offset A0001)
    (aset af-syntax-protected-body        af-token-match-next-offset     A0081)
    (aset af-syntax-protected             af-token-match-previous-offset A0001)
    (aset af-syntax-protected             af-token-match-next-offset     A0070)
    (aset af-syntax-procedure             af-token-match-previous-offset A0082)
    (aset af-syntax-procedure             af-token-match-next-offset     A0085)
    (aset af-syntax-private-section       af-token-match-previous-offset A0086)
    (aset af-syntax-private-section       af-token-match-next-offset     A0081)
    (aset af-syntax-package-body          af-token-match-previous-offset A0001)
    (aset af-syntax-package-body          af-token-match-next-offset     A0097)
    (aset af-syntax-package               af-token-match-previous-offset A0082)
    (aset af-syntax-package               af-token-match-next-offset     A0070)
    (aset af-syntax-or-select             af-token-match-previous-offset A0098)
    (aset af-syntax-or-select             af-token-match-next-offset     A0077)
    (aset af-syntax-loop                  af-token-match-previous-offset A0099)
    (aset af-syntax-loop                  af-token-match-next-offset     A0102)
    (aset af-syntax-if                    af-token-match-previous-offset A0001)
    (aset af-syntax-if                    af-token-match-next-offset     A0103)
    (aset af-syntax-generic               af-token-match-previous-offset A0001)
    (aset af-syntax-generic               af-token-match-next-offset     A0104)
    (aset af-syntax-function              af-token-match-previous-offset A0082)
    (aset af-syntax-function              af-token-match-next-offset     A0106)
    (aset af-syntax-for                   af-token-match-previous-offset A0001)
    (aset af-syntax-for                   af-token-match-next-offset     A0110)
    (aset af-syntax-exception-section     af-token-match-previous-offset A0113)
    (aset af-syntax-exception-section     af-token-match-next-offset     A0114)
    (aset af-syntax-end-select            af-token-match-previous-offset A0117)
    (aset af-syntax-end-select            af-token-match-next-offset     A0001)
    (aset af-syntax-end-record            af-token-match-previous-offset A0119)
    (aset af-syntax-end-record            af-token-match-next-offset     A0001)
    (aset af-syntax-end-func              af-token-match-previous-offset A0124)
    (aset af-syntax-end-func              af-token-match-next-offset     A0001)
    (aset af-syntax-end-name              af-token-match-previous-offset A0128)
    (aset af-syntax-end-name              af-token-match-next-offset     A0001)
    (aset af-syntax-end-loop              af-token-match-previous-offset A0139)
    (aset af-syntax-end-loop              af-token-match-next-offset     A0001)
    (aset af-syntax-end-case              af-token-match-previous-offset A0141)
    (aset af-syntax-end-case              af-token-match-next-offset     A0001)
    (aset af-syntax-end-if                af-token-match-previous-offset A0143)
    (aset af-syntax-end-if                af-token-match-next-offset     A0001)
    (aset af-syntax-end                   af-token-match-previous-offset A0128)
    (aset af-syntax-end                   af-token-match-next-offset     A0001)
    (aset af-syntax-elsif                 af-token-match-previous-offset A0145)
    (aset af-syntax-elsif                 af-token-match-next-offset     A0103)
    (aset af-syntax-else                  af-token-match-previous-offset A0146)
    (aset af-syntax-else                  af-token-match-next-offset     A0147)
    (aset af-syntax-do                    af-token-match-previous-offset A0148)
    (aset af-syntax-do                    af-token-match-next-offset     A0081)
    (aset af-syntax-declare               af-token-match-previous-offset A0001)
    (aset af-syntax-declare               af-token-match-next-offset     A0073)
    (aset af-syntax-case                  af-token-match-previous-offset A0001)
    (aset af-syntax-case                  af-token-match-next-offset     A0149)
    (aset af-syntax-begin                 af-token-match-previous-offset A0151)
    (aset af-syntax-begin                 af-token-match-next-offset     A0152)
    (aset af-syntax-accept                af-token-match-previous-offset A0001)
    (aset af-syntax-accept                af-token-match-next-offset     A0154)
    (aset af-syntax-right-square          af-token-match-previous-offset A0156)
    (aset af-syntax-right-square          af-token-match-next-offset     A0001)
    (aset af-syntax-left-square           af-token-match-previous-offset A0001)
    (aset af-syntax-left-square           af-token-match-next-offset     A0159)
    (aset af-syntax-right-paren           af-token-match-previous-offset A0162)
    (aset af-syntax-right-paren           af-token-match-next-offset     A0001)
    (aset af-syntax-left-paren            af-token-match-previous-offset A0001)
    (aset af-syntax-left-paren            af-token-match-next-offset     A0165)
    (aset af-syntax-right-label           af-token-match-previous-offset A0168)
    (aset af-syntax-right-label           af-token-match-next-offset     A0001)
    (aset af-syntax-left-label            af-token-match-previous-offset A0001)
    (aset af-syntax-left-label            af-token-match-next-offset     A0171)
    (aset af-syntax-right-curly           af-token-match-previous-offset A0174)
    (aset af-syntax-right-curly           af-token-match-next-offset     A0001)
    (aset af-syntax-left-curly            af-token-match-previous-offset A0001)
    (aset af-syntax-left-curly            af-token-match-next-offset     A0177)
    (aset af-syntax-end-of-line           af-token-match-previous-offset A0180)
    (aset af-syntax-end-of-line           af-token-match-next-offset     A0180)
  )

  );; if (not) nil
);; if nil

);; let ((fixup-list nil)...)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; AF-STATE-1 - What do do with each possible character as we go forward
;;;; trying to lex a line of Ada code.  This Finite State Machine lexes
;;;; Ada tokens in the forward direction.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defconst af-state-1 (make-vector 256 'af-bad-ada-character)
  ;;--Tells us what to do when we are about to try to scan a line and turn
  ;;  it into a series of Ada lexical elements.
  )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; AF-STATE-1 support routines.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-defsubst af-state-keyword (keywords)

  ;;--Called when we have a letter and we don't yet know if we have a keyword.
  ;;  The argument is a list of cons'ed keyword name strings and keyword flag
  ;;  values.

  (skip-syntax-forward "w_")
  (if (setq keywords
            (assoc (upcase (buffer-substring lex-result-point (point)))
                   keywords))
      (progn
        (af-debug (car keywords))
        (cdr keywords)
        )

    (af-debug "Identifier")
    af-syntax-identifier
    )

  );; af-state-keyword


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun af-bad-ada-character ()

  ;;--Called when we are attempting to lex a line of Ada source text and we
  ;;  encounter a character which is not legally part of an Ada source unit."

  (error
   "Character '%c'/0%o is not allowed (outside a comment) in Ada source text."
   (following-char) (following-char))
  );; af-bad-ada-character


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun af-bad-number-at (what)

  ;;--Put out an error message about a number of invalid form.

  (error "Badly formed number (%s) at line %d."
         what
         (af-what-line-number)
         )
  );; af-bad-number-at


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun af-bad-string-at (place)

  ;;--Put out an error message about a string of invalid form.

  (error "Badly formed string (missing ending quote) at line %d."
         (af-what-line-number place)
         )
  );; af-bad-string-at


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(let ((edebug-all-defuns nil))

(defun af-what-line-number (&optional position)
  "Return the current line number (in the buffer) of point.  Optionally we
move to the argument position first."
  (save-restriction
    (widen)
    (save-excursion
      (if position (goto-char position))
      (beginning-of-line)
      (1+ (count-lines 1 (point)))
      )
    )
  );; af-what-line-number

);; edebug-all-defuns

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;-Set the contents of af-state-1.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(let ((pairs
       (list

;;;-We don't regard end-of-line as "true" whitespace.  We use it to notice
;;; when we have finished scanning a line of text.

        (cons "\f\r\v"                  ; End-Of-Line characters (sometimes)
              (lambda ()
                (forward-char 1)
                (af-debug "End of line.")
                af-syntax-end-of-line))

        (cons "\n"                      ; End-Of-Line characters
              (lambda ()
                (forward-char 1)
                (af-debug "End of line.")
                af-syntax-end-of-line))

;;;-Null may really be end-of-buffer.

        (cons "\000"                    ; Null
              (lambda ()
                (if (eobp)
                    (progn
                      (af-debug "End of line/buffer.")
                      af-syntax-end-of-line
                      )
                  (af-bad-ada-character)
                  )
                ))

;;;-This is what we do for identifiers that cannot be keywords.

        (cons "hjkqvyzHJKQVYZ"
              (lambda ()
                (skip-syntax-forward "w_")
                (af-debug "Identifier")
                af-syntax-identifier))

;;;-This is what we do for identifiers that might be keywords.

        (cons "aA" (lambda () (af-state-keyword af-keywords-a)))
        (cons "bB" (lambda () (af-state-keyword af-keywords-b)))
        (cons "cC" (lambda () (af-state-keyword af-keywords-c)))
        (cons "dD" (lambda () (af-state-keyword af-keywords-d)))
        (cons "eE" (lambda () (af-state-keyword af-keywords-e)))
        (cons "fF" (lambda () (af-state-keyword af-keywords-f)))
        (cons "gG" (lambda () (af-state-keyword af-keywords-g)))
        (cons "iI" (lambda () (af-state-keyword af-keywords-i)))
        (cons "lL" (lambda () (af-state-keyword af-keywords-l)))
        (cons "mM" (lambda () (af-state-keyword af-keywords-m)))
        (cons "nN" (lambda () (af-state-keyword af-keywords-n)))
        (cons "oO" (lambda () (af-state-keyword af-keywords-o)))
        (cons "pP" (lambda () (af-state-keyword af-keywords-p)))
        (cons "rR" (lambda () (af-state-keyword af-keywords-r)))
        (cons "sS" (lambda () (af-state-keyword af-keywords-s)))
        (cons "tT" (lambda () (af-state-keyword af-keywords-t)))
        (cons "uU" (lambda () (af-state-keyword af-keywords-u)))
        (cons "wW" (lambda () (af-state-keyword af-keywords-w)))
        (cons "xX" (lambda () (af-state-keyword af-keywords-x)))

;;;-This is what we do for the various compound operators and symbols.

        (cons "=" (lambda ()
                    (forward-char 1)
                    (cond
                     ((= ?> (following-char))
                      (forward-char 1)
                      (af-debug "Arrow =>")
                      af-syntax-arrow
                      )
                     (t
                      (af-debug "Equal =")
                      af-syntax-equal
                      ))))
                        
        (cons "'" (lambda ()
                    (forward-char 1)
                    (cond
                     ((= ?' (char-after (1+ (point))))
                      (forward-char 2)
                      (af-debug "Character")
                      af-syntax-character
                      )
                     (t
                      (af-debug "Apostrophe '")
                      af-syntax-apostrophe
                      ))))

        (cons "<" (lambda ()
                    (forward-char 1)
                    (cond
                     ((= ?= (following-char))
                      (forward-char 1)
                      (af-debug "Less-Than-Or_Equal <=")
                      af-syntax-less-than-or-equal
                      )
                     ((= ?< (following-char))
                      (forward-char 1)
                      (af-debug "Left-Label <<")
                      af-syntax-left-label
                      )
                     ((= ?> (following-char))
                      (forward-char 1)
                      (af-debug "Box <>")
                      af-syntax-box
                      )   
                     (t
                      (af-debug "Less-Than <")
                      af-syntax-less-than
                      ))))

        (cons ">" (lambda ()
                    (forward-char 1)
                    (cond
                     ((= ?= (following-char))
                      (forward-char 1)
                      (af-debug "Greater-Than-Or-Equal >=")
                      af-syntax-greater-than-or-equal
                      )
                     ((= ?> (following-char))
                      (forward-char 1)
                      (af-debug "Right-Label >>")
                      af-syntax-right-label
                      )
                     (t
                      (af-debug "Greater-Than >")
                      af-syntax-greater-than
                      ))))

        (cons ":" (lambda ()
                    (forward-char 1)
                    (cond
                     ((= ?= (following-char))
                      (forward-char 1)
                      (af-debug "Assignment :=")
                      af-syntax-assignment
                      )
                     (t
                      (af-debug "Colon :")
                      af-syntax-colon
                      ))))
              
        (cons "/" (lambda ()
                    (forward-char 1)
                    (cond
                     ((= ?= (following-char))
                      (forward-char 1)
                      (af-debug "Not-Equal /=")
                      af-syntax-not-equal
                      )
                     (t
                      (af-debug "Divide /")
                      af-syntax-divide
                      ))))
                        
        (cons "*" (lambda ()
                    (forward-char 1)
                    (cond
                     ((= ?* (following-char))
                      (forward-char 1)
                      (af-debug "Double-Star **")
                      af-syntax-double-star
                      )
                     (t
                      (af-debug "Multiply *")
                      af-syntax-multiply
                      ))))
                        
        (cons "." (lambda ()
                    (forward-char 1)
                    (cond
                     ((= ?. (following-char))
                      (forward-char 1)
                      (af-debug "Double-Dot ..")
                      af-syntax-double-dot
                      )
                     (t
                      (af-debug "Dot .")
                      af-syntax-dot
                      ))))
                        
        (cons "-" (lambda ()
                    (forward-char 1)
                    (cond
                     ((= ?- (following-char))
                      (af-end-of-line)
                      (if (not (eobp)) (forward-char 1))
                      (af-debug "Comment")
                      af-syntax-comment
                      )
                     (t
                      (af-debug "Minus -")
                      af-syntax-minus
                      ))))

;;;-These are all single character symbols that do not participate in
;;; multi-character symbols.

        (cons "(" (lambda ()
                    (forward-char 1)
                    (af-debug "Left-Paren (")
                    af-syntax-left-paren
                    ))

        (cons ")" (lambda ()
                    (forward-char 1)
                    (af-debug "Right-Paren )")
                    af-syntax-right-paren
                    ))

        (cons "[" (lambda ()
                    (forward-char 1)
                    (af-debug "Left-Square [")
                    af-syntax-left-square
                    ))

        (cons "]" (lambda ()
                    (forward-char 1)
                    (af-debug "Right-Square ]")
                    af-syntax-right-square
                    ))

        (cons "{" (lambda ()
                    (forward-char 1)
                    (af-debug "Left-Curly {")
                    af-syntax-left-curly
                    ))

        (cons "}" (lambda ()
                    (forward-char 1)
                    (af-debug "Right-Curly }")
                    af-syntax-right-curly
                    ))

        (cons "&" (lambda ()
                    (forward-char 1)
                    (af-debug "Ampersand &")
                    af-syntax-ampersand
                    ))

        (cons "," (lambda ()
                    (forward-char 1)
                    (af-debug "Comma ,")
                    af-syntax-comma
                    ))	

        (cons "+" (lambda ()
                    (forward-char 1)
                    (af-debug "Plus +")
                    af-syntax-plus
                    ))

        (cons ";" (lambda ()
                    (forward-char 1)
                    (af-debug "Semicolon ;")
                    af-syntax-semicolon
                    ))

        (cons "|!" (lambda ()
                     (forward-char 1)
                     (af-debug "Vertical-Bar |")
                     af-syntax-vertical-bar
                     ))

;;;-This is what we do for numbers.  We semi-lex them.  We don't try to catch
;;; all errors.  We mainly want to correctly skip over the number regardless of
;;; its form.

        (cons "0123456789"
              (lambda ()
                (let ((hash nil)        ; Not a based number.
                      (dot nil)         ; No decimal point.
                      (exp nil))        ; No exponent.
                  (forward-char 1)      ; Skip initial digit.
                  (while                ; Scan until we read end-of-number.
                      (cond            

                       ;;-- Skip all digits and underscores

                       ((and (<= ?0 (following-char))
                             (>= ?9 (following-char))))

                       ((= ?_ (following-char)))

                       ((and (numberp hash)
                             (<= ?A (upcase (following-char)))
                             (>= ?F (upcase (following-char)))))

                       ;;--See if we have a decimal point in our number.
                       ;;  There should never be more than one dot in a
                       ;;  number.

                       ((= ?. (following-char))
                        (if (or dot
                                exp)
                            (af-bad-number-at "invalid .")
                          )
                        (forward-char 1)
                        (setq dot t)
                        )

                       ;;--See if we have a based number.

                       ((or (= ?# (following-char))
                            (= ?: (following-char)))
                        (forward-char 1)
                        (if hash
                            (progn
                              (if (/= hash (preceding-char))
                                  (af-bad-number-at
                                   (message "invalid #/:, doesn't match %c"
                                            hash))
                                )
                              (setq hash t
                                    dot  t)
                              )
                          
                          (if (or exp
                                  dot)
                              (af-bad-number-at
                               "invalid #/:, comes after . or Exp.")
                            )
                          (setq hash (preceding-char))
                          )
                        )


                       ;;--We have an exponent.

                       ((= ?E (upcase (following-char)))
                        (if exp (af-bad-number-at "invalid exponent"))
                        (forward-char 1)
                        (if (or (= ?- (following-char))
                                (= ?+ (following-char)))
                            (forward-char 1)
                          )
                        (setq exp t
                              dot t
                              hash t)
                        )
                       );; cond
                    (forward-char 1)
                    );; while
                  );; let
                (af-debug "Number")
                af-syntax-number
                ))

;;;-This is what we do for strings.

        (cons "\""
              (lambda ()
                (forward-char 1)        ; Skip initial quote.
                (skip-chars-forward "^\"\f\n\r\v")
                (while                  ; Scan until we read end-of-string.
                    (cond

                     ;;--We have "" if the string is to continue.

                     ((= ?\" (following-char))
                      (forward-char 1)
                      (= ?\" (following-char))
                      )

                     ;;--We have end-of-line in a string.

                     ((af-bad-string-at (point)))

                     );; cond
                  (forward-char 1)      ; Skip 2nd quote of 2 quotes.
                  (skip-chars-forward "^\"\f\n\r\v")
                  );; while
                (af-debug "String")
                af-syntax-string
                ))

        (cons "%"
              (lambda ()
                (forward-char 1)        ; Skip initial quote.
                (skip-chars-forward "^%\f\n\r\v")
                (while                  ; Scan until we read end-of-string.
                    (cond

                     ;;--We have %% if the string is to continue.

                     ((= ?% (following-char))
                      (forward-char 1)
                      (= ?% (following-char))
                      )

                     ;;--We have end-of-line in a string.

                     ((af-bad-string-at (point)))

                     );; cond
                  (forward-char 1)      ; Skip 2nd quote of 2 quotes.
                  (skip-chars-forward "^%\f\n\r\v")
                  );; while
                (af-debug "%String%")
                af-syntax-string
                ))
        ))
      head)

;;;-Loop over all pairs in the list.

  (while pairs
    (setq head  (car pairs))
    (setq pairs (cdr pairs))

;;;-Each pair is a string and a function.  The function is the one to call
;;; for each letter in the string.

    (let ((chars (car head))
          (count (length (car head)))
          (func  (cdr head))
          char)

;;;-Loop over all string characters and set the af-state-1 to func.

      (while (> count 0)
        (setq count (1- count))
        (setq char (aref chars count))
        (aset af-state-1 char func)
        )
      )
    )
  )

;;;-All characters in the 0200..0377 range are "identifier" characters.

(let ((char ?\200))
  (while (< char ?\400)
    (aset af-state-1 char (aref af-state-1 ?Z))
    (setq char (1+ char))
    )
  )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; AF replacement functions.  We need special treatment of the end-of-line
;;;; concept.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun af-beginning-of-line (&optional arg)
  "Called to do (beginning-of-line) but paying attention to af-unix-eol.
Move point to beginning of current line.
With argument ARG not nil or 1, move forward ARG - 1 lines first.
If scan reaches end of buffer, stop there without error."
  (if af-unix-eol
      (beginning-of-line arg)
    (cond ((eq arg nil)
           (skip-chars-backward "^\f\n\r\v"))
          ((= arg 1)
           (skip-chars-backward "^\f\n\r\v"))
          ((> arg 1)
           (while (> arg 1)
             (skip-chars-forward "^\f\n\r\v")
             (if (not (eobp)) (forward-char 1))
             (setq arg (1- arg))
             )
           )
          (t
           (while (<= arg 0)
             (skip-chars-backward "^\f\n\r\v")
             (if (not (bobp)) (backward-char 1))
             (setq arg (1+ arg))
             )
           (skip-chars-backward "^\f\n\r\v")
           )
          )
    )
  );; af-beginning-of-line

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun af-end-of-line (&optional arg)
  "Called to do (end-of-line) but paying attention to af-unix-eol.
Move point to end of current line.
With argument ARG not nil or 1, move forward ARG - 1 lines first.
If scan reaches end of buffer, stop there without error."
  (if af-unix-eol
      (end-of-line arg)
    (cond ((eq arg nil)
           (skip-chars-forward "^\f\n\r\v"))
          ((>= arg 1)
           (while (> arg 1)
             (skip-chars-forward "^\f\n\r\v")
             (if (not (eobp)) (forward-char 1))
             (setq arg (1- arg))
             )
           (skip-chars-forward "^\f\n\r\v")
           )
          (t
           (while (<= arg 0)
             (skip-chars-backward "^\f\n\r\v")
             (if (not (bobp)) (backward-char 1))
             (setq arg (1+ arg))
             )
           )
          )
    )
  );; af-end-of-line

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun af-current-column (&optional here current)
  "Called to do (current-column) but paying attention to af-unix-eol."
  (if af-unix-eol
      (current-column)

    (setq here (point)
          current (current-column))
    (af-beginning-of-line)
    (setq current (- current (current-column)))
    (goto-char here)
    current
    )
  );; af-current-column

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun af-forward-to-indentation (arg)
  "Called to do (forward-to-indentation arg) but paying attention to
af-unix-eol.  Go forward (positive) or backward (negative) arg lines, 0
means don't move.  Then advance forward past all spaces and tabs."
  (if af-unix-eol
      (forward-to-indentation arg)
    (progn
      (while (> arg 0)
        (af-end-of-line)
        (if (not (eobp)) (forward-char 1))
        (setq arg (1- arg))
        )
      (af-beginning-of-line)
      (while (< arg 0)
        (if (not (bobp)) (backward-char 1))
        (af-beginning-of-line)
        (setq arg (1+ arg))
        )
      (skip-chars-forward " \t")
      )
    )
  );; af-forward-to-indentation


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; af-indent-line - Basic indentation function for Ada mode.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(af-defsubst af-indent-line-guts ()

  ;;--Remember our starting point.  This also implicitly allows us to remember
  ;;  our starting column.

  (setq starting-point  (point-marker))

  ;;--Go to the beginning of the current line.  Remember this point.
  ;;  Calculate our preferred indentation for this line of code
  ;;  (which may be blank or may be a comment not starting in
  ;;  column 1).

  (af-beginning-of-line)
  (setq bol         (point)
        indentation (af-calculate-indentation))

  ;;--Skip all initial whitespace on this line.  We only
  ;;  reindent the line (change the buffer) if the indentation
  ;;  needs to change.  If the indentation does not change then
  ;;  we want to insert a tab at our starting point unless it is
  ;;  at the start of a properly indented line.

  (goto-char bol)
  (skip-chars-forward " \t")
  (if (/= indentation (af-current-column))

      ;;--We are changing the indentation.  Delete the initial
      ;;  whitespace.
      ;;--If our starting point was in the whitespace that was
      ;;  deleted then we want dot to be at the end of the new
      ;;  whitespace when we insert it.
      ;;--If our starting point was not in the whitespace that was
      ;;  deleted then we want to return dot to wherever it was
      ;;  after we insert the new whitespace.

      (progn
        (delete-region bol (point))
        (if (= (point) starting-point)
            (indent-to indentation indentation)
          (progn
            (indent-to indentation indentation)
            (goto-char starting-point)
            )
          )
        )

    ;;--We are not changing the indentation.  Go back to the
    ;;  starting point and insert a tab if the starting point is
    ;;  beyond the starting point of the line.

    (if (< (point) starting-point)
        (progn
          (goto-char starting-point)
          (insert-tab)
          )
      )
    )

  ;;--Reset our marker.  (The ELisp manual recommends doing this to
  ;;  markers you aren't planning on reusing even if they in variables
  ;;  which are about to be unbound causing the marker to become
  ;;  gc-able.)

  (set-marker starting-point nil)

  );; af-indent-line-guts

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defconst af-max-specpdl-size    25000
  "*max-specpdl-size used during af-indent-line processing.")
(defconst af-max-lisp-eval-depth  5000
  "*max-lisp-eval-depth used during af-indent-line processing.")

(if af-byte-compiling

(defun af-indent-line ()
  "Reindent the current line.  The buffer is not changed by this reindentation
if the line is already at the correct indentation.  If point is at the start
of the line, or anywhere in the leading whitespace of the line, then point is
left at the end of that whitespace.  If point is past the first nonwhitespace
character on the line, and if the line indentation has not been changed, then
tab to the next tab stop within the line.  (The arguments are ignored and
exist solely because creating optional arguments is faster than creating
local variables.)"
  (interactive)

  (let (starting-point
        bol
        indentation
        blink-point
        (max-specpdl-size    (max max-specpdl-size    af-max-specpdl-size))
        (max-lisp-eval-depth (max max-lisp-eval-depth af-max-lisp-eval-depth)))

    (condition-case
        error-data

        (af-indent-line-guts)

      ;;--If we get any kind of signal/error coming out, then push the
      ;;  starting point as the topmost mark (so user can get back there
      ;;  easily) and then resignal the error.

      (error (message "We are at the error.")
             (sit-for 1 0 t)
             (message "The point is %d." starting-point)
             (sit-for 1 0 t)
             (push-mark (marker-position starting-point) t t)
             (set-marker starting-point nil)
             (signal (car error-data) (cdr error-data))
             )
      )

  ;;--Blink the cursor if that is desired and possible.

    (and blink-point
         af-blink-matching-group-begin
         (af-blink-matching-group-begin blink-point)
         )
    )

  );; af-indent-line

(defun af-indent-line ()
  "Debug version only."
  (interactive)

  (let (starting-point
        bol
        indentation
        blink-point
        (max-specpdl-size    (max max-specpdl-size    af-max-specpdl-size))
        (max-lisp-eval-depth (max max-lisp-eval-depth af-max-lisp-eval-depth)))

    (af-indent-line-guts)

  ;;--Blink the cursor if that is desired and possible.

    (and blink-point
         af-blink-matching-group-begin
         (af-blink-matching-group-begin blink-point)
         )
    )

  );; af-indent-line
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; af-calculate-indentation - Basic indentation calculation function for Ada
;;;; mode.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun af-calculate-indentation ( &optional previous-token
                                            previous-data
                                            previous-symbol
                                            current-token
                                            current-data
                                            current-symbol
                                            position
                                            indentation
                                            starting-point )
  "We are at the beginning of some line.  We want to calculate the appropriate
indentation to use for that line.  The result should be in some 0..N range.
It will only be negative is the various indentation control variables contains
some semi-bizarre values (eg. large negatives).  (The optional arguments are
actually 'fast' local function variables and not read at all.)"

  (setq starting-point (point))         ; Our initial position at some bol.
  (catch 'return

    ;;--If the line starts with -- at column 0 then we use an indentation of
    ;;  0.

    (if (looking-at "--")
        (throw 'return (setq indentation 0))
      )

    ;;--We have to do things by lexing around.  First we clear the token
    ;;  list because we don't know what's in it or what's changed in the
    ;;  buffer since it was last used.  Then we look forward from (point)
    ;;  and see what we see on this line.  The line could be a blank/comment
    ;;  line.

    (clear-af-token-list starting-point)
    (setq current-token (initial-indent-af-token))

    ;;--Our first check is to see if we are in the middle of a complex token.
    ;;  e.g. Between the END and the IF of END IF.  We have the forward
    ;;  simple token.  We check the token.  If the location of the token
    ;;  changes (which also implies that the symbol of the location changed,
    ;;  but a change of symbol would not always imply a position change, so
    ;;  we check the point for change rather than the symbol for change)
    ;;  then we know we are in the middle of a complex token and we just
    ;;  indent relative to the start of that token (which is the new point
    ;;  of the token).

    (setq position       (af-token-point current-token)
          current-token  (af-token-check-both current-token))
    (if (/= position (af-token-point current-token))
        (progn
          (goto-char (af-token-point current-token))
          (setq indentation (+ af-indent-middle-complex
                               (af-current-column)))
          (throw 'return t)
          )
      )
    (setq current-data   (af-token-data current-token)
          previous-token (af-token-check-previous
                          (previous-af-token current-token))
          previous-data  (af-token-data previous-token))

    ;;--If the current token is some kind of group continue or end, then we
    ;;  want to indent relative to the group-begin.  This token is at bol,
    ;;  so record his special indentation as a starting indentation.

    (if (af-testbit current-data (logior af-flag-group-begin-continue
                                         af-flag-group-continue
                                         af-flag-group-end))
        (progn
          (setq indentation (af-token-special-indentation current-token))

          ;;--Search for a previous group token.  We will only fail (get a
          ;;  nil) if this token is a begin-continue and there turns out to
          ;;  be no group-begin.

          (if (af-calc-backup-inside-group current-token current-data)
              (progn

                ;;--We have a previous group token.  Get its column and,
                ;;  if it is at bol, subtract its special indentation.

                (goto-char (af-token-point current-token))
                (setq blink-point (point))
                (if (af-testbit (af-token-flags current-token)
                                af-token-flag-bol)
                    (setq indentation
                          (+ indentation
                             (- (af-token-special-indentation current-token))
                             (af-current-column)))
                  (setq indentation (+ indentation
                                       (af-current-column)))
                  )
                (throw 'return t)
                )
            )

          ;;--We had a begin-continue with no previous begin.  Restore the
          ;;  old previous/current-token position and continue on.

          (goto-char starting-point)     ; Our initial position at some bol.
          (clear-af-token-list starting-point)
          (setq current-token (initial-indent-af-token))
          (setq current-token  (af-token-check-both current-token)
                current-data   (af-token-data current-token)
                previous-token (af-token-check-previous
                                (previous-af-token current-token))
                previous-data  (af-token-data previous-token))
          )
      )

    ;;--If the current-token is the special af-syntax-name-at symbol then
    ;;  we have "ID['ID] at x range x..x;", which may follow a clause like
    ;;  "record at mod x", and so it may not be preceded by an
    ;;  af-flag-any-sep (see below), but we know it should be treated as
    ;;  if it were so.

    (setq current-symbol (af-token-symbol current-token))
    (if (eq current-symbol af-syntax-name-at)
        (progn
          (setq indentation   (af-token-special-indentation current-token))

          ;;--Backup to (a) a separator at eol that separates two previous
          ;;  elements or (b) any begin/begin-continue/continue token.
          ;;  We don't care which.  In both cases we want to indent exactly
          ;;  the same as the token that follows what we find unless what
          ;;  we find is a RECORD followed by an AT, in that case we indent
          ;;  relative to the RECORD.

          (if (af-testbit previous-data af-flag-element-separator)
              (af-calc-backup-element previous-token previous-data 'eol-only)
            (af-calc-backup-element current-token current-data 'eol-only)
            )
          (if (af-testbit current-data af-flag-group-begin)

              ;;--We have a RECORD AT, we are indenting the first clause
              ;;  of a record rep clause.  Indent relative to the RECORD.

              (progn
                (goto-char (af-token-point current-token))
                (setq indentation
                      (+ indentation
                         af-indent-new-element
                         (af-current-column)))
                (throw 'return t)
                )

            ;;--We are indenting normally, get the next token and indent
            ;;  as it is indented.

            (setq current-token (next-af-token current-token))
            (goto-char (af-token-point current-token))
            (setq indentation
                  (+ indentation
                     (- (af-token-special-indentation current-token))
                     (af-current-column)))
            (throw 'return t)
            )
          )
      )

    ;;--If the current-token is an IS then we need to run backwards until
    ;;  we find a begin/continue/begin-continue to align with or something
    ;;  else like a TYPE/SUBTYPE to indent within as a continue.

    (if (eq current-symbol af-syntax-is)
        (progn
          (setq indentation (af-token-special-indentation current-token))
          (if (af-calc-backup-for-is current-token)

              ;;--We backed up and found a begin style token.  We want to
              ;;  align with it.

              (progn
                (goto-char (af-token-point current-token))
                (setq indentation (+ indentation
                                     (af-current-column)))
                (throw 'return t)
                )

            ;;--We backed up and found a non-begin style token.  We want to
            ;;  indent within it as a continuation.

            (throw 'return (af-calc-continuation-on-current-token))
            )
          )
      )

    ;;--If the current-token is an => then we need to run backwards until
    ;;  we find a WHEN/','/'(' to tell us what to do.

    (if (eq current-symbol af-syntax-arrow)
        (progn
          (setq indentation (af-token-special-indentation current-token))
          (if (af-calc-backup-for-arrow current-token)

              ;;--We backed up and found a WHEN token.  We want to
              ;;  align inside it.

              (progn
                (goto-char (af-token-point current-token))
                (setq indentation (+ indentation
                                     af-indent-new-element
                                     (af-current-column)))
                (throw 'return t)
                )

            ;;--We backed up and found a '('/',' token.  We want to
            ;;  indent as a new element with the element that follows it.

            (setq current-token  (next-af-token current-token)
                  current-symbol (af-token-symbol current-token))
            (throw 'return (af-calc-continuation-on-current-token))
            )
          )
      )

    ;;--If the previous token was a group-begin, then we are indenting the
    ;;  first element of a new group.  Indent relative to that begin token.

    (setq previous-symbol (af-token-symbol previous-token))
    (if (af-testbit previous-data af-flag-group-begin)
        (cond

         ;;--If we are at the beginning of the buffer.  Indent 0.

         ((eq previous-symbol af-syntax-end-of-line)
          (setq indentation 0)
          (throw 'return t)
          )

         ;;--Indent within a normal begin token.  () are a special case.

         ((eq previous-symbol af-syntax-left-paren)
          (if (numberp af-indent-after-left-paren)
              (progn
                (goto-char (af-token-point previous-token))
                (setq indentation
                      (+ (af-token-special-indentation current-token)
                         af-indent-after-left-paren
                         (af-current-column)))
                (throw 'return t)
                )
            )
          )

         ;;--Other begins indent their first element this much.

         (t
          (goto-char (af-token-point previous-token))
          (setq indentation
                (+ (af-token-special-indentation current-token)
                   af-indent-new-element
                   (af-current-column)))
          (throw 'return t)
          )
         );; cond
      )

    ;;--((This test must come *after* the test for af-flag-group-begin.))
    ;;--If the current token is something that could be part of a "tight"
    ;;  expression group, then we want to indent relative to the start of
    ;;  that tight expression.  Tight groups are things like:
    ;;
    ;;  tight ::= ( ID | "str" | 'x' ) 
    ;;          | tight . ( ID | "str" | 'x' | ALL )
    ;;          | tight ' ( ID | (...) )
    ;;          | tight (...)

    (if (and (af-testbit current-data  af-flag-tight-expression)
             (af-testbit previous-data af-flag-tight-expression))
        (progn
          (af-calc-find-tight-expression-start previous-token previous-data)
          (goto-char (af-token-point current-token))
          (setq indentation (+ af-indent-tight-expression
                               (af-current-column)))
          (throw 'return t)
          )
      )

    ;;--If the previous token is a continue, then we are indenting the first
    ;;  element of a new group.  If it has no force-indent and does not
    ;;  occur at bol, then search backwards looking for a begin or a
    ;;  continue with one of those attributes.

    (if (af-testbit previous-data (logior af-flag-group-begin-continue
                                          af-flag-group-continue))
        (progn
          (setq indentation   (af-token-special-indentation current-token))
          (if (or (af-testbit previous-data af-flag-force-indent)
                  (af-testbit (af-token-flags previous-token)
                              af-token-flag-bol))
              (progn
                
                ;;--We have a continue with force-indent or at bol.  Use
                ;;  it as it it were a begin.

                (goto-char (af-token-point previous-token))
                (setq indentation (+ indentation
                                     af-indent-new-element
                                     (af-current-column)))
                (throw 'return t)
                )
            )

          ;;--We have a continue, but a weak one.  Go find a strong one.
          ;;  Remember where we started in case this is a begin-continue
          ;;  with no previous real begin.

          (setq position      (af-token-point previous-token))
          (if (af-calc-backup-inside-group previous-token
                                           (af-token-data previous-token))
              ;;--Found matching group token to indent with.
              (goto-char (af-token-point current-token))
            ;;--Found outer group, revert to our begin-continue
            ;;  and use it as a begin.
            (goto-char position)
            )
          (setq indentation (+ indentation
                               af-indent-new-element
                               (af-current-column)))
          (throw 'return t)
          )
      )

    ;;--If the previous-token is an IS then we need to run backwards until
    ;;  we find a begin/continue/begin-continue to indent within or something
    ;;  else like a TYPE/SUBTYPE to indent within as a continue.

    (if (eq previous-symbol af-syntax-is)
        (if (af-calc-backup-for-is previous-token)

            ;;--We have a begin style token.  We are a first new element.

            (progn
              (goto-char (af-token-point current-token))
              (setq indentation (+ af-indent-new-element
                                   (af-current-column)))
              (throw 'return t)
              )

          ;;--We have a non-begin style token.  We are a continuation.
          ;;  (Except for certain special cases.)

          (setq indentation 0)
          (throw 'return (af-calc-continuation-on-current-token))
          )
      )

    ;;--If the previous-token is an => then we need to run backwards until
    ;;  we find a WHEN/'('/',' to indent within.

    (if (eq previous-symbol af-syntax-arrow)
        (progn
          (setq indentation (af-token-special-indentation current-token))
          (if (af-calc-backup-for-arrow previous-token)

              ;;--We backed up and found a WHEN token.  We want to
              ;;  align inside it.

              (progn
                (goto-char (af-token-point current-token))
                (setq indentation (+ indentation
                                     af-indent-new-element
                                     (af-current-column)))
                (throw 'return t)
                )

            ;;--We backed up and found a '('/',' token.  We want to
            ;;  indent as a continuation with the aggregate element that
            ;;  follows it.

            (setq current-token  (next-af-token current-token)
                  current-symbol (af-token-symbol current-token))
            (throw 'return (af-calc-continuation-on-current-token))
            )
          )
      )

    ;;--If the previous token is an and-then or an or-else, and if we are
    ;;  treating those specially, then search  back for a previous
    ;;  and-then/or-else and indent as if they were separators.  That is,
    ;;  we treat this continuation-of-an-expression situation as if it were
    ;;  a sequence of group elements instead.

    (if (and af-control-and-then-or-else-special
             (or (eq previous-symbol af-syntax-and-then)
                 (eq previous-symbol af-syntax-or-else)))
        (progn

          ;;--Backup to (a) a separator at eol that separates two previous
          ;;  elements or (b) any begin/begin-continue/continue token.
          ;;  We don't care which.  In both cases we want to indent exactly
          ;;  the same as the token that follows what we find.
          ;;  We also do this if we find a comma.

          (setq indentation   (af-token-special-indentation current-token))
          (if (or (af-calc-backup-and-then-or-else previous-token)
                  (eq (af-token-symbol current-token) af-syntax-comma))

              ;;--We backed up and found a previous and-then/or-else at
              ;;  eol, or, we found a begin/continue/begin-continue.
              ;;  We want to indent exactly the same as the token that
              ;;  follows what we found.

              (progn
                (setq current-token (next-af-token current-token))
                (goto-char (af-token-point current-token))
                (setq indentation (af-current-column))
                (throw 'return t)
                )

            ;;--We backed up and found a separator (probably a ";").
            ;;  We want to treat our expression as a normal continuation.

            (setq current-token  (next-af-token current-token)
                  current-symbol (af-token-symbol current-token))
            (throw 'return (af-calc-continuation-on-current-token t))
            )
          )
      )

    ;;--If the previous token is a pseudo-begin then we want to indent
    ;;  relative to it's position.

    (if (af-testbit previous-data af-flag-after-pseudo-begin)
        (progn
          (goto-char (af-token-point previous-token))
          (setq indentation (+ (af-token-special-indentation current-token)
                               af-indent-continuation
                               (af-current-column)))
          (throw 'return t)
          )
      )

    ;;--If the previous token is an element-separator, then we are indenting
    ;;  the first element of a new group.  And, we are doing it in the
    ;;  middle of a group.  We want to find the start of the previous element
    ;;  in the group.  Otherwise, we are doing a continuation line.

    (if (af-testbit previous-data af-flag-element-separator)
        (progn

          ;;--Backup to (a) a separator at eol that separates two previous
          ;;  elements or (b) any begin/begin-continue/continue token.
          ;;  We don't care which.  If we find a begin-continue that has
          ;;  an is-group-start flag, we indent relative to the
          ;;  begin-continue.  In all other cases we want to indent exactly
          ;;  the same as the token that follows what we find.

          (setq indentation   (af-token-special-indentation current-token))
          (af-calc-backup-element previous-token previous-data 'eol-only)
          (if (af-testbit current-data af-flag-is-group-start)

              ;;--Indent within the begin-continue.

              (progn
                (goto-char (af-token-point current-token))
                (setq indentation
                      (+ indentation
                         af-indent-new-element
                         (af-current-column)))
                (throw 'return t)
                )

            ;;--Indent the same as the token after the begin/separator.

            (setq current-token (next-af-token current-token))
            (goto-char (af-token-point current-token))
            (setq indentation
                  (+ indentation
                     (- (af-token-special-indentation current-token))
                     (af-current-column)))
            (throw 'return t)
            )
          )
      )

    ;;--We are doing a continuation line.  Find the start of the current
    ;;  element and indent within that.

    (setq indentation    (af-token-special-indentation current-token)
          current-symbol (af-token-symbol current-token))
    (if (af-calc-backup-element
         previous-token
         previous-data
         (if (af-testbit current-data (logior af-flag-element-separator
                                              af-flag-after-pseudo-begin))
             'eol-only
           'not-eol-only))

        ;;--We found a separator, indent relative to the token that follows
        ;;  the separator.

        (progn
          (setq current-token  (next-af-token current-token)
                current-symbol (af-token-symbol current-token))
          (af-calc-continuation-on-current-token)
          )

      ;;--We found some kind of begin, indent relative to the token that
      ;;  follows it unless we have a pseudo-begin, in which case we indent
      ;;  the same as the token that follows it.

      (if (or (af-testbit current-data af-flag-after-pseudo-begin)
              (eq current-symbol af-syntax-record)
              (eq current-symbol af-syntax-use))

          ;;--Psedo-begin, indent same as follower token.

          (progn
            (setq current-token  (next-af-token current-token))
            (goto-char (af-token-point current-token))
            (setq indentation (+ indentation
                                 (af-current-column)))
            )

        ;;--Any other begin, take continuation on next token.

        (setq current-token  (next-af-token current-token)
              current-symbol (af-token-symbol current-token))
        (af-calc-continuation-on-current-token)
        )
      )
    );; catch 'return

  ;;--Restore our starting point so our callers don't have to do it
  ;;  themselves.

  (goto-char starting-point)
  indentation

  );; af-calculate-indentation


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun af-calc-continuation-on-current-token (&optional and-then-or-else)
  ;;--Called when current-token is something which we are to be a continuation
  ;;  of.  We calculate the new indentation relative to that element.  We
  ;;  expect the indentation varaible to already be initialized with some
  ;;  slight BOL offset value.  We add to that value.
  ;;
  ;;  Returns the indentation amount.

  ;;--See if the current-token is a BOL token, apply any special indentation
  ;;  rule that applies to that token.

  (if (af-testbit (af-token-flags current-token) af-token-flag-bol)
      (setq indentation (- indentation
                           (af-token-special-indentation current-token)))
    )

  ;;--We want to do special things for RETURN keywords.

  (if (eq current-symbol af-syntax-return)
      ;;--Special rule, we want to indent "within" a RETURN statement as if
      ;;  the RETURN was some kind of group-begin.
      (setq current-token (next-af-token current-token))
    )

  ;;--Indent relative to the token.

  (goto-char (af-token-point current-token))
  (cond (and-then-or-else
         (setq indentation (+ indentation
                              (af-current-column))))

        ((eq current-symbol af-syntax-vertical-bar)
         (setq indentation (+ indentation
                              (af-current-column))))

        ((and (eq current-symbol af-syntax-record)
              (eq previous-symbol af-syntax-is))
         (setq indentation (+ af-indent-record-after-is
                              (af-current-column))))

        (t
         (setq indentation (+ indentation
                              af-indent-continuation
                              (af-current-column))))
        )

  );; af-calc-continuation-on-next-token


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun af-calc-find-tight-expression-start (start-token start-data)
  ;;--Called when current-token and previous-token are both tight-expression
  ;;  tokens.  Called with previous-token and previous-data.  We search
  ;;  backwards looking for the start of the tight part of the enclosing
  ;;  expression.  current-token and current-data contains the last token
  ;;  of the tight expression when we return.

  (while (progn

           ;;--If we find a group-end, then find the group-begin.  We are in
           ;;  an expression so there should be no chance of a loose
           ;;  begin-continue.

           (if (af-testbit start-data af-flag-group-end)
               (if (af-calc-backup-group start-token start-data)

                   ;;--We found the begin that matches the end.  Now get the
                   ;;  token in front of that.

                   (setq start-token (af-token-check-previous
                                      (previous-af-token current-token))
                         start-data  (af-token-data start-token))

                 ;;--This will not happen in valid source.  It means that we
                 ;;  got something like PROCEDURE...END in the expression?

                 (goto-char (af-token-point current-token))
                 (error
                  "Unmatched begin-continue found in tight expression at %d."
                  (af-what-line-number starting-point))
                 )

             ;;--We are still in a tight expression, get the next token.

             (setq current-token start-token
                   current-data  start-data
                   start-token   (af-token-check-previous
                                  (previous-af-token start-token))
                   start-data    (af-token-data start-token))
             )

           ;;--Loop until we find something that is not a tight-expression.
           ;;  We also stop when we find a group-begin, eg. "("

           (and (af-testbit start-data af-flag-tight-expression)
                (af-zerobit start-data af-flag-group-begin))
           )
    )

  );; af-calc-find-tight-expression-start


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun af-calc-backup-element (start-token
                               start-data
                               eol-only
                               &optional last-is-semicolon ; Local variable.
                               )
  ;;--Called with start-token and start-data pointing at a separator, or,
  ;;  at some token which is not a group begin/continue/begin-continue.
  ;;  Backs up until it discovers:
  ;;
  ;;    - (eq eol-only 'eol-only)       - any element separator, other than the
  ;;                                      first token seen
  ;;    - (eq eol-only 'not-eol-only)   - any element-separator at eol, other
  ;;                                      than the first token
  ;;    - any group begin/begin-continue/continue element
  ;;
  ;;  Returns t if we stopped on a separator and nil if we stopped on a group
  ;;  token.  current-token will be the separator or group token that stopped
  ;;  the search.  Skips an initial element-separator on the assumption that
  ;;  it is merely the end of the element we are backing over.  

  ;;--If we have a group-end, then skip backwards over that group.  If we
  ;;  find an enclosing group as part of that skip, then stop.  Otherwise
  ;;  loop searching for what we want.

  (if (or (and (af-zerobit start-data af-flag-group-end)
               (setq current-token start-token))
          (af-calc-backup-group start-token start-data))
      (catch 'return
        (setq last-is-semicolon nil)
        (while t

          ;;--Get the previous token.

          (setq start-token   current-token
                current-token (af-token-check-previous
                               (previous-af-token start-token))
                current-data  (af-token-data current-token))

          ;;--An optimization, the only tokens we care about are these, do a
          ;;  quick one-operation check.

          (if (af-testbit current-data (logior af-flag-element-separator
                                               af-flag-group-begin
                                               af-flag-group-begin-continue
                                               af-flag-group-continue
                                               af-flag-group-end
                                               af-flag-is-group-start
                                               af-flag-simple-is-start
                                               af-flag-is-going-backwards
                                               af-flag-simple-is-follower
                                               af-flag-after-pseudo-begin))
              (cond

               ;;--If the token is a separator, and the next token was at
               ;;  bol, then the separator is a eol and we stop.

               ((af-testbit current-data af-flag-element-separator)
                (if (or (eq eol-only 'not-eol-only)
                        (af-testbit (af-token-flags start-token)
                                    af-token-flag-bol))
                    (throw 'return t)
                  ))

               ;;--If the token is any kind of group begin/continue then we
               ;;  stop.  If the token is an is-group-start token, and if
               ;;  we didn't "just now" see an IS as we went by, then this
               ;;  token isn't really a begin-continue, its more of an end.

               ((af-testbit current-data (logior af-flag-group-begin
                                                 af-flag-group-begin-continue
                                                 af-flag-group-continue))
                (if (or 
                     ;;--Ordinary begin/continue.
                     (af-zerobit current-data af-flag-is-group-start)
                     ;;--Non-ordinary, but followed by IS.
                     (eq last-is-semicolon af-syntax-is)
                     ;;--Non-ordinary, may have a GENERIC in front of
                     ;;  it, so must go look.
                     (and last-is-semicolon
                          (af-zerobit (aref last-is-semicolon
                                            af-token-data-offset)
                                      af-flag-simple-is-follower)
                          ;;--Didn't find GENERIC, found other begin.
                          (not (af-calc-backup-group current-token
                                                     current-data))))
                    (throw 'return nil)
                  )
                )

               ;;--If the token is a group end then skip that group and stop
               ;;  if that takes us to an enclosing group begin/continue.

               ((af-testbit current-data af-flag-group-end)
                (if (not (af-calc-backup-group current-token current-data))
                    (throw 'return nil)
                  )
                )

               ;;--This must come before the IS tests below.
               ;;--If we are doing a simple continuation search, and if this
               ;;  token is a pseudo-begin, then stop the search here.

               ((and (eq eol-only 'not-eol-only)
                     (af-testbit current-data af-flag-after-pseudo-begin))
                (throw 'return nil)
                )

               ;;--If the token is something that changes the interpretation
               ;;  of PROCEDURE/FUNCTION as we go by, record that.

               ((af-testbit current-data (logior af-flag-is-group-start
                                                 af-flag-simple-is-start
                                                 af-flag-is-going-backwards
                                                 af-flag-simple-is-follower))
                (setq start-data (af-token-symbol current-token))
                (if (not (and last-is-semicolon
                              (eq start-data af-syntax-is)
                              (af-testbit (aref last-is-semicolon
                                                af-token-data-offset)
                                          af-flag-simple-is-follower)))
                    (setq last-is-semicolon start-data)
                  (if (af-testbit current-data
                                  (logior af-flag-is-group-start
                                          af-flag-simple-is-start))
                      (setq last-is-semicolon af-syntax-semicolon)
                    )

                  )
                )
               );; cond
            );; if
          );; while
        );; catch
    );; if

  );; af-calc-backup-element


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun af-calc-backup-and-then-or-else (start-token &optional symbol)
  ;;--Called with start-token and start-data pointing at an and-then or an
  ;;  or-else.  Backs up until it discovers:
  ;;
  ;;    - any element separator
  ;;    - any previous and-then or or-else occurring at bol
  ;;    - any group begin/begin-continue/continue element
  ;;
  ;;  Returns t if we stopped on an and-then/or-else or a group token, and nil
  ;;  if we stopped on a normal separator.  current-token and current-data
  ;;  will be the token that stopped the search.

  ;;--If we have a group-end, then skip backwards over that group.  If we
  ;;  find an enclosing group as part of that skip, then stop.  Otherwise
  ;;  loop searching for what we want.

  (catch 'return
    (while t

      ;;--Get the previous token.

      (setq current-token (af-token-check-previous
                           (previous-af-token start-token))
            current-data  (af-token-data current-token))
      (cond

       ;;--If the token is not a group token, see if it is an and-then/or-else.

       ((or (eq (setq symbol (af-token-symbol current-token))
                af-syntax-and-then)
            (eq symbol af-syntax-or-else))
        (if (af-testbit (af-token-flags start-token) af-token-flag-bol)
            (throw 'return t)
          )
        )
       ;;--If the token is any kind of group token, handle it here.

       ((af-testbit current-data (logior af-flag-element-separator
                                         af-flag-group-begin
                                         af-flag-group-begin-continue
                                         af-flag-group-continue
                                         af-flag-group-end))
        (cond

         ;;--If the token is a separator, we stop.

         ((af-testbit current-data af-flag-element-separator)
          (throw 'return nil))

         ;;--If the token is any kind of group begin/continue then we
         ;;  stop.

         ((af-testbit current-data (logior af-flag-group-begin
                                           af-flag-group-begin-continue
                                           af-flag-group-continue))
          (throw 'return nil))

         ;;--If the token is a group end then skip that group and stop
         ;;  if that takes us to an enclosing group begin/continue.

         ((af-testbit current-data af-flag-group-end)
          (if (not (af-calc-backup-group current-token current-data))
              (throw 'return nil)
            )
          )
         );; cond
        )
       );; cond
      (setq start-token current-token)
      );; while
    );; catch

  );; af-calc-backup-and-then-or-else


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun af-calc-backup-group (start-token
                             start-data
                             &optional  ; Local variables
                             match-previous
                             last-is-semicolon
                             end-symbol
                             end-data
                             end-position)
  ;;--Called to skip backwards over a group.  The start-token is the
  ;;  group-end of the group.  We run backwards until we find a matching
  ;;  group-begin.
  ;;
  ;;  We return t if we find, and stop at, the matching group-begin.  
  ;;  current-token and current-data will be the group-begin in this case.
  ;;
  ;;  We return nil in the case where the group was actually begun by a
  ;;  group-begin-continue token, and we didn't figure this out until we ran
  ;;  into a begin/continue belonging to an enclosing group.  current-token
  ;;  and current-data will be the enclosing begin/continue/begin-continue
  ;;  in this case.

  (setq last-is-semicolon nil
        end-symbol   (af-token-symbol start-token)
        end-data     start-data
        end-position (af-token-point  start-token))

  (setq match-previous (af-token-match-previous start-token))
  (setq current-token start-token)
  (catch 'return
    (while t

      ;;--Search backwards for any begin/continue/end.

      (setq current-token (af-token-check-previous
                           (previous-af-token current-token))
            current-data  (af-token-data current-token))
      (cond

       ;;--If the token is a group-end, then skip the group forward.

       ((and (af-testbit current-data af-flag-group-end)
             (if (af-calc-backup-group current-token current-data)
                 t
               (setq last-is-semicolon af-syntax-is)
               nil)))

       ;;--If the token is some kind of group token and...

       ((and (af-testbit current-data (logior af-flag-group-begin
                                              af-flag-group-continue
                                              af-flag-group-begin-continue))

             ;;--...and not just a PROCEDURE/FUNCTION that needs skipped.

             (or 
              ;;--Ordinary begin/continue.
              (af-zerobit current-data af-flag-is-group-start)
              ;;--Non-ordinary, but followed by IS.
              (eq last-is-semicolon af-syntax-is)
              ;;--Non-ordinary, may have a GENERIC in front of
              ;;  it, so must go look.
              (and last-is-semicolon
                   (af-zerobit (aref last-is-semicolon
                                     af-token-data-offset)
                               af-flag-simple-is-follower)
                   ;;--Didn't find GENERIC, found other begin.
                   (not (af-calc-backup-group current-token
                                              current-data)))))
        (cond

         ;;--We have a begin/continue/begin-continue token.
         ;;  If this applies to our group, and our match-previous list
         ;;  continues, then continue the loop.
         
         ((consp (setq end-data
                       (af-token-match-check match-previous current-token)))
          (setq match-previous  end-data
                end-symbol      (af-token-symbol current-token)
                end-data        current-data
                end-position    (af-token-point  current-token))
          )

         ;;--If we just ran off the end of our match list, then we are done,
         ;;  return t.

         ((= 0 end-data)
          (throw 'return t))

         ;;--We may have a begin/continue/begin-continue that
         ;;  encloses the group we were traversing.  If our last
         ;;  end-token was a begin-continue, then we return a nil to
         ;; indicate that our search went too far before stopping.

         ((= 1 end-data)
          (throw 'return nil)
          )

         ;;--We have a begin/continue/begin-continue that does not
         ;; match our group, so this must be a syntax error in the
         ;; source.

         (t
          (goto-char (af-token-point current-token))
          (error
           "Token %s at line %d is preceded by this %s token.  Expected (%s)."
           (aref end-symbol af-token-name-offset)
           (af-what-line-number end-position)
           (af-token-string-name current-token)
           (mapconcat (lambda (assoc)
                        (if (eq t (car assoc))
                            ""
                          (aref (car assoc) af-token-name-offset)))
                      match-previous " "))
          )
         );; cond
        )

       ;;--Remember IS's and ;'s in case we run into a FUNC/PROC and wonder
       ;;  what to do with it.

       ((af-testbit current-data (logior af-flag-is-group-start
                                         af-flag-simple-is-start
                                         af-flag-is-going-backwards
                                         af-flag-simple-is-follower))
        (setq start-data (af-token-symbol current-token))
        (if (not (and last-is-semicolon
                      (eq start-data af-syntax-is)
                      (af-testbit (aref last-is-semicolon
                                        af-token-data-offset)
                                  af-flag-simple-is-follower)))
            (setq last-is-semicolon start-data)
          (if (af-testbit current-data
                          (logior af-flag-is-group-start
                                  af-flag-simple-is-start))
              (setq last-is-semicolon af-syntax-semicolon)
            )
          
          )
        )

       ;;--Otherwise continue the loop.

       (t)

       );; cond
      );; while t
    );; catch 'return
 
  );; af-calc-backup-group


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun af-calc-backup-inside-group (start-token
                                    start-data
                                    &optional ; Local variables
                                    match-previous
                                    last-is-semicolon
                                    force-indent
                                    end-symbol
                                    end-data
                                    end-position
                                    short-circuit)
  ;;--Called to skip backwards over the next segment of a group.  The
  ;;  start-token and start-data is some begin-continue/continue of the group.
  ;;  We run backwards until we find a matching:
  ;;
  ;;    - begin
  ;;    - begin-continue at bol
  ;;    - continue at bol
  ;;
  ;;  Or, if the start-token does not have af-flag-force-indent, a matching:
  ;;
  ;;    - begin-continue with force-indent
  ;;    - continue with force-indent
  ;;
  ;;  We return t if we find, and stop at, some matching point mentioned above.
  ;;  current-token and current-data will be the matching token in this case.
  ;;
  ;;  We return nil in the case where we started with a begin-continue
  ;;  and we found a begin/continue/begin-continue belonging to an enclosing
  ;;  group.  current-token and current-data will be the enclosing
  ;;  begin/continue/begin-continue in this case.

  (setq last-is-semicolon nil
        force-indent   (if (af-testbit start-data af-flag-force-indent)
                           0
                         af-flag-force-indent)
        end-symbol     (af-token-symbol current-token)
        end-data       start-data
        end-position   (af-token-point  current-token)
        match-previous (af-token-match-previous start-token)
        short-circuit  (if match-previous
                           (car (car match-previous))
                         nil)
        short-circuit  (cond ((eq short-circuit af-syntax-generic)
                              af-flag-not-generic-formal)
                             ((eq short-circuit af-syntax-declare)
                              af-flag-not-declare-begin)
                             (t
                              0)
                             )
        current-token start-token)
  (catch 'return
    (while t

      ;;--Search backwards for any begin/continue/end.

      (setq current-token (af-token-check-previous
                           (previous-af-token current-token))
            current-data  (af-token-data current-token))
      (cond

       ;;--If the token is a short-circuit token then stop now and return a
       ;;  nil.

       ((af-testbit current-data short-circuit)
        ;; GEB (goto-char (af-token-point current-token))
        ;; GEB (message "stopping here")
        ;; GEB (sit-for 4)
        (throw 'return nil))

       ;;--If the token is a group-end, then skip the group forward.

       ((and (af-testbit current-data af-flag-group-end)
             (if (af-calc-backup-group current-token current-data)
                 t
               (setq last-is-semicolon af-syntax-is)
               nil)))

       ;;--If the token is some kind of group token and...

       ((and (af-testbit current-data (logior af-flag-group-begin
                                              af-flag-group-continue
                                              af-flag-group-begin-continue))

             ;;--...and not just a PROCEDURE/FUNCTION that needs skipped.

             (or 
              ;;--Ordinary begin/continue.
              (af-zerobit current-data af-flag-is-group-start)
              ;;--Non-ordinary, but followed by IS.
              (eq last-is-semicolon af-syntax-is)
              ;;--Non-ordinary, may have a GENERIC in front of
              ;;  it, so must go look.
              (and last-is-semicolon
                   (af-zerobit (aref last-is-semicolon
                                     af-token-data-offset)
                               af-flag-simple-is-follower)
                   ;;--Didn't find GENERIC, found other begin.
                   (not (af-calc-backup-group current-token
                                              current-data)))))
        (cond

         ;;--We have a begin/continue/begin-continue token.
         ;;  If this applies to our group then:
         ;;    - if it is a begin, we are done, return t
         ;;    - if it is at bol, or if it has force-indent, return t
         ;;    - otherwise, continue the loop

         ((consp (setq end-data
                       (af-token-match-check match-previous current-token)))
          (if (or (af-testbit current-data
                              (logior af-flag-group-begin
                                      af-flag-force-indent))
                  (af-testbit (af-token-flags current-token)
                              af-token-flag-bol))
              (throw 'return t)
            (setq match-previous end-data
                  end-symbol     (af-token-symbol current-token)
                  end-data       current-data
                  end-position   (af-token-point  current-token))
            )
          )

         ;;--If we just ran off the end of our match list, then we are done,
         ;;  return t.

         ((= 0 end-data)
          (throw 'return t))

         ;;--We have a begin/continue/begin-continue that
         ;;  encloses the group we were traversing.  If our last
         ;;  end-token was a begin-continue, then we return a nil to
         ;; indicate that our search went too far before stopping.

         ((= 1 end-data)
          (throw 'return nil))

         ;;--We have a begin/continue/begin-continue that does not
         ;; match our group, so this must be a syntax error in the
         ;; source.

         (t
          (goto-char (af-token-point current-token))
          (error
           "Token %s at line %d is preceded by this %s token.  Expected (%s)."
           (aref end-symbol af-token-name-offset)
           (af-what-line-number end-position)
           (af-token-string-name current-token)
           (mapconcat (lambda (assoc)
                        (if (eq t (car assoc))
                            ""
                          (aref (car assoc) af-token-name-offset)))
                      match-previous " "))
          )
         );; cond
        )

       ;;--Remember IS's and ;'s in case we run into a FUNC/PROC and wonder
       ;;  what to do with it.

       ((af-testbit current-data (logior af-flag-is-group-start
                                         af-flag-simple-is-start
                                         af-flag-is-going-backwards
                                         af-flag-simple-is-follower))
        (setq start-data (af-token-symbol current-token))
        (if (not (and last-is-semicolon
                      (eq start-data af-syntax-is)
                      (af-testbit (aref last-is-semicolon
                                        af-token-data-offset)
                                  af-flag-simple-is-follower)))
            (setq last-is-semicolon start-data)
          (if (af-testbit current-data
                          (logior af-flag-is-group-start
                                  af-flag-simple-is-start))
              (setq last-is-semicolon af-syntax-semicolon)
            )
          
          )
        )

       ;;--Otherwise continue the loop.

       (t)

       );; cond
      );; while t
    );; catch 'return
 
  );; af-calc-backup-inside-group


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun af-calc-backup-for-is (start-token
                              &optional ; Local variables
                              follower
                              starting-point)
  ;;--Called with start-token as an IS.  We back up until we find something
  ;;  that tells us what the IS is for.  For some things, the IS is a kind
  ;;  of continue token, eg. PROCEDURE x IS.  For other things, the IS is just
  ;;  another token in the stream, eg. TYPE x IS.
  ;;
  ;;  Returns t if we find a begin style token.  Returns nil if we find
  ;;  an ordinary token.  current-token and current-data will reflect the
  ;;  token found.

  ;;--Remember whatever follows the IS token.  Then loop reading tokens.

  (setq current-token start-token
        follower (af-token-symbol (next-af-token start-token))
        starting-point (af-token-point start-token))
  (catch 'return
    (while
        (cond

         ;;--The next token is a PROCEDURE/TYPE or whatever.  Our search
         ;;  is over.  Figure out what result to return.

         ((af-testbit (setq current-token (af-token-check-previous
                                           (previous-af-token current-token))
                            current-data  (af-token-data current-token))
                      (logior af-flag-simple-is-start
                              af-flag-is-group-start))
          (throw 'return
                 (not (or (af-testbit current-data af-flag-simple-is-start)
                          (eq follower af-syntax-abstract)
                          (eq follower af-syntax-separate)
                          (eq follower af-syntax-new))))
          )

         ;;--Found a group end.  Skip the group.  It should only be a ()
         ;;  group so it should be impossible to get a begin-continue
         ;;  problem.
         ;;
         ;;--Found a group separator.  Either there is a missing end,
         ;;  eg. a missing ), or there is a missing begin, eg. PROCEDURE
         ;;  or TYPE.

         ((or (and (af-testbit current-data af-flag-group-end)
                   (not (af-calc-backup-group current-token current-data)))
              (af-testbit current-data af-flag-element-separator))
          (goto-char (af-token-point current-token))
          (error
           (concat "Found '%s' instead of the start of declaration"
                   " containing the IS at line %d.")
           (af-token-string-name current-token)
           (af-what-line-number starting-point))
          )

         ;;--Otherwise just loop.

         (t)

         );; cond
      );; while
    );; catch

  );; af-calc-backup-for-is


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun af-calc-backup-for-arrow (start-token
                                 &optional ; Local variables
                                 starting-point
                                 symbol)
  ;;--Called with start-token as an =>.  We back up until we find something
  ;;  that tells us what the => is for.  eg. "WHEN x =>" or "(A =>" or 
  ;;  "(..., A =>".
  ;;
  ;;  Returns t if we find a WHEN token.  Returns nil if we find an aggregate
  ;;  token ('(' or ',').  current-token and current-data will reflect the
  ;;  token found.

  ;;--Loop reading tokens.

  (setq current-token start-token
        starting-point (af-token-point start-token))
  (catch 'return
    (while
        (cond

         ;;--These tokens stop the search.

         ((eq (setq current-token (af-token-check-previous
                                   (previous-af-token current-token))
                    current-data  (af-token-data current-token)
                    symbol        (af-token-symbol current-token))
              af-syntax-when)
          (throw 'return t))

         ((or (eq symbol af-syntax-left-paren)
              (eq symbol af-syntax-comma))
          (throw 'return nil))

         ((eq symbol af-syntax-vertical-bar)) ; Just skip these.

         ;;--Found a group end.  Skip the group.  It should only be a ()
         ;;  group so it should be impossible to get a begin-continue
         ;;  problem.
         ;;
         ;;--Found a group separator.  Either there is a missing end,
         ;;  eg. a missing ), or there is a missing begin, eg. PROCEDURE
         ;;  or TYPE.

         ((or (and (af-testbit current-data af-flag-group-end)
                   (not (af-calc-backup-group current-token current-data)))
              (af-testbit current-data af-flag-element-separator))
          (goto-char (af-token-point current-token))
          (error
           (concat "Found '%s' instead of WHEN/'('/','"
                   " containing the => at line %d.")
           (af-token-string-name current-token)
           (af-what-line-number starting-point))
          )

         ;;--Otherwise just loop.

         (t)

         );; cond
      );; while
    );; catch
 
  );; af-calc-backup-for-arrow


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun af-blink-matching-group-begin (begin &optional bol eol)
  ;;--if blink-matching-paren is turned on we blink matching group tokens
  ;;  when we indent them.

  (save-excursion
    (goto-char begin)
    (if (pos-visible-in-window-p begin)
        (if (numberp af-blink-matching-group-begin)
            (sit-for af-blink-matching-group-begin)
          (if (consp af-blink-matching-group-begin)
              (apply 'sit-for af-blink-matching-group-begin)
            (sit-for 1)
            )
          )

      (setq bol (progn (af-beginning-of-line) (point)))
      (setq eol (progn (af-end-of-line) (point)))
      (cond ((= bol begin)
             (message "^%s" (buffer-substring begin eol)))
            ((= eol begin)
             (message "%s^" (buffer-substring bol begin)))
            (t
             (message "%s^%s"
                      (buffer-substring bol (1- begin))
                      (buffer-substring begin eol)))
            )
      )
    )

  );; af-blink-matching-group-begin


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; Profiling Tools
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(if (or t  ;;; t turns off profiling, nil turns it on.
        af-byte-compiling)
    nil

(message "***DOING PROFILING***")
(sit-for 1 0 t)

(defun do-profile ()
  (interactive)
  (let ((blink-matching-paren nil))
    (af-indent-line)
    (profile-report)
    )
  );; do-profile

(defconst profile-list                        ; List of funcs to profile
  (list 'af-indent-line
;;;        'af-token-match-check
        'af-lex-one
        'af-token-check-previous
        'af-token-check-next
        'af-state-keyword
;;;        'af-testbit
;;;        'af-zerobit
        'af-calculate-indentation
        'af-calc-find-tight-expression-start
        'af-calc-backup-element
        'af-calc-backup-and-then-or-else
        'af-calc-backup-group
        'af-calc-backup-inside-group
        'af-calc-backup-for-is
        'af-calc-backup-for-arrow))

(defvar profile-data    nil)                  ; All profile data
(defvar profile-current nil)                  ; Current profile function
(defvar profile-count   0)                    ; Tells us when to GC

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun profile-time (start stop)

  ;;--Pass in two (current-time) values.  Returns the difference in floating
  ;;  seconds.

  (- (+ (* (float (nth 0 stop)) 65536.0)
        (float (nth 1 stop))
        (/ (float (nth 2 stop)) 1000000.0))
     (+ (* (float (nth 0 start)) 65536.0)
        (float (nth 1 start))
        (/ (float (nth 2 start)) 1000000.0)))

  );; profile-time

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun profile-setup ( &optional funcs func data var)

  ;;--Set up functions to be profiled.

  (setq funcs           profile-list)
  (setq profile-current nil)
  (setq profile-data    nil)
  (setq profile-count   0)
  (while funcs

    ;;--Get next function.  Create the data structure for that function.
    ;;  Put the data on our profile-list.

    (setq func (car funcs))
    (setq data (list (symbol-function func)   ; 0 Function to call
                     func                     ; 1 Function to call
                     nil                      ; 2 Time of last call
                     0                        ; 3 Number of calls
                     0.0                      ; 4 Total time spent
                     ))
    (setq profile-data (nconc profile-data (list data)))
    (setq var (eval (read (format "(defconst %S_dataX data)" func))))

    ;;--Create the new function that will be replacing the function we are
    ;;  to profile.

    (fset
     func
     (eval
      (`
       (lambda (&rest rest)
         (let ((this-time             (current-time))
               (outer-profile-current (car profile-current))
               the-result)

           ;;--Compute the elapsed time used so far in the call that is 
           ;;  outside of this call.

           (if outer-profile-current
               (setcar (nthcdr 4 outer-profile-current)
                       (+  (nth 4 outer-profile-current)
                           (profile-time (nth 2 outer-profile-current)
                                         this-time)))
             )

           ;;--Add this function to our call-in-progress list.

           (setq profile-current (cons (, var) profile-current))

           ;;--Increment the call count on our function.  Record the current
           ;;  time on this call.  Force a GC so we don't charge our routines
           ;;  for that overhead.

           (setcar (nthcdr 3 (, var)) (1+ (car (nthcdr 3 (, var)))))
           (if (> 0 (setq profile-count (1- profile-count)))
               (progn
                 (setq gc-cons-threshold (max gc-cons-threshold 10000000))
                 (garbage-collect)
                 (setq profile-count 500)
                 )
             )
           (setcar (nthcdr 2 (, var)) (current-time))

           ;;--Call the function we are trying to profile.

           (setq the-result (apply (nth 0 (, var)) rest))

           ;;--Get the current time, and compute the elapsed time for this
           ;;  function.

           (setq this-time (current-time))
           (setcar (nthcdr 4 (, var))
                   (+  (car (nthcdr 4 (, var)))
                       (profile-time (car (nthcdr 2 (, var)))
                                     this-time)))

           ;;--Remove us from the call-in-progress list and return our result.

           (setq profile-current (cdr profile-current))

           ;;--Resume the outer call by recording the time in his data.
           ;;  Return our result.

           (if outer-profile-current
               (setcar (nthcdr 2 outer-profile-current) (current-time))
             )
           the-result
           )
         )))
     )

    ;;--Do the next function on the list.

    (setq funcs (cdr funcs))
    )

  );; profile-setup

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun profile-report ( &optional funcs data )
  "Report on the last profile performed and reset all counters."
  (interactive)

  (setq gc-cons-threshold 100000)
  (setq funcs profile-data)
  (switch-to-buffer-other-window "*profile*")
  (beginning-of-buffer)
  (insert "\n")
  (beginning-of-buffer)
  (insert "Calls Time     Time/Call Name\n")
  (while funcs
    (setq data (car funcs))
    (if (/= 0 (nth 3 data))
        (progn
          (insert (format "%5d %7f %7f  %S\n"
                          (nth 3 data)
                          (nth 4 data)
                          (/ (nth 4 data) (nth 3 data))
                          (nth 1 data)))
          (setcar (nthcdr 3 data) 0)
          (setcar (nthcdr 4 data) 0.0)
          )
      )
    (setq funcs (cdr funcs))
    )
  (beginning-of-buffer)

  );; profile-report

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(profile-setup)
)



(defun af-indent-line-function ()
  "ada-format version of the indent-line-function."

  (let ((starting-point (point-marker)))
    (af-beginning-of-line)
    (af-indent-line)
    (if (< (point) starting-point)
        (goto-char starting-point)
      )
    (set-marker starting-point nil)
    )

  );; af-indent-line-function


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; (define-key ada-mode-map [tab] 'af-indent-line)
; (setq indent-line-function 'af-indent-line-function)

;;; M-; should see if the current line is empty and do a tab-- if it
;;; is.  Although, if the previous line has a -- on it, and that line
;;; is not otherwise empty, perhaps we want to line up with it
;;; instead.  We probably want to do that line-it-up thing regardless?

;;; have an option that does this
;;;    WHEN foo               WHEN foo |
;;;        | bar                  bar |
;;;        | gorp =>              gorp =>
;;; that is, it just indents the choices normally, and then have one that
;;; does this
;;;    WHEN foo               WHEN foo |
;;;         | bar                  bar |
;;;         | gorp =>              gorp =>
;;; that is, it treats WHEN like a ( and indents relative to the first 
;;; non-blank after the WHEN and not relative to the 'W" in WHEN.

;;; Need to add continuation-indent variables for all continuation tokens.
;;; Need to start using the special indentation values as well.
;;; And need a flag for WHEN and ELSIF which indicate that we do *not*
;;; indent relative to the token following them on the same line like we would
;;; for tokens following ( and THEN on the same line.  We also want a flag
;;; that says how to indent the 2nd group element if the first group element
;;; starts on the same line as a previous continue token.  And maybe for begin
;;; tokens as well.  That whole mechanism needs to be customizable.

;;; Need a flag that indicates how far to indent the statement that follows
;;; a label.  And also a loop/block that follows a name:.

;;; Need a flag that indicates whether to indent inside a () relative to the
;;; ( or relative to the start of a tight-expression that precedes the (.
;;; This would presumably only affect what we do when the ( and the following
;;; token are on separate lines.

;;; have an af-make-blank-lines-empty, which, during a region indent, make
;;; blank lines be empty of spaces and tabs

;;; Replace all of those dynamic logior calls with pre-computed constants of
;;; the same value at some point.

(af-debug-check (message "All eval'ed.")) ;; GEB remove after debugging

;;;    (let ((buffer (current-buffer)))          ; GEB
;;;      (switch-to-buffer "*edebug-trace*")
;;;      (end-of-buffer)
;;;      (insert (format "shared-forward: %S\n" shared-forward))
;;;      (insert (format "shared-backward: %S\n" shared-backward))
;;;      (sit-for 0)
;;;      (switch-to-buffer buffer)
;;;      )

(provide 'ada-format)

;;; Local Variables:
;;; comment-column: 46
;;; End:


